//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

package com.aleax.mufu.world2;

/**
 * AABB - rectangular dynamic object
 */

public class AABB extends DynamicObject
{
   /** Half-width of this AABB  */
   final double xw;
   
   /** Half-height of this AABB  */
   final double yw;

   /** Hack - current cell that this object is in */
   TileMapCell cell;
   
   /**
    * @param pos - initial position for this AABB
    * @param xw - half-width for this AABB
    * @param yw - helf-height for this AABB;
    */
   public AABB(Vector2 pos, double xw, double yw)
   {
      super(pos);
      this.xw = Math.abs(xw);
      this.yw = Math.abs(yw);
   }
   
   /** Draws this AABB using the given renderer */
   public void draw(VectorRenderer rend)
   {
      rend.drawAABB(pos, xw, yw);
   }
   
   public void verlet()
   {
      super.integrateVerlet(); 
   }
   
   /**
    *  Tests for a ray intersection with this AABB
    * 
    * this uses an interpretation of woo's method for ray-vs-AABB intersection
    * (i've only heard of woo's method third-hand so i have no idea if this is really the way he suggests)
    * 
    *  -determine closest two sides of AABB to ray origin (by looking at origin position relative to AABB center)
    *  -calculate distance along ray (i.e ray parameter) from ray origin to the lines containing these sides
    *  -the side with the furthest 'distance' is a candidate for intersection; test it with ray-vs-lineseg
    *  
    *  @param out - where to store the intersection with the ray (if there is one)
    *  @param px - x coordinate of the ray origin
    *  @param py - x coordinate of the ray origin
    *  @param dx - x coordinate of point on ray (for direction)
    *  @param dy - x coordinate of point on ray (for direction)
    *  
    *  @return true if the ray intersects this AABB, false otherwise
    */
   
   public boolean testRay(Vector2 out, double px, double py, double dx, double dy)
   {
      //DEBUG
      System.out.println("--- intersecting ray vs. an AABB");
   
      double bx = pos.x;
      double by = pos.y;
      
      //use woo-based ray-vs-AABB
      
      //step 1: find the two AABB edges closest to the ray origin
      //note that we do this 
      double xval, yval;//these store the "potential candidate" sides
      
      if (px < bx)
      {
         //origin is left of box; test min. edge
         xval = bx - xw;
      }
      else
      {
         //origin is right of box; test max. edge
         xval = bx + xw;
      }
   
      if (py < by)
      {
         //origin is above box; test min. edge
         yval = by - yw;
      }
      else
      {
         //origin is below box; test max. edge
         yval = by + yw;
      }
         
      //step 2: calculate the ray parameter from the ray origin to the lines containing the two closest edges to the origin      
      
      double t;//this holds the candidate ray parameter
      double x0,y0,x1,y1;//these hold the coordinates of the cadidate edge/lineseg
      
      if (dx == 0)
      {
         if (dy == 0)
         {
            //big problem..
            System.out.println("Warning!! testRay_AABB was sent a ray with a null direction vector!!");
            return false;
         }
         else
         {
            //the ray is vertical so it can only intersect with y-bounds;
            //prepare lineseg for intersection test
            x0 = bx - xw;
            x1 = bx + xw;
            y0 = y1 = yval;
            t = (yval - py)/dy;
         }
      }
      else if (dy == 0)
      {
            //the ray is horizontal so it can only intersect with x-bounds;
            //prepare lineseg for intersection test
            y0 = by - yw;
            y1 = by + yw;
            x0 = x1 = xval;
            t = (xval - px)/dx;
      }
      else
      {
         //we need to find the potential candidate edge with the greatest ray parameter
         
         double tX = (xval - px)/dx;//calculate ray paramaters; note that we've already made sure
         double tY = (yval - py)/dy;//that div-by-0 can't happen here
   
         if (tX < tY)
         {
            //y-edge is candidate;
            //prepare lineseg for intersection test
            x0 = bx - xw;
            x1 = bx + xw;
            y0 = y1 = yval;
            t = tY;
         }
         else
         {
            //x-edge is candidate;
            //prepare lineseg for intersection test
            y0 = by - yw;
            y1 = by + yw;
            x0 = x1 = xval;
            t = tX;
         }
      }
      
      //if the ray parameter of the intersection betwen ray and line is negative, the ray points away from the
      //line and we know there can't be an intersection
      
      if (0 < t)
      {
         //there MAY be an intersection
      
         //determine boolean result of ray-lineseg intersection test
         //since we need two points to describe the ray, use a point 10units along the ray
         double qx = px + 100*dx;
         double qy = py + 100*dy;
         
         double areaPQ0 = (((qx-px)*(y0-py)) - ((x0-px)*(qy-py)));
         double areaPQ1 = (((qx-px)*(y1-py)) - ((x1-px)*(qy-py)));
         
         if ( (areaPQ0 * areaPQ1) < 0)
         {
            //DEBUG
            System.out.println("intersection found!: " + t);
            
            //signed areas are different signs; intersection is true
                     
            //ray intersects lineseg
            out.x = px + t*dx;
            out.y = py + t*dy;
            return true;
         }
         else
         {
            //DEBUG
            System.out.println("no intersection; areas: " + areaPQ0 + "," + areaPQ1);        
            
            //no intersection
            return false;
         }
      }
      else
      {
         //DEBUG
         System.out.println("ray parameter negative: " + t);
         
         //ray parameter is negative, ray poitns away from AABB and thus no intersection possible
         return false;
      }
   }
   
   public void collideVsWorldBounds()
   {
      Vector2 p = pos;
      
      //collide vs. x-bounds
      //test XMIN
      double dx = demo5.XMIN - (p.x - xw);
      if (0 < dx)
      {
         //object is colliding with XMIN
         reportCollisionVsWorld(dx,0,1,0,null);
      }
      else
      {
         //test XMAX
         dx = (p.x + xw) - demo5.XMAX;
         if (0 < dx)
         {
            //object is colliding with XMAX
            reportCollisionVsWorld(-dx,0,-1,0,null);
         }
      }
      
      //collide vs. y-bounds
      //test YMIN
      double dy = demo5.YMIN - (p.y - yw);
      if (0 < dy)
      {
         //object is colliding with YMIN
         reportCollisionVsWorld(0,dy,0,1,null);
      }
      else
      {
         //test YMAX
         dy = (p.y + yw) - demo5.YMAX;
         if (0 < dy)
         {
            //object is colliding with YMAX
            reportCollisionVsWorld(0,-dy,0,-1,null);
         }
      }  
   }
   
   public void collideVsTile(TileMapCell tile)
   {
      TileMapCell c = tile;
         
      double tx = c.pos.x;
      double ty = c.pos.y;
      double txw = c.xw;
      double tyw = c.yw;
   
      double dx = pos.x - tx;//tile->obj delta
      double px = (txw + xw) - Math.abs(dx);//penetration depth in x
   
      if (0 < px)
      {
         double dy = pos.y - ty;//tile->obj delta
         double py = (tyw + yw) - Math.abs(dy);//pen depth in y
         
         if (0 < py)
         {
            //object may be colliding with tile; call tile-specific collision function
            
            //calculate projection vectors
            if (px < py)
            {
               //project in x
               if (dx < 0)
               {
                  //project to the left
                  px *= -1;
                  py = 0;
               }
               else
               {
                  //proj to right
                  py = 0;
               }
            }
            else
            {           
               //project in y
               if (dy < 0)
               {
                  //project up
                  px = 0;
                  py *= -1;
               }
               else
               {
                  //project down
                  px = 0;
               }
            }
            
            resolveBoxTile(px,py,this,c);
         }
      }
   }
   
   /**
    * this method resolves the collision between an object and a tile,
    * based on the tile type. (x,y) is the  projection vector.
    * this function returns true IF it moves the object by the specified
    * projection vector
    */
   
   public int resolveBoxTile(double x, double y, AABB box, TileMapCell t)
   {
      switch (t.CTYPE)
      {
         case TileMapCell.CTYPE_FULL:
            return projAABB_Full(x,y,box,t);
         case TileMapCell.CTYPE_45DEG:
            return projAABB_45Deg(x,y,box,t);
         case TileMapCell.CTYPE_CONCAVE:
            return projAABB_Concave(x,y,box,t);
         case TileMapCell.CTYPE_CONVEX:
            return projAABB_Convex(x,y,box,t);
         case TileMapCell.CTYPE_22DEGs:
            return projAABB_22DegS(x,y,box,t);
         case TileMapCell.CTYPE_22DEGb:
            return projAABB_22DegB(x,y,box,t);
         case TileMapCell.CTYPE_67DEGs:
            return projAABB_67DegS(x,y,box,t);
         case TileMapCell.CTYPE_67DEGb:
            return projAABB_67DegB(x,y,box,t);
         case TileMapCell.CTYPE_HALF:
            return projAABB_Half(x,y,box,t);            
         default:
            System.out.println("resolveBoxTile() was called with an empty (or unknown) tile!: ID=" + t.ID + " ("+ t.i + "," + t.j + ")");
            return -1;
      }  
   }
   
   public int projAABB_67DegB(double x, double y, AABB obj, TileMapCell t)
   {
      int signx = t.signx;
      int signy = t.signy;
         
      double ox = (obj.pos.x - (signx*obj.xw)) - (t.pos.x + (signx*t.xw));//this gives is the coordinates of the innermost
      double oy = (obj.pos.y - (signy*obj.yw)) - (t.pos.y - (signy*t.yw));//point on the AABB, relative to a point on the slope
                                          
      double sx = t.sx;//get slope unit normal
      double sy = t.sy;
         
      //if the dotprod of (ox,oy) and (sx,sy) is negative, the corner is in the slope
      //and we need toproject it out by the magnitude of the projection of (ox,oy) onto (sx,sy)
      double dp = (ox*sx) + (oy*sy);
      
      if (dp < 0)
      {
         //collision; project delta onto slope and use this to displace the object
         sx *= -dp;//(sx,sy) is now the projection vector
         sy *= -dp;     
            
         double lenN = Math.sqrt(sx*sx + sy*sy);
         double lenP = Math.sqrt(x*x + y*y);   
            
         if (lenP < lenN)
         {
            obj.reportCollisionVsWorld(x,y,x/lenP, y/lenP, t);
   
            return COL_AXIS;
         }
         else
         {     
            obj.reportCollisionVsWorld(sx,sy,t.sx,t.sy,t);
               
            return COL_OTHER;
         }
      }
         
      return COL_NONE;  
   }

   public int projAABB_67DegS(double x, double y, AABB obj, TileMapCell t)
   {
      int signx = t.signx;
      int signy = t.signy;
   
      double px = obj.pos.x - (signx*obj.xw);
      double penX = t.pos.x - px;
      
      if (0 < (penX*signx))
      {
         
         double ox = (obj.pos.x - (signx*obj.xw)) - (t.pos.x - (signx*t.xw));//this gives is the coordinates of the innermost
         double oy = (obj.pos.y - (signy*obj.yw)) - (t.pos.y + (signy*t.yw));//point on the AABB, relative to a point on the slope
   
         double sx = t.sx;//get slope unit normal
         double sy = t.sy;
         
         //if the dotprod of (ox,oy) and (sx,sy) is negative, the corner is in the slope
         //and we need to project it out by the magnitude of the projection of (ox,oy) onto (sx,sy)
         double dp = (ox*sx) + (oy*sy);
         if (dp < 0)
         {
            //collision; project delta onto slope and use this to displace the object
            sx *= -dp;//(sx,sy) is now the projection vector
            sy *= -dp;     
   
            double lenN = Math.sqrt(sx*sx + sy*sy);
            double lenP = Math.sqrt(x*x + y*y);
   
            double aX = Math.abs(penX);
            if (lenP < lenN)
            {
               if (aX < lenP)
               {
                  obj.reportCollisionVsWorld(penX, 0, penX/aX, 0, t);
                  
                  return COL_OTHER;
               }
               else
               {
                  obj.reportCollisionVsWorld(x,y,x/lenP, y/lenP, t);
                  
                  return COL_AXIS;
               }
            }
            else
            {
               if (aX < lenN)
               {
                  obj.reportCollisionVsWorld(penX, 0, penX/aX, 0, t);
                  
                  return COL_OTHER;
               }
               else
               {           
                  obj.reportCollisionVsWorld(sx,sy,t.sx,t.sy,t);
   
                  return COL_OTHER;
               }
            }
         }
      }
      
      //if we've reached this point, no collision has occured
      return COL_NONE;  
   }

   public int projAABB_22DegB(double x, double y, AABB obj, TileMapCell t)
   {
      int signx = t.signx;
      int signy = t.signy;
   
      double ox = (obj.pos.x - (signx*obj.xw)) - (t.pos.x - (signx*t.xw));//this gives is the coordinates of the innermost
      double oy = (obj.pos.y - (signy*obj.yw)) - (t.pos.y + (signy*t.yw));//point on the AABB, relative to a point on the slope
                                    
         
      double sx = t.sx;//get slope unit normal
      double sy = t.sy;
         
      //if the dotprod of (ox,oy) and (sx,sy) is negative, the corner is in the slope
      //and we need toproject it out by the magnitude of the projection of (ox,oy) onto (sx,sy)
      double dp = (ox*sx) + (oy*sy);
      if (dp < 0)
      {
         //collision; project delta onto slope and use this to displace the object
         sx *= -dp;//(sx,sy) is now the projection vector
         sy *= -dp;     
   
         double lenN = Math.sqrt(sx*sx + sy*sy);
         double lenP = Math.sqrt(x*x + y*y);
   
         if (lenP < lenN)
         {
            obj.reportCollisionVsWorld(x,y,x/lenP, y/lenP, t);
               
            return COL_AXIS;
         }
         else
         {     
            obj.reportCollisionVsWorld(sx,sy,t.sx,t.sy,t);
                           
            return COL_OTHER;
         }
      
      }
         
      return COL_NONE;
   }

   public int projAABB_22DegS(double x, double y, AABB obj, TileMapCell t)
   {
      
      int signx = t.signx;
      int signy = t.signy;
      
   
      //first we need to check to make sure we're colliding with the slope at all
      double py = obj.pos.y - (signy*obj.yw);
      double penY = t.pos.y - py;//this is the vector from the innermost point on the box to the highest point on
                        //the tile; if it is positive, this means the box is above the tile and
                        //no collision is occuring
      if (0 < (penY*signy))
      {
         
         double ox = (obj.pos.x - (signx*obj.xw)) - (t.pos.x + (signx*t.xw));//this gives is the coordinates of the innermost
         double oy = (obj.pos.y - (signy*obj.yw)) - (t.pos.y - (signy*t.yw));//point on the AABB, relative to a point on the slope
                                          
         double sx = t.sx;//get slope unit normal
         double sy = t.sy;
         
         //if the dotprod of (ox,oy) and (sx,sy) is negative, the corner is in the slope
         //and we need toproject it out by the magnitude of the projection of (ox,oy) onto (sx,sy)
         double dp = (ox*sx) + (oy*sy);
         
         if (dp < 0)
         {
            //collision; project delta onto slope and use this to displace the object
            sx *= -dp;//(sx,sy) is now the projection vector
            sy *= -dp;     
   
            double lenN = Math.sqrt(sx*sx + sy*sy);
            double lenP = Math.sqrt(x*x + y*y);
            
            double aY = Math.abs(penY);
            
            if (lenP < lenN)
            {
               if (aY < lenP)
               {
                  obj.reportCollisionVsWorld(0, penY, 0, penY/aY, t);
                  
                  return COL_OTHER;
               }
               else
               {
                  obj.reportCollisionVsWorld(x,y,x/lenP, y/lenP, t);
                  
                  return COL_AXIS;
               }
            }
            else
            {
               if (aY < lenN)
               {
                  obj.reportCollisionVsWorld(0, penY, 0, penY/aY, t);
                  
                  return COL_OTHER;
               }
               else
               {
                  obj.reportCollisionVsWorld(sx,sy,t.sx,t.sy,t);
   
                  return COL_OTHER;
               }
            }
         }
      }
      
      //if we've reached this point, no collision has occured
      return COL_NONE;
   }

   public int projAABB_45Deg(double x, double y, AABB obj, TileMapCell t)
   {
      int signx = t.signx;
      int signy = t.signy;
   
      double ox = (obj.pos.x - (signx*obj.xw)) - t.pos.x;//this gives is the coordinates of the innermost
      double oy = (obj.pos.y - (signy*obj.yw)) - t.pos.y;//point on the AABB, relative to the tile center
   
      double sx = t.sx;
      double sy = t.sy;
         
      //if the dotprod of (ox,oy) and (sx,sy) is negative, the corner is in the slope
      //and we need toproject it out by the magnitude of the projection of (ox,oy) onto (sx,sy)
      double dp = (ox*sx) + (oy*sy);
      
      if (dp < 0)
      {
         //collision; project delta onto slope and use this to displace the object
         sx *= -dp;//(sx,sy) is now the projection vector
         sy *= -dp;     
         
         double lenN = Math.sqrt(sx*sx + sy*sy);
         double lenP = Math.sqrt(x*x + y*y);
   
         if (lenP < lenN)
         {
            //project along axis
            obj.reportCollisionVsWorld(x,y,x/lenP, y/lenP, t);
   
            return COL_AXIS;
         }
         else
         {
            //project along slope
            obj.reportCollisionVsWorld(sx,sy,t.sx,t.sy, null);
   
            return COL_OTHER;
         }
      }
      
      return COL_NONE;
   }

   public int projAABB_Convex(double x, double y, AABB obj, TileMapCell t)
   {
      //if distance from "innermost" corner of AABB is less than than tile radius,
      //collision is occuring and we need to project
         
      int signx = t.signx;
      int signy = t.signy;
         
      double ox = (obj.pos.x - (signx*obj.xw)) - (t.pos.x - (signx*t.xw));//(ox,oy) is the vector from the circle center to
      double oy = (obj.pos.y - (signy*obj.yw)) - (t.pos.y - (signy*t.yw));//the AABB
      double len = Math.sqrt(ox*ox + oy*oy);
      
      double twid = t.xw*2;
      double rad = Math.sqrt(twid*twid + 0);//this gives us the radius of a circle centered on the tile's corner and extending to the opposite edge of the tile;
                                 //note that this should be precomputed at compile-time since it's constant
               
      double pen = rad - len;
      
      if (((signx*ox) < 0) || ((signy*oy) < 0))
      {
         //the test corner is "outside" the 1/4 of the circle we're interested in
         double lenP = Math.sqrt(x*x + y*y);
         obj.reportCollisionVsWorld(x,y,x/lenP, y/lenP, t);
            
         return COL_AXIS;//we need to report       
      }     
      else if (0 < pen)
      {
         //project along corner->circle vector
         ox /= len;
         oy /= len;        
         obj.reportCollisionVsWorld(ox*pen, oy*pen, ox, oy, t);
   
         return COL_OTHER;
      }
   
      return COL_NONE;
   }

   public int projAABB_Concave(double x, double y, AABB obj, TileMapCell t)
   {     
      //if distance from "innermost" corner of AABB is further than tile radius,
      //collision is occuring and we need to project
         
      int signx = t.signx;
      int signy = t.signy;
         
      double ox = (t.pos.x + (signx*t.xw)) - (obj.pos.x - (signx*obj.xw));//(ox,oy) is the vector form the innermost AABB corner to the
      double oy = (t.pos.y + (signy*t.yw))- (obj.pos.y - (signy*obj.yw));//circle's center
   
      double twid = t.xw*2;
      double rad = Math.sqrt(twid*twid + 0);//this gives us the radius of a circle centered on the tile's corner and extending to the opposite edge of the tile;
                                 //note that this should be precomputed at compile-time since it's constant
         
      double len = Math.sqrt(ox*ox + oy*oy);
      double pen = len - rad;
      
      if (0 < pen)
      {
         //collision; we need to either project along the axes, or project along corner->circlecenter vector
            
         double lenP = Math.sqrt(x*x + y*y);
         
         if (lenP < pen)
         {
            //it's shorter to move along axis directions
            obj.reportCollisionVsWorld(x,y,x/lenP, y/lenP, t);
               
            return COL_AXIS;
         }
         else
         {
            //project along corner->circle vector
            ox /= len;//len should never be 0, since if it IS 0, rad should be > than len
            oy /= len;//and we should never reach here
            
            obj.reportCollisionVsWorld(ox*pen, oy*pen, ox, oy, t);
               
            return COL_OTHER;
         }
   
      }
   
      return COL_NONE;         
   }

   //signx or signy must be 0; the other must be -1 or 1
   public int projAABB_Half(double x, double y, AABB obj, TileMapCell t)
   {
      //calculate the projection vector for the half-edge, and then 
      //(if collision is occuring) pick the minimum
      
      double sx = t.signx;
      double sy = t.signy;
         
      double ox = (obj.pos.x - (sx*obj.xw)) - t.pos.x;//this gives is the coordinates of the innermost
      double oy = (obj.pos.y - (sy*obj.yw)) - t.pos.y;//point on the AABB, relative to the tile center
         
   
      //we perform operations analogous to the 45deg tile, except we're using 
      //an axis-aligned slope instead of an angled one..
   
      //if the dotprod of (ox,oy) and (sx,sy) is negative, the corner is in the slope
      //and we need toproject it out by the magnitude of the projection of (ox,oy) onto (sx,sy)
      double dp = (ox*sx) + (oy*sy);
      
      if (dp < 0)
      {
         //collision; project delta onto slope and use this to displace the object
         sx *= -dp;//(sx,sy) is now the projection vector
         sy *= -dp;     
            
         double lenN = Math.sqrt(sx*sx + sy*sy);
         double lenP = Math.sqrt(x*x + y*y);
         
         if (lenP < lenN)
         {
            //project along axis; note that we're assuming that this tile is horizontal OR vertical
            //relative to the AABB's current tile, and not diagonal OR the current tile.
            obj.reportCollisionVsWorld(x,y,x/lenP, y/lenP, t);
   
            return COL_AXIS;
         }
         else
         {     
            //note that we could use -= instead of -dp
            obj.reportCollisionVsWorld(sx,sy,t.signx, t.signy, t);
               
            return COL_OTHER;
         }
      }
         
      return COL_NONE;
   }

   public int projAABB_Full(double x, double y, AABB obj, TileMapCell t)
   {
      double l = Math.sqrt(x*x + y*y);
      obj.reportCollisionVsWorld(x,y,x/l,y/l,t);
      
      return COL_AXIS;
   }
   
   //this method detects a collision between the player and an edge-based tilemap,
   //and (if collision is found) calls player.ReportCollisionVsWorld() to report the collision   
   
   public void collideVsTileMap(TileMap tiles)
   {   
      TileMapCell c = tiles.getTile_V(pos);
      
      //store the cell in the box; this is SUPER hacky but will allow ragdoll-vs-object collisions
      this.cell = c;
      
      double tx = c.pos.x;
      double ty = c.pos.y;
      double txw = c.xw;
      double tyw = c.yw;
      
   
      double dx = (pos.x - tx);//tile->obj delta
      double dy = (pos.y - ty);
      
      if (0 < c.ID)
      {
         //current tile is full!! 
         //for now, move object to oldpos; later, we'll need to determine projection direction and resolve
         //the collision
         //System.out.println("CollideAABBvsTileMap() found collision with current tile!!");
      
         double px = (txw + xw) - Math.abs(dx);//penetration depth in x  
         double py = (tyw + yw) - Math.abs(dy);//pen depth in y
         
         //calculate projection vectors
         if (px < py)
         {
            //project in x
            if (dx < 0)
            {
               //project to the left
               px *= -1;
               py = 0;
            }
            else
            {
               //proj to right
               py = 0;
            }
         }
         else
         {
            //project in y
            if (dy < 0)
            {
               //project up
               px = 0;
               py *= -1;
            }
            else
            {
               //project down
               px = 0;
            }
         }
               
         resolveBoxTile(px,py,this,c);
      }
   
      //current tile is empty, test borders
   
      // test in y
      boolean crossV = false;//these flags will indicate whether or not we should collide vs. the diagonal tile
      int hitV = COL_NONE;

      dy = (pos.y - ty);//tile->obj delta
      double py = (Math.abs(dy) + yw) - tyw;//pen depth in y
      
      //DEBUG
      //System.out.println("..raw penetration vector in y: " + py);

      // store edge, neighbor, and direction info
      double eH = 0, pH = 0, signX = 0;
      TileMapCell nH;
            
      double eV = 0, pV = 0, signY = 0;
      TileMapCell nV;
      
      if (0 < py)
      {
         crossV = true;
         
         if (dy < 0)
         {
            eV = c.eU;
            nV = c.nU;
            pV = py;
            signY = 1;
         }
         else
         {
            eV = c.eD;
            nV = c.nD;
            pV = -py;
            signY = -1;
         }
            
         //aabb crosses vertical edge
         if (0 < eV)
         {
            //edge is solid or interesting
            if (eV == TileMapCell.EID_SOLID)
            {
               //we're colliding with a solid edge; resolve right away
               hitV = COL_AXIS;
               reportCollisionVsWorld(0,pV,0,signY,nV);
            }
            else
            {

               //edge is interesting; resolve using tile-specific function
               hitV = resolveBoxTile(0,pV,this,nV);

            }
         }
      }
      
      // test in x
      boolean crossH = false;//these flags will indicate whether or not we should collide vs. the diagonal tile
      int hitH = COL_NONE;
      
      dx = (pos.x - tx);//tile->obj delta
      double px = (Math.abs(dx) + xw) - txw;//penetration depth in x

      //DEBUG
      //System.out.println("..raw penetration vector in x: " + px );

      if (0 < px)
      {
         crossH = true;//aabb crosses horizontal edge
         
         if (dx < 0)
         {
            eH = c.eL;
            nH = c.nL;
            pH = px;
            signX = 1;
         }
         else
         {
            eH = c.eR;
            nH = c.nR;
            pH = -px;
            signX = -1;
         }
      
         //aabb crosses vertical edge
         if (0 < eH)
         {
            //edge is solid or interesting
            if (eH == TileMapCell.EID_SOLID)
            {
               //we're colliding with a solid edge; resolve right away
               hitH = COL_AXIS;
               reportCollisionVsWorld(pH, 0, signX, 0, nH);

            }
            else
            {

               //edge is interesting; resolve using tile-specific function
               hitH = resolveBoxTile(pH,0,this,nH);
            }
         }
      }     
      
      //we should collide vs. the diagonal ONLY if the following are true:
      //-object overlaps cell edge vertically but wasn't moved vertically to resolve this
      //-object overlaps cell edge horizontally but wasn't moved horizontally to resolve this
      //-adx and ady should both NOT be 0; if they ARE 0, the object is in the center of the tile
      //and thus (since objects must be smaller than tiles) should never be able to collide
      //with the diagonal
      
      //note that we can REALLY clean+speed this up! but do it later.
      
      //DEBUG
      //System.out.println("diag?: " + crossH + "," + (!hitH) + "," + crossV + "," + (!hitV));
      
      //the basic reasonning is: if we've already projected a long an axis, that means we projected
      //out of the diagonal tile. otherwise, IF we're touchign the diagonal tile, test for collision.
      
      if ((crossH)&&(hitH != COL_AXIS)&&(crossV)&&(hitV != COL_AXIS))
      {
         //DEBUG
         //System.out.println("TESTING DIAGONAL!");
         
         
         //we didn't collide with horiz or vert neighbors; test the diagonal neighbor
         //NOTE: just as we used the current cell's edges and the h/v neighbor's cell states
         //to test horizontally/vertically, we use the h/v neighbor's edges and the diagonal
         //neighb's cell state to test diagonally
         
         //NOTE: we already DID all these tests, so there _should_ be a way to cache the results
         //instead of doing it all over again.. however, for now we'll just get things working
         //and not worry about speed
         
         dx = (pos.x - tx);//tile->obj delta
         dy = (pos.y - ty);//tile->obj delta         
         px = (Math.abs(dx) + xw) - txw;//penetration depth in x        
         py = (Math.abs(dy) + yw) - tyw;//pen depth in y       
         
         double rx = 0;//potential projection vector
         double ry = 0;
         //boolean hit = false;//flag to indocate if we should resolve collision or not
            
         //double eH, eV;//stores the two diagonal edges we're testing 
         TileMapCell dTile = null;//this should hold a pointer to the diagonal neighbor we're testing against

         //note that we can assume that the object is hitting the diagonal edges. we KNOW this
         //due to the testD, crossH and crossV flags

         //currently this logic is very, very crude and brute-force; 
         //however, once it's working we can worry about speed 
         if ((dx < 0) && (dy < 0))
         {
            //test top-left neighbor

            //DEBUG
            //System.out.println("...test UL: " + c.nU.eL + "," + c.nL.eU);

            eH = c.nU.eL;
            eV = c.nL.eU;
            dTile = c.nU.nL;
            
         }
         else if ((dx < 0) && (0 < dy))
         {
            //test bottom-left neighbor
            
            //DEBUG
            //System.out.println("...test DL: " + c.nD.eL + "," + c.nL.eD);
            
            eH = c.nD.eL;
            eV = c.nL.eD;
            dTile = c.nD.nL;
            
         }        
         else if ((0 < dx) && (0 < dy))
         {
            //test bottom-right neighbor
            
            //DEBUG
            //System.out.println("...test DR: " + c.nD.eR + "," + c.nR.eD);
         
            eH = c.nD.eR;
            eV = c.nR.eD;
            dTile = c.nD.nR;

         }        
         else if ((0 < dx) && (dy < 0))
         {
            //test top-right neighbor
            
            //DEBUG
            //System.out.println("...test UR: " + c.nU.eR + "," + c.nR.eU);
            
            eH = c.nU.eR;
            eV = c.nR.eU;
            dTile = c.nU.nR;
            
         }
         else
         {
            //object is in exact center of current cell.. we should never have reached here because
            //they could never have touched any cell edges
            //System.out.println("WARNING!! we're colliding vs. diagonal but the object is in the dead center of it's cell!!");
         }
         
         //we give precedence to "interesting" edges OVER solid edges, because interesting edges are
         //guaranteed to project the object by AT MOST the same as the solid edges, and at least a lot less.

         //our collision resolution routines expect ONE of the axis penetration components to
         //be zero; this means we need to find the smallest axial projection vector right now..
         double nx,ny;
         
         if (px < py)
         {
            ry = ny = 0;
            if (dx < 0)
            {
               rx = px;
               nx = 1;
            }
            else
            {
               rx = -px;
               nx = -1;
            }           
         }
         else
         {
            rx = nx = 0;
            if (dy < 0)
            {
               ry = py;
               ny = 1;
            }
            else
            {
               ry = -py;
               ny = -1;
            }
         }

         if (0 < eH)
         {
            //horizontal edge is non-empty
            
            if (0 < eV)
            {
               //both edges are non-empty
               if (eH == TileMapCell.EID_SOLID)
               {
                  if (eV == TileMapCell.EID_SOLID)
                  {
                     //both edges are solid; we need to project along the shortest axis
                     reportCollisionVsWorld(rx,ry,nx,ny, dTile);
                  }
                  else
                  {
                     //vert edge is intersecting horiz is solid; investigate vert edge
                     int temp = resolveBoxTile(rx,ry,this,dTile);

                     if ( temp == COL_NONE)
                     {
                        //we didn't collide with the vertical "interesting" edge, so we're still
                        //colliding with the solid edge; resolve!
                        reportCollisionVsWorld(pH,0,signX,0, dTile);
                     }
                  }
               }
               else
               {
                  if (eV == TileMapCell.EID_SOLID)
                  {
                     //vert edge is solid, horiz is interesting; investigate horiz
                     int temp = resolveBoxTile(rx,ry,this,dTile);

                     if (temp == COL_NONE)
                     {
                        //we didn't collide with the horiz "interesting" edge, so we're still
                        //colliding with the solid edge; resolve!
                        reportCollisionVsWorld(0,pV,0,signY,dTile);
                     }
                  }
                  else
                  {
                     //one or both of the edges is interesting; investigate/project
                     resolveBoxTile(rx,ry,this,dTile);
                  }
               }
            }
            else
            {
               //vert is empty, horiz is not; collide vs. horiz
               if (eH == TileMapCell.EID_SOLID)
               {
                  reportCollisionVsWorld(pH,0,signX,0,dTile);
               }
               else
               {
                  resolveBoxTile(rx,ry,this,dTile);
               }
            }
         }
         else if (0 < eV)
         {
            //horiz is empty, vertical is not; collide vs. vertical
            if (eV == TileMapCell.EID_SOLID)
            {
               reportCollisionVsWorld(0,pV,0,signY, dTile);
            }
            else
            {
               resolveBoxTile(rx,ry,this,dTile);
            }                    
         }
         else
         {
            //both edges are empty; do nothing
         }
      }
   }   
}
