//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

package com.aleax.mufu.world2;

/**
 * Circular dynamic object 
 */

public class Circle extends DynamicObject
{
   private static final double SQRT2 = Math.sqrt(2);
   
   /** Radius of this Circle */
   final double r;
   
   public Circle(Vector2 pos, double r)
   {
      super(pos);
      this.r = Math.abs(r);
   }
   
   public void draw(VectorRenderer rend)
   {
      rend.drawCircle(pos, r);
   }
   
   public void verlet()
   {
      integrateVerlet(); 
   }
   
   /**
    *  Tests for a ray intersection with this circle
    * 
    *  @param out - where to store the intersection with the ray (if there is one)
    *  @param px - x coordinate of the ray origin
    *  @param py - x coordinate of the ray origin
    *  @param dx - x coordinate of point on ray (for direction)
    *  @param dy - x coordinate of point on ray (for direction)
    *  
    *  @return true if the ray intersects this circle, false otherwise
    */
   
   public boolean testRay(Vector2 out, double px, double py, double dx, double dy)
   {
      //DEBUG
      System.out.println("--- intersecting ray vs. a circle");
      
      //we use gamasutra swept-circle code to test for collision;
      //we ASSUME that the ray doesn't collide initially..
            
      double abx = px - pos.x;//(abx,aby) should now be the vector from the circle's center to the
      double aby = py - pos.y;//ray position px,py
            
      //DEBUG
      System.out.println("--- abx,aby: " + abx  + "," + aby + " dx,dy: " + dx + "," + dy);
            
      double a = (dx*dx)+(dy*dy);// const SCALAR a = vab.dot(vab); u*u coefficient
      double b = 2*( (dx*abx) + (dy*aby) );// const SCALAR b = 2*vab.dot(AB); u coefficient
      double rad = r;
      double c = ((abx*abx)+(aby*aby)) - (rad*rad);// const SCALAR c = AB.dot(AB) - rab*rab; constant term
            
      double q = b*b - (4*a*c);// const SCALAR q = b*b - 4*a*c; 
      
      if (0 <= q)
      {
         //if ( q >= 0 )
      
         double sq = Math.sqrt(q);// const SCALAR sq = sqrt(q);
         double d = 1/(2*a);// const SCALAR d = 1 / (2*a);
         double u0 = (-b + sq) * d;
         double u1 = (-b - sq) * d;

         //DEBUG
         System.out.println("roots: " + u0 + "," + u1);

         double u;
         
         if (u1 < 0)
         {
            //u1 points the wrong direction along the line containing the ray..
            if (u0 < 0)
            {
               //both roots are negative -- the circle is "behind" the ray
               return false;
            }
            else
            {
               //u0 is the correct root
               u = u0;
            }
         }
         else if (u0 < 0)
         {
            //u0 points the wrong direction along the line containing the ray..
            //u1 is the correct root
            u = u1;
         }
         else
         {
            //both roots point the right way; choose the first one..
            if ( u1 < u0 )
            {              
                  //u1 is time of first collision
                  u = u1;
            }
            else
            {              
                  //u0 is time of first collision; u1 is our solution
                  u = u0;
            }
         }
         
         out.x = px + u*dx;
         out.y = py + u*dy;
         
         return true;

      }
      else
      {
         //DEBUG
         System.out.println("complex: " + q);
         
         //complex roots
         return false;
      }     
   }
   
   public void collideVsWorldBounds()
   {
      Vector2 p = this.pos;
      
      //collide vs. x-bounds
      //test XMIN
      double dx = demo5.XMIN - (p.x - r);
      if (0 < dx)
      {
         //object is colliding with XMIN
         reportCollisionVsWorld(dx,0,1,0,null);
      }
      else
      {
         //test XMAX
         dx = (p.x + r) - demo5.XMAX;
         if (0 < dx)
         {
            //object is colliding with XMAX
            reportCollisionVsWorld(-dx,0,-1,0,null);
         }
      }
      
      //collide vs. y-bounds
      //test YMIN
      double dy = demo5.YMIN - (p.y - r);
      if (0 < dy)
      {
         //object is colliding with YMIN
         reportCollisionVsWorld(0,dy,0,1,null);
      }
      else
      {
         //test YMAX
         dy = (p.y + r) - demo5.YMAX;
         if (0 < dy)
         {
            //object is colliding with YMAX
            reportCollisionVsWorld(0,-dy,0,-1,null);
         }
      }  
   }
   
   public void collideVsTile(TileMapCell tile)
   {
      TileMapCell c = tile;
         
      double tx = c.pos.x;
      double ty = c.pos.y;
      double txw = c.xw;
      double tyw = c.yw;
      
      double dx = pos.x - tx;//tile->obj delta
      double px = (txw + r) - Math.abs(dx);//penetration depth in x
   
      if (0 < px)
      {
         double dy = pos.y - ty;//tile->obj delta
         double py = (tyw + r) - Math.abs(dy);//pen depth in y
         
         if (0 < py)
         {
            //object may be colliding with tile
                     
            //determine grid/voronoi region of circle center
            double oH = 0;
            double oV = 0;
            
            if (dx < -txw)
            {
               //circle is on left side of tile
               oH = -1;
            }
            else if (txw < dx)
            {
               //circle is on right side of tile
               oH = 1;
            }
         
            if (dy < -tyw)
            {
               //circle is on top side of tile
               oV = -1;
            }
            else if (tyw < dy)
            {
               //circle is on bottom side of tile
               oV = 1;
            }        
   
            resolveCircleTile(px,py,oH,oV,this,c);   
         }
      }     
   }   
   
   /**
    * this function resolves the collision between an object and a tile,
    * based on the tile type. (x,y) is the  projection vector.
    * this function returns true IF it moves the object by the specified
    * projection vector
    * 
    * NOTE: the cell offset vector (oH,oV) relative to the _tile_, i.e if the
    * circle is being tested against the tile to its left, oH = 1
    */
   
   public int resolveCircleTile(double x, double y, double oH, double oV, Circle obj, TileMapCell t)
   {
      switch(t.CTYPE)
      {
         case TileMapCell.CTYPE_FULL:
            return projCircle_Full(x,y,oH,oV,obj,t);
         case TileMapCell.CTYPE_45DEG:
            return projCircle_45Deg(x,y,oH,oV,obj,t);
         case TileMapCell.CTYPE_CONCAVE:
            return projCircle_Concave(x,y,oH,oV,obj,t);
         case TileMapCell.CTYPE_CONVEX:
            return projCircle_Convex(x,y,oH,oV,obj,t);
         case TileMapCell.CTYPE_22DEGs:
            return projCircle_22DegS(x,y,oH,oV,obj,t);
         case TileMapCell.CTYPE_22DEGb:
            return projCircle_22DegB(x,y,oH,oV,obj,t);
         case TileMapCell.CTYPE_67DEGs:
            return projCircle_67DegS(x,y,oH,oV,obj,t);
         case TileMapCell.CTYPE_67DEGb:
            return projCircle_67DegB(x,y,oH,oV,obj,t);
         case TileMapCell.CTYPE_HALF:
            return projCircle_Half(x,y,oH,oV,obj,t);
            
         default:
            System.out.println("ResolveCircleTile() was called with an empty (or unknown) tile!: ID=" + t.ID + " ("+ t.i + "," + t.j + ")");
            return -1;
      }
   }
      
   public int projCircle_67DegB(double x, double y, double oH, double oV, Circle obj, TileMapCell t)
   {
      //if we're colliding diagonally:
      //  -if we're in the cell pointed at by the normal, collide vs slope, else
      //  collide vs. the appropriate corner/vertex
      //
      //if obj is in this tile: collide as with aabb
      //
      //if obj is horiz or vertical neighbor AGAINST the slope: collide with edge
      //
      //if obj is vert neighb in direction of slope: collide vs. slope or vertex or halfedge
      //
      //if obj is horiz neighb in direction of slope: collide vs. slope or vertex
   
      int signx = t.signx;
      int signy = t.signy;
   
      if (oH == 0)
      {
         if (oV == 0)
         {
            //colliding with current cell
   
            double sx = t.sx;
            double sy = t.sy;
      
            double r = obj.r;
            double ox = (obj.pos.x - (sx*r)) - (t.pos.x + (signx*t.xw));//this gives is the coordinates of the innermost
            double oy = (obj.pos.y - (sy*r)) - (t.pos.y - (signy*t.yw));//point on the AABB, relative to a point on the slope
         
            //if the dotprod of (ox,oy) and (sx,sy) is negative, the point on the circle is in the slope
            //and we need toproject it out by the magnitude of the projection of (ox,oy) onto (sx,sy)
            double dp = (ox*sx) + (oy*sy);
                  
            if (dp < 0)
            {
               //collision; project delta onto slope and use this to displace the object
               sx *= -dp;//(sx,sy) is now the projection vector
               sy *= -dp;     
                        
               double lenN = Math.sqrt(sx*sx + sy*sy);
               double lenP;
               
               //find the smallest axial projection vector
               if (x < y)
               {              
                  //penetration in x is smaller
                  lenP = x;
                  y = 0;   
                  //get sign for projection along x-axis    
                  if ((obj.pos.x - t.pos.x) < 0)
                  {
                     x *= -1;
                  }
               }
               else
               {     
                  //penetration in y is smaller
                  lenP = y;
                  x = 0;   
                  //get sign for projection along y-axis    
                  if ((obj.pos.y - t.pos.y)< 0)
                  {
                     y *= -1;
                  }        
               }
      
               if (lenP < lenN)
               {
                  obj.reportCollisionVsWorld(x,y,x/lenP, y/lenP, t);
                  
                  return COL_AXIS;
               }
               else
               {
                  obj.reportCollisionVsWorld(sx, sy, t.sx, t.sy, t);
                  
                  return COL_OTHER;
               }
      
            }              
         }
         else
         {
            //colliding vertically
         
            if ((signy*oV) < 0)
            {
               //colliding with face/edge
               obj.reportCollisionVsWorld(0, y*oV, 0, oV, t);
   
               return COL_AXIS;
            }
            else
            {
               //colliding with edge, slope, or vertex
            
               double ox = obj.pos.x - t.pos.x;//this gives is the coordinates of the innermost
               double oy = obj.pos.y - (t.pos.y + (signy*t.yw));//point on the circle, relative to the closest tile vert   
                  
               if ((ox*signx) < 0)
               {
                  //we're colliding with the halfface
                  obj.reportCollisionVsWorld(0, y*oV, 0, oV, t);
   
                  return COL_AXIS;        
               }
               else
               {
                  //colliding with the vertex or slope
   
                  double sx = t.sx;
                  double sy = t.sy;
                              
                  //if the component of (ox,oy) parallel to the normal's righthand normal
                  //has the same sign as the slope of the slope (the sign of the slope's slope is signx*signy)
                  //then we project by the vertex, otherwise by the slope.
                  //note that this is simply a VERY tricky/weird method of determining 
                  //if the circle is in side the slope/face's voronio region, or that of the vertex.                                  
                  double perp = (ox*-sy) + (oy*sx);
                  
                  if (0 < (perp*signx*signy))
                  {
                     //collide vs. vertex
                     double len = Math.sqrt(ox*ox + oy*oy);
                     double pen = obj.r - len;
                     
                     if (0 < pen)
                     {
                        //note: if len=0, then perp=0 and we'll never reach here, so don't worry about div-by-0
                        ox /= len;
                        oy /= len;
         
                        obj.reportCollisionVsWorld(ox*pen, oy*pen, ox, oy, t);
                        
                        return COL_OTHER;
                     }              
                  }
                  else
                  {
                     //collide vs. slope
                        
                     //if the component of (ox,oy) parallel to the normal is less than the circle radius, we're
                     //penetrating the slope. note that this method of penetration calculation doesn't hold
                     //in general (i.e it won't work if the circle is in the slope), but works in this case
                     //because we know the circle is in a neighboring cell
                     double dp = (ox*sx) + (oy*sy);
                     double pen = obj.r - Math.abs(dp);//note: we don't need the abs because we know the dp will be positive, but just in case..
                     
                     if (0 < pen)
                     {
                        //collision; circle out along normal by penetration amount
                        obj.reportCollisionVsWorld(sx*pen, sy*pen, sx, sy, t);
                        
                        return COL_OTHER;
                     }
                  }  
               }
            }
         }
      }
      else if (oV == 0)
      {
         //colliding horizontally
            
         if ((signx*oH) < 0)
         {
            //colliding with face/edge
            obj.reportCollisionVsWorld(x*oH, 0, oH, 0, t);
            
            return COL_AXIS;
         }
         else
         {
            //we could only be colliding vs the slope OR a vertex
            //look at the vector form the closest vert to the circle to decide
   
            double slen = Math.sqrt(2*2 + 1*1);//the raw slope is (-2,-1)
            double sx = (signx*2) / slen;//get slope _unit_ normal;
            double sy = (signy*1) / slen;//raw RH normal is (1,-2)
               
            double ox = obj.pos.x - (t.pos.x + (signx*t.xw));//this gives is the coordinates of the innermost
            double oy = obj.pos.y - (t.pos.y - (signy*t.yw));//point on the circle, relative to the closest tile vert   
   
            //if the component of (ox,oy) parallel to the normal's righthand normal
            //has the same sign as the slope of the slope (the sign of the slope's slope is signx*signy)
            //then we project by the slope, otherwise by the vertex.
            //note that this is simply a VERY tricky/weird method of determining 
            //if the circle is in side the slope/face's voronio region, or that of the vertex.                                  
            double perp = (ox*-sy) + (oy*sx);
            
            if ((perp*signx*signy) < 0)
            {
               //collide vs. vertex
               double len = Math.sqrt(ox*ox + oy*oy);
               double pen = obj.r - len;
               
               if (0 < pen)
               {
                  //note: if len=0, then perp=0 and we'll never reach here, so don't worry about div-by-0
                  ox /= len;
                  oy /= len;
   
                  obj.reportCollisionVsWorld(ox*pen, oy*pen, ox, oy, t);
                  
                  return COL_OTHER;
               }              
            }
            else
            {
               //collide vs. slope
                  
               //if the component of (ox,oy) parallel to the normal is less than the circle radius, we're
               //penetrating the slope. note that this method of penetration calculation doesn't hold
               //in general (i.e it won't work if the circle is in the slope), but works in this case
               //because we know the circle is in a neighboring cell
               double dp = (ox*sx) + (oy*sy);
               double pen = obj.r - Math.abs(dp);//note: we don't need the abs because we know the dp will be positive, but just in case..
               
               if (0 < pen)
               {
                  //collision; circle out along normal by penetration amount
                  obj.reportCollisionVsWorld(sx*pen, sy*pen, t.sx, t.sy, t);
                  
                  return COL_OTHER;
               }
            }
         }
      }
      else
      {
         //colliding diagonally
         if ( 0 < ((signx*oH) + (signy*oV)) ) 
         {
            //the dotprod of slope normal and cell offset is strictly positive,
            //therefore obj is in the diagonal neighb pointed at by the normal.
            
            //collide vs slope
   
            double sx = t.sx;
            double sy = t.sy;
      
            double r = obj.r;
            double ox = (obj.pos.x - (sx*r)) - (t.pos.x + (signx*t.xw));//this gives is the coordinates of the innermost
            double oy = (obj.pos.y - (sy*r)) - (t.pos.y - (signy*t.yw));//point on the circle, relative to a point on the slope
         
            //if the dotprod of (ox,oy) and (sx,sy) is negative, the point on the circle is in the slope
            //and we need toproject it out by the magnitude of the projection of (ox,oy) onto (sx,sy)
            double dp = (ox*sx) + (oy*sy);
                  
            if (dp < 0)
            {
               //collision; project delta onto slope and use this to displace the object  
               //(sx,sy)*-dp is the projection vector
   
               obj.reportCollisionVsWorld(-sx*dp, -sy*dp, t.sx, t.sy, t);
   
               return COL_OTHER;
            }
            return COL_NONE;
         }
         else
         {
            
            //collide vs the appropriate vertex
            double vx = t.pos.x + (oH*t.xw);
            double vy = t.pos.y + (oV*t.yw);
            
            double dx = obj.pos.x - vx;//calc vert->circle vector    
            double dy = obj.pos.y - vy;
            
            double len = Math.sqrt(dx*dx + dy*dy);
            double pen = obj.r - len;
            
            if (0 < pen)
            {
               //vertex is in the circle; project outward
               if (len == 0)
               {
                  //project out by 45deg
                  dx = oH / SQRT2;
                  dy = oV / SQRT2;
               }
               else
               {
                  dx /= len;
                  dy /= len;
               }
   
               obj.reportCollisionVsWorld(dx*pen, dy*pen, dx, dy, t);
   
               return COL_OTHER;
            }
               
         }     
      }
      
      return COL_NONE;
   }
   
   public int projCircle_67DegS(double x, double y, double oH, double oV, Circle obj, TileMapCell t)
   {
      //if the object is in a cell pointed at by signx, no collision will ever occur
      //otherwise,
      //
      //if we're colliding diagonally:
      //  -collide vs. the appropriate vertex
      //if obj is in this tile: collide vs slope or vertex or axis
      //if obj is vert neighb in direction of slope: collide vs. slope or vertex
      //if obj is vert neighb against the slope:
      //   if (distance in y from circle to 90deg corner of tile < 1/2 tileheight, collide vs. face)
      //   else(collide vs. corner of slope) (vert collision with a non-grid-aligned vert)
      //if obj is horiz neighb against direction of slope: collide vs. face
   
      int signx = t.signx;
      int signy = t.signy;
   
      if (0 < (signx*oH))
      {
         //object will never collide vs tile, it can't reach that far
   
         return COL_NONE;
      }
      else if (oH == 0)
      {
         if (oV == 0)
         {
            //colliding with current tile
            //we could only be colliding vs the slope OR a vertex
            //look at the vector form the closest vert to the circle to decide
      
            double sx = t.sx;
            double sy = t.sy;
            
            double r = obj.r;
            double ox = obj.pos.x - t.pos.x;//this gives is the coordinates of the innermost
            double oy = obj.pos.y - (t.pos.y - (signy*t.yw));//point on the circle, relative to the tile corner   
         
            //if the component of (ox,oy) parallel to the normal's righthand normal
            //has the same sign as the slope of the slope (the sign of the slope's slope is signx*signy)
            //then we project by the normal or axis, otherwise by the corner/vertex
            //note that this is simply a VERY tricky/weird method of determining 
            //if the circle is in side the slope/face's voronoi region, or that of the vertex.
               
            double perp = (ox*-sy) + (oy*sx);
            
            if ((perp*signx*signy) < 0)
            {
               //collide vs. vertex
               double len = Math.sqrt(ox*ox + oy*oy);
               double pen = r - len;
               
               if (0 < pen)
               {
                  //note: if len=0, then perp=0 and we'll never reach here, so don't worry about div-by-0
                  ox /= len;
                  oy /= len;
   
                  obj.reportCollisionVsWorld(ox*pen, oy*pen, ox, oy, t);
                  return COL_OTHER;
               }              
            }
            else
            {
               //collide vs. slope or vs axis
               ox -= r*sx;//this gives us the vector from  
               oy -= r*sy;//a point on the slope to the innermost point on the circle
         
               //if the dotprod of (ox,oy) and (sx,sy) is negative, the point on the circle is in the slope
               //and we need toproject it out by the magnitude of the projection of (ox,oy) onto (sx,sy)
               double dp = (ox*sx) + (oy*sy);
               
               if (dp < 0)
               {
                  //collision; project delta onto slope and use this to displace the object
                  sx *= -dp;//(sx,sy) is now the projection vector
                  sy *= -dp;     
                     
                  double lenN = Math.sqrt(sx*sx + sy*sy);
                  double lenP;
            
                  //find the smallest axial projection vector
                  if (x < y)
                  {              
                     //penetration in x is smaller
                     lenP = x;
                     y = 0;   
                     //get sign for projection along x-axis    
                     if ((obj.pos.x - t.pos.x) < 0)
                     {
                        x *= -1;
                     }
                  }
                  else
                  {     
                     //penetration in y is smaller
                     lenP = y;
                     x = 0;   
                     //get sign for projection along y-axis    
                     if ((obj.pos.y - t.pos.y)< 0)
                     {
                        y *= -1;
                     }        
                  }
   
                  if (lenP < lenN)
                  {
                     obj.reportCollisionVsWorld(x,y,x/lenP, y/lenP, t);
                     
                     return COL_AXIS;
                  }
                  else
                  {     
                     obj.reportCollisionVsWorld(sx,sy,t.sx,t.sy,t);
                     
                     return COL_OTHER;
                  }  
               }
            }
            
         }
         else
         {
            //colliding vertically
            
            if ((signy*oV) < 0)
            {
               //colliding with face/edge OR with corner of wedge, depending on our position vertically
                  
               //collide vs. vertex
               //get diag vertex position
               double vx = t.pos.x;
               double vy = t.pos.y - (signy*t.yw);
                     
               double dx = obj.pos.x - vx;//calc vert->circle vector    
               double dy = obj.pos.y - vy;
                     
               if ((dx*signx) < 0)
               {  
                  //colliding vs face
                  obj.reportCollisionVsWorld(0, y*oV, 0, oV, t);
                  
                  return COL_AXIS;              
               }
               else
               {
                  //colliding vs. vertex
                     
                  double len = Math.sqrt(dx*dx + dy*dy);
                  double pen = obj.r - len;
                  
                  if (0 < pen)
                  {
                     //vertex is in the circle; project outward
                     if (len == 0)
                     {
                        //project out by 45deg
                        dx = oH / SQRT2;
                        dy = oV / SQRT2;
                     }
                     else
                     {
                        dx /= len;
                        dy /= len;
                     }
                        
                     obj.reportCollisionVsWorld(dx*pen, dy*pen, dx, dy, t);
                     
                     return COL_OTHER;
                  }
               }
            }
            else
            {
               //we could only be colliding vs the slope OR a vertex
               //look at the vector form the closest vert to the circle to decide
         
               double sx = t.sx;
               double sy = t.sy;
                  
               double ox = obj.pos.x - (t.pos.x - (signx*t.xw));//this gives is the coordinates of the innermost
               double oy = obj.pos.y - (t.pos.y + (oV*t.yw));//point on the circle, relative to the closest tile vert   
         
               //if the component of (ox,oy) parallel to the normal's righthand normal
               //has the same sign as the slope of the slope (the sign of the slope's slope is signx*signy)
               //then we project by the vertex, otherwise by the normal.
               //note that this is simply a VERY tricky/weird method of determining 
               //if the circle is in side the slope/face's voronio region, or that of the vertex.                                  
               double perp = (ox*-sy) + (oy*sx);
               
               if (0 < (perp*signx*signy))
               {
                  //collide vs. vertex
                  double len = Math.sqrt(ox*ox + oy*oy);
                  double pen = obj.r - len;
                  
                  if (0 < pen)
                  {
                     //note: if len=0, then perp=0 and we'll never reach here, so don't worry about div-by-0
                     ox /= len;
                     oy /= len;
   
                     obj.reportCollisionVsWorld(ox*pen, oy*pen, ox, oy, t);
                     
                     return COL_OTHER;
                  }              
               }
               else
               {
                  //collide vs. slope
                        
                  //if the component of (ox,oy) parallel to the normal is less than the circle radius, we're
                  //penetrating the slope. note that this method of penetration calculation doesn't hold
                  //in general (i.e it won't work if the circle is in the slope), but works in this case
                  //because we know the circle is in a neighboring cell
                  double dp = (ox*sx) + (oy*sy);
                  double pen = obj.r - Math.abs(dp);//note: we don't need the abs because we know the dp will be positive, but just in case..            
   
                  if (0 < pen)
                  {
                     //collision; circle out along normal by penetration amount
                     obj.reportCollisionVsWorld(sx*pen, sy*pen, t.sx, t.sy, t);
                     
                     return COL_OTHER;
                  }
               }
            }
         }     
      }
      else if (oV == 0)
      {
         //colliding horizontally; we can assume that (signy*oV) < 0
         //due to the first conditional far above
   
            obj.reportCollisionVsWorld(x*oH, 0, oH, 0, t);
            
            return COL_AXIS;
      }
      else
      {     
         //colliding diagonally; due to the first conditional above,
         //obj is vertically offset against slope, and offset in either direction horizontally
   
         //collide vs. vertex
         //get diag vertex position
         double vx = t.pos.x + (oH*t.xw);
         double vy = t.pos.y + (oV*t.yw);
            
         double dx = obj.pos.x - vx;//calc vert->circle vector    
         double dy = obj.pos.y - vy;
            
         double len = Math.sqrt(dx*dx + dy*dy);
         double pen = obj.r - len;
         
         if (0 < pen)
         {
            //vertex is in the circle; project outward
            if (len == 0)
            {
               //project out by 45deg
               dx = oH / SQRT2;
               dy = oV / SQRT2;
            }
            else
            {
               dx /= len;
               dy /= len;
            }
   
            obj.reportCollisionVsWorld(dx*pen, dy*pen, dx, dy, t);
            
            return COL_OTHER;
         }
      }
   
      return COL_NONE;
   
   }
   
   public int projCircle_22DegB(double x, double y, double oH, double oV, Circle obj, TileMapCell t)
   {
   
      //if we're colliding diagonally:
      //  -if we're in the cell pointed at by the normal, collide vs slope, else
      //  collide vs. the appropriate corner/vertex
      //
      //if obj is in this tile: collide as with aabb
      //
      //if obj is horiz or vertical neighbor AGAINST the slope: collide with edge
      //
      //if obj is horiz neighb in direction of slope: collide vs. slope or vertex or edge
      //
      //if obj is vert neighb in direction of slope: collide vs. slope or vertex
   
      int signx = t.signx;
      int signy = t.signy;
   
      if (oH == 0)
      {
         if (oV == 0)
         {
            //colliding with current cell
   
            double sx = t.sx;
            double sy = t.sy;
      
            double r = obj.r;
            double ox = (obj.pos.x - (sx*r)) - (t.pos.x - (signx*t.xw));//this gives is the coordinates of the innermost
            double oy = (obj.pos.y - (sy*r)) - (t.pos.y + (signy*t.yw));//point on the AABB, relative to a point on the slope
         
            //if the dotprod of (ox,oy) and (sx,sy) is negative, the point on the circle is in the slope
            //and we need toproject it out by the magnitude of the projection of (ox,oy) onto (sx,sy)
            double dp = (ox*sx) + (oy*sy);
                  
            if (dp < 0)
            {
               //collision; project delta onto slope and use this to displace the object
               sx *= -dp;//(sx,sy) is now the projection vector
               sy *= -dp;     
                        
               double lenN = Math.sqrt(sx*sx + sy*sy);
               double lenP;
               
               //find the smallest axial projection vector
               if (x < y)
               {              
                  //penetration in x is smaller
                  lenP = x;
                  y = 0;   
                  //get sign for projection along x-axis    
                  if ((obj.pos.x - t.pos.x) < 0)
                  {
                     x *= -1;
                  }
               }
               else
               {     
                  //penetration in y is smaller
                  lenP = y;
                  x = 0;   
                  //get sign for projection along y-axis    
                  if ((obj.pos.y - t.pos.y)< 0)
                  {
                     y *= -1;
                  }        
               }
      
               if (lenP < lenN)
               {
                  obj.reportCollisionVsWorld(x, y, x/lenP, y/lenP, t);
                  
                  return COL_AXIS;
               }
               else
               {        
                  obj.reportCollisionVsWorld(sx, sy, t.sx, t.sy, t);
            
                  return COL_OTHER;
               }  
            }              
         }
         else
         {
            //colliding vertically
            
            if ((signy*oV) < 0)
            {
               //colliding with face/edge
               obj.reportCollisionVsWorld(0, y*oV, 0, oV, t);
               
               return COL_AXIS;
            }
            else
            {
               //we could only be colliding vs the slope OR a vertex
               //look at the vector form the closest vert to the circle to decide
   
               double sx = t.sx;
               double sy = t.sy;
               
               double ox = obj.pos.x - (t.pos.x - (signx*t.xw));//this gives is the coordinates of the innermost
               double oy = obj.pos.y - (t.pos.y + (signy*t.yw));//point on the circle, relative to the closest tile vert   
   
               //if the component of (ox,oy) parallel to the normal's righthand normal
               //has the same sign as the slope of the slope (the sign of the slope's slope is signx*signy)
               //then we project by the vertex, otherwise by the normal.
               //note that this is simply a VERY tricky/weird method of determining 
               //if the circle is in side the slope/face's voronio region, or that of the vertex.                                  
               double perp = (ox*-sy) + (oy*sx);
               
               if (0 < (perp*signx*signy))
               {
                  //collide vs. vertex
                  double len = Math.sqrt(ox*ox + oy*oy);
                  double pen = obj.r - len;
                  
                  if (0 < pen)
                  {
                     //note: if len=0, then perp=0 and we'll never reach here, so don't worry about div-by-0
                     ox /= len;
                     oy /= len;
   
                     obj.reportCollisionVsWorld(ox*pen, oy*pen, ox, oy, t);
                     
                     return COL_OTHER;
                  }              
               }
               else
               {
                  //collide vs. slope
                  
                  //if the component of (ox,oy) parallel to the normal is less than the circle radius, we're
                  //penetrating the slope. note that this method of penetration calculation doesn't hold
                  //in general (i.e it won't work if the circle is in the slope), but works in this case
                  //because we know the circle is in a neighboring cell
                  double dp = (ox*sx) + (oy*sy);
                  double pen = obj.r - Math.abs(dp);//note: we don't need the abs because we know the dp will be positive, but just in case..
                  
                  if (0 < pen)
                  {
                     //collision; circle out along normal by penetration amount
                     obj.reportCollisionVsWorld(sx*pen, sy*pen,sx, sy, t);
                     
                     return COL_OTHER;
                  }
               }
            }
         }
      }
      else if (oV == 0)
      {
         //colliding horizontally
         
         if ((signx*oH) < 0)
         {
            //colliding with face/edge
            obj.reportCollisionVsWorld(x*oH, 0, oH, 0, t);
            
            return COL_AXIS;
         }
         else
         {
            //colliding with edge, slope, or vertex
         
            double ox = obj.pos.x - (t.pos.x + (signx*t.xw));//this gives is the coordinates of the innermost
            double oy = obj.pos.y - t.pos.y;//point on the circle, relative to the closest tile vert  
               
            if ((oy*signy) < 0)
            {
               //we're colliding with the halfface
               obj.reportCollisionVsWorld(x*oH, 0, oH, 0, t);
               
               return COL_AXIS;        
            }
            else
            {
               //colliding with the vertex or slope
   
               double sx = t.sx;
               double sy = t.sy;
                           
               //if the component of (ox,oy) parallel to the normal's righthand normal
               //has the same sign as the slope of the slope (the sign of the slope's slope is signx*signy)
               //then we project by the slope, otherwise by the vertex.
               //note that this is simply a VERY tricky/weird method of determining 
               //if the circle is in side the slope/face's voronio region, or that of the vertex.                                  
               double perp = (ox*-sy) + (oy*sx);
               
               if ((perp*signx*signy) < 0)
               {
                  //collide vs. vertex
                  double len = Math.sqrt(ox*ox + oy*oy);
                  double pen = obj.r - len;
                  
                  if (0 < pen)
                  {
                     //note: if len=0, then perp=0 and we'll never reach here, so don't worry about div-by-0
                     ox /= len;
                     oy /= len;
      
                     obj.reportCollisionVsWorld(ox*pen, oy*pen, ox, oy, t);
                     
                     return COL_OTHER;
                  }              
               }
               else
               {
                  //collide vs. slope
                     
                  //if the component of (ox,oy) parallel to the normal is less than the circle radius, we're
                  //penetrating the slope. note that this method of penetration calculation doesn't hold
                  //in general (i.e it won't work if the circle is in the slope), but works in this case
                  //because we know the circle is in a neighboring cell
                  double dp = (ox*sx) + (oy*sy);
                  double pen = obj.r - Math.abs(dp);//note: we don't need the abs because we know the dp will be positive, but just in case..
                  
                  if (0 < pen)
                  {
                     //collision; circle out along normal by penetration amount
                     obj.reportCollisionVsWorld(sx*pen, sy*pen, t.sx, t.sy, t);
                     
                     return COL_OTHER;
                  }
               }  
            }
         }
      }
      else
      {
         //colliding diagonally
         if ( 0 < ((signx*oH) + (signy*oV)) ) 
         {
            //the dotprod of slope normal and cell offset is strictly positive,
            //therefore obj is in the diagonal neighb pointed at by the normal.
            
            //collide vs slope
   
            //we should really precalc this at compile time, but for now, fuck it
            double slen = Math.sqrt(2*2 + 1*1);//the raw slope is (-2,-1)
            double sx = (signx*1) / slen;//get slope _unit_ normal;
            double sy = (signy*2) / slen;//raw RH normal is (1,-2)
      
            double r = obj.r;
            double ox = (obj.pos.x - (sx*r)) - (t.pos.x - (signx*t.xw));//this gives is the coordinates of the innermost
            double oy = (obj.pos.y - (sy*r)) - (t.pos.y + (signy*t.yw));//point on the circle, relative to a point on the slope
         
            //if the dotprod of (ox,oy) and (sx,sy) is negative, the point on the circle is in the slope
            //and we need toproject it out by the magnitude of the projection of (ox,oy) onto (sx,sy)
            double dp = (ox*sx) + (oy*sy);
                  
            if (dp < 0)
            {
               //collision; project delta onto slope and use this to displace the object  
               //(sx,sy)*-dp is the projection vector
               obj.reportCollisionVsWorld(-sx*dp, -sy*dp, t.sx, t.sy, t);
               
               return COL_OTHER;
            }
            return COL_NONE;
         }
         else
         {
            //collide vs the appropriate vertex
            double vx = t.pos.x + (oH*t.xw);
            double vy = t.pos.y + (oV*t.yw);
            
            double dx = obj.pos.x - vx;//calc vert->circle vector    
            double dy = obj.pos.y - vy;
            
            double len = Math.sqrt(dx*dx + dy*dy);
            double pen = obj.r - len;
            
            if (0 < pen)
            {
               //vertex is in the circle; project outward
               if (len == 0)
               {
                  //project out by 45deg
                  dx = oH / SQRT2;
                  dy = oV / SQRT2;
               }
               else
               {
                  dx /= len;
                  dy /= len;
               }
   
               obj.reportCollisionVsWorld(dx*pen, dy*pen, dx, dy, t);
   
               return COL_OTHER;
            }
               
         }     
      }
      
      return COL_NONE;
   }
   
   public int projCircle_22DegS(double x, double y, double oH, double oV, Circle obj, TileMapCell t)
   {
      
      //if the object is in a cell pointed at by signy, no collision will ever occur
      //otherwise,
      //
      //if we're colliding diagonally:
      //  -collide vs. the appropriate vertex
      //if obj is in this tile: collide vs slope or vertex
      //if obj is horiz neighb in direction of slope: collide vs. slope or vertex
      //if obj is horiz neighb against the slope:
      //   if (distance in y from circle to 90deg corner of tile < 1/2 tileheight, collide vs. face)
      //   else(collide vs. corner of slope) (vert collision with a non-grid-aligned vert)
      //if obj is vert neighb against direction of slope: collide vs. face
   
      double signx = t.signx;
      double signy = t.signy;
   
      if (0 < (signy*oV))
      {
         //object will never collide vs tile, it can't reach that far
         
         return COL_NONE;
      }
      else if (oH == 0)
      {
         if (oV == 0)
         {
            //colliding with current tile
            //we could only be colliding vs the slope OR a vertex
            //look at the vector form the closest vert to the circle to decide
      
            double sx = t.sx;
            double sy = t.sy;
            
            double r = obj.r;
            double ox = obj.pos.x - (t.pos.x - (signx*t.xw));//this gives is the coordinates of the innermost
            double oy = obj.pos.y - t.pos.y;//point on the circle, relative to the tile corner  
         
            //if the component of (ox,oy) parallel to the normal's righthand normal
            //has the same sign as the slope of the slope (the sign of the slope's slope is signx*signy)
            //then we project by the vertex, otherwise by the normal or axially.
            //note that this is simply a VERY tricky/weird method of determining 
            //if the circle is in side the slope/face's voronio region, or that of the vertex.
               
            double perp = (ox*-sy) + (oy*sx);
            
            if (0 < (perp*signx*signy))
            {
               //collide vs. vertex
               double len = Math.sqrt(ox*ox + oy*oy);
               double pen = r - len;
               
               if (0 < pen)
               {
                  //note: if len=0, then perp=0 and we'll never reach here, so don't worry about div-by-0
                  ox /= len;
                  oy /= len;
   
                  obj.reportCollisionVsWorld(ox*pen, oy*pen, ox, oy, t);
                  
                  return COL_OTHER;
               }              
            }
            else
            {
               //collide vs. slope or vs axis
               ox -= r*sx;//this gives us the vector from  
               oy -= r*sy;//a point on the slope to the innermost point on the circle
         
               //if the dotprod of (ox,oy) and (sx,sy) is negative, the point on the circle is in the slope
               //and we need toproject it out by the magnitude of the projection of (ox,oy) onto (sx,sy)
               double dp = (ox*sx) + (oy*sy);
               
               if (dp < 0)
               {
                  //collision; project delta onto slope and use this to displace the object
                  sx *= -dp;//(sx,sy) is now the projection vector
                  sy *= -dp;     
                     
                  double lenN = Math.sqrt(sx*sx + sy*sy);
                  double lenP;
            
                  //find the smallest axial projection vector
                  if (x < y)
                  {              
                     //penetration in x is smaller
                     lenP = x;
                     y = 0;   
                     //get sign for projection along x-axis    
                     if ((obj.pos.x - t.pos.x) < 0)
                     {
                        x *= -1;
                     }
                  }
                  else
                  {     
                     //penetration in y is smaller
                     lenP = y;
                     x = 0;   
                     //get sign for projection along y-axis    
                     if ((obj.pos.y - t.pos.y)< 0)
                     {
                        y *= -1;
                     }        
                  }
   
                  if (lenP < lenN)
                  {
                     obj.reportCollisionVsWorld(x,y,x/lenP, y/lenP, t);
   
                     return COL_AXIS;
                  }
                  else
                  {           
                     obj.reportCollisionVsWorld(sx,sy,t.sx,t.sy,t);
   
                     return COL_OTHER;
                  }
            
               }
            }
            
         }
         else
         {
            //colliding vertically; we can assume that (signy*oV) < 0
            //due to the first conditional far above
   
            obj.reportCollisionVsWorld(0,y*oV, 0, oV, t);
               
            return COL_AXIS;
         }     
      }
      else if (oV == 0)
      {
         //colliding horizontally
         if ((signx*oH) < 0)
         {
            //colliding with face/edge OR with corner of wedge, depending on our position vertically
               
            //collide vs. vertex
            //get diag vertex position
            double vx = t.pos.x - (signx*t.xw);
            double vy = t.pos.y;
                  
            double dx = obj.pos.x - vx;//calc vert->circle vector    
            double dy = obj.pos.y - vy;
                  
            if ((dy*signy) < 0)
            {
               //colliding vs face
               obj.reportCollisionVsWorld(x*oH, 0, oH, 0, t);
               
               return COL_AXIS;              
            }
            else
            {
               //colliding vs. vertex
                  
               double len = Math.sqrt(dx*dx + dy*dy);
               double pen = obj.r - len;
               
               if (0 < pen)
               {
                  //vertex is in the circle; project outward
                  if (len == 0)
                  {
                     //project out by 45deg
                     dx = oH / SQRT2;
                     dy = oV / SQRT2;
                  }
                  else
                  {
                     dx /= len;
                     dy /= len;
                  }
   
                  obj.reportCollisionVsWorld(dx*pen, dy*pen, dx, dy, t);
                  
                  return COL_OTHER;
               }
            }
         }
         else
         {
            //we could only be colliding vs the slope OR a vertex
            //look at the vector form the closest vert to the circle to decide
      
            double sx = t.sx;
            double sy = t.sy;
               
            double ox = obj.pos.x - (t.pos.x + (oH*t.xw));//this gives is the coordinates of the innermost
            double oy = obj.pos.y - (t.pos.y - (signy*t.yw));//point on the circle, relative to the closest tile vert   
      
            //if the component of (ox,oy) parallel to the normal's righthand normal
            //has the same sign as the slope of the slope (the sign of the slope's slope is signx*signy)
            //then we project by the normal, otherwise by the vertex.
            //(NOTE: this is the opposite logic of the vertical case;
            // for vertical, if the perp prod and the slope's slope agree, it's outside.
            // for horizontal, if the perp prod and the slope's slope agree, circle is inside.
            //  ..but this is only a property of flahs' coord system (i.e the rules might swap
            // in righthanded systems))
            //note that this is simply a VERY tricky/weird method of determining 
            //if the circle is in side the slope/face's voronio region, or that of the vertex.                                  
            double perp = (ox*-sy) + (oy*sx);
            
            if ((perp*signx*signy) < 0)
            {
               //collide vs. vertex
               double len = Math.sqrt(ox*ox + oy*oy);
               double pen = obj.r - len;
               
               if (0 < pen)
               {
                  //note: if len=0, then perp=0 and we'll never reach here, so don't worry about div-by-0
                  ox /= len;
                  oy /= len;
   
                  obj.reportCollisionVsWorld(ox*pen, oy*pen, ox, oy, t);
                  
                  return COL_OTHER;
               }              
            }
            else
            {
               //collide vs. slope
                     
               //if the component of (ox,oy) parallel to the normal is less than the circle radius, we're
               //penetrating the slope. note that this method of penetration calculation doesn't hold
               //in general (i.e it won't work if the circle is in the slope), but works in this case
               //because we know the circle is in a neighboring cell
               double dp = (ox*sx) + (oy*sy);
               double pen = obj.r - Math.abs(dp);//note: we don't need the abs because we know the dp will be positive, but just in case..            
   
               if (0 < pen)
               {
                  //collision; circle out along normal by penetration amount
                  obj.reportCollisionVsWorld(sx*pen, sy*pen, sx, sy, t);
                  
                  return COL_OTHER;
               }
            }
         }
      }
      else
      {
   
         //colliding diagonally; due to the first conditional above,
         //obj is vertically offset against slope, and offset in either direction horizontally
   
         //collide vs. vertex
         //get diag vertex position
         double vx = t.pos.x + (oH*t.xw);
         double vy = t.pos.y + (oV*t.yw);
            
         double dx = obj.pos.x - vx;//calc vert->circle vector    
         double dy = obj.pos.y - vy;
            
         double len = Math.sqrt(dx*dx + dy*dy);
         double pen = obj.r - len;
         
         if (0 < pen)
         {
            //vertex is in the circle; project outward
            if (len == 0)
            {
               //project out by 45deg
               dx = oH / SQRT2;
               dy = oV / SQRT2;
            }
            else
            {
               dx /= len;
               dy /= len;
            }
   
            obj.reportCollisionVsWorld(dx*pen, dy*pen, dx, dy, t);
            
            return COL_OTHER;
         }
      }
   
      return COL_NONE;
   
   }
   
   //projCircle_22DegSO
   
   public int projCircle_Convex(double x, double y, double oH, double oV, Circle obj, TileMapCell t)
   {
      //if the object is horiz AND/OR vertical neighbor in the normal (signx,signy)
      //direction, collide vs. tile-circle only.
      //if we're colliding diagonally:
      //  -else, collide vs. the appropriate vertex
      //if obj is in this tile: perform collision as for aabb
      //if obj is horiz or vert neigh against direction of slope: collide vs. face
   
      int signx = t.signx;
      int signy = t.signy;
   
      if (oH == 0)
      {
         if (oV == 0)
         {
            //colliding with current tile
               
               
            double ox = obj.pos.x - (t.pos.x - (signx*t.xw));//(ox,oy) is the vector from the tile-circle to 
            double oy = obj.pos.y - (t.pos.y - (signy*t.yw));//the circle's center
         
            double twid = t.xw*2;
            double trad = Math.sqrt(twid*twid + 0);//this gives us the radius of a circle centered on the tile's corner and extending to the opposite edge of the tile;
                                       //note that this should be precomputed at compile-time since it's constant
               
            double len = Math.sqrt(ox*ox + oy*oy);
            double pen = (trad + obj.r) - len;
            double lenP;
   
            if (0 < pen)
            {
               //find the smallest axial projection vector
               if (x < y)
               {              
                  //penetration in x is smaller
                  lenP = x;
                  y = 0;
                  
                  //get sign for projection along x-axis    
                  if ((obj.pos.x - t.pos.x) < 0)
                  {
                     x *= -1;
                  }
               }
               else
               {     
                  //penetration in y is smaller
                  lenP = y;
                  x = 0;
                  
                  //get sign for projection along y-axis    
                  if ((obj.pos.y - t.pos.y)< 0)
                  {
                     y *= -1;
                  }        
               }

               
               if (lenP < pen)
               {
                  obj.reportCollisionVsWorld(x, y, x/lenP, y/lenP, t);
                  
                  return COL_AXIS;
               }
               else
               {
                  //note: len should NEVER be == 0, because if it is, 
                  //projeciton by an axis shoudl always be shorter, and we should
                  //never arrive here
                  ox /= len;
                  oy /= len;
                  
                  obj.reportCollisionVsWorld(ox*pen, oy*pen, ox, oy, t);
                  
                  return COL_OTHER;
                  
               }
            }
         }
         else
         {
            //colliding vertically
            if ((signy*oV) < 0)
            {
               //colliding with face/edge
               obj.reportCollisionVsWorld(0, y*oV, 0, oV, t);
               
               return COL_AXIS;
            }
            else
            {
               //obj in neighboring cell pointed at by tile normal;
               //we could only be colliding vs the tile-circle surface
   
               double ox = obj.pos.x - (t.pos.x - (signx*t.xw));//(ox,oy) is the vector from the tile-circle to 
               double oy = obj.pos.y - (t.pos.y - (signy*t.yw));//the circle's center
         
               double twid = t.xw*2;
               double trad = Math.sqrt(twid*twid + 0);//this gives us the radius of a circle centered on the tile's corner and extending to the opposite edge of the tile;
                                       //note that this should be precomputed at compile-time since it's constant
               
               double len = Math.sqrt(ox*ox + oy*oy);
               double pen = (trad + obj.r) - len;
   
               if (0 < pen)
               {
   
                  //note: len should NEVER be == 0, because if it is, 
                  //obj is not in a neighboring cell!
                  ox /= len;
                  oy /= len;
                     
                  obj.reportCollisionVsWorld(ox*pen, oy*pen, ox, oy, t);
                  
                  return COL_OTHER;
               }
            }
         }     
      }
      else if (oV == 0)
      {
         //colliding horizontally
         if ((signx*oH) < 0)
         {
            //colliding with face/edge
            obj.reportCollisionVsWorld(x*oH, 0, oH, 0, t);
            
            return COL_AXIS;
         }
         else
         {
               //obj in neighboring cell pointed at by tile normal;
               //we could only be colliding vs the tile-circle surface
   
            double ox = obj.pos.x - (t.pos.x - (signx*t.xw));//(ox,oy) is the vector from the tile-circle to 
            double oy = obj.pos.y - (t.pos.y - (signy*t.yw));//the circle's center
         
            double twid = t.xw*2;
            double trad = Math.sqrt(twid*twid + 0);//this gives us the radius of a circle centered on the tile's corner and extending to the opposite edge of the tile;
                                       //note that this should be precomputed at compile-time since it's constant
               
            double len = Math.sqrt(ox*ox + oy*oy);
            double pen = (trad + obj.r) - len;
         
            if (0 < pen)
            {

               //note: len should NEVER be == 0, because if it is, 
               //obj is not in a neighboring cell!
               ox /= len;
               oy /= len;

               obj.reportCollisionVsWorld(ox*pen, oy*pen, ox, oy, t);
               
               return COL_OTHER;
            }  
         }
      }
      else
      {
         //colliding diagonally
         if ( 0 < ((signx*oH) + (signy*oV)) ) 
         {
               //obj in diag neighb cell pointed at by tile normal;
               //we could only be colliding vs the tile-circle surface
   
            double ox = obj.pos.x - (t.pos.x - (signx*t.xw));//(ox,oy) is the vector from the tile-circle to 
            double oy = obj.pos.y - (t.pos.y - (signy*t.yw));//the circle's center
         
            double twid = t.xw*2;
            double trad = Math.sqrt(twid*twid + 0);//this gives us the radius of a circle centered on the tile's corner and extending to the opposite edge of the tile;
                                       //note that this should be precomputed at compile-time since it's constant
               
            double len = Math.sqrt(ox*ox + oy*oy);
            double pen = (trad + obj.r) - len;
               
            if (0 < pen)
            {

               //note: len should NEVER be == 0, because if it is, 
               //obj is not in a neighboring cell!
               ox /= len;
               oy /= len;

               obj.reportCollisionVsWorld(ox*pen, oy*pen, ox, oy, t);
               
               return COL_OTHER;
            }
         }
         else
         {
            //collide vs. vertex
            //get diag vertex position
            double vx = t.pos.x + (oH*t.xw);
            double vy = t.pos.y + (oV*t.yw);
            
            double dx = obj.pos.x - vx;//calc vert->circle vector    
            double dy = obj.pos.y - vy;
            
            double len = Math.sqrt(dx*dx + dy*dy);
            double pen = obj.r - len;
            
            if (0 < pen)
            {
               //vertex is in the circle; project outward
               if (len == 0)
               {
                  //project out by 45deg
                  dx = oH / SQRT2;
                  dy = oV / SQRT2;
               }
               else
               {
                  dx /= len;
                  dy /= len;
               }
   
               obj.reportCollisionVsWorld(dx*pen, dy*pen, dx, dy, t);
               
               return COL_OTHER;
            }
      
         }
      
      }
   
      return COL_NONE;
      
   }
   
   public int projCircle_Concave(double x, double y, double oH, double oV, Circle obj, TileMapCell t)
   {
   
      //if we're colliding diagonally:
      // -if obj is in the diagonal pointed to by the slope normal: we can't collide, do nothing
      //  -else, collide vs. the appropriate vertex
      //if obj is in this tile: perform collision as for aabb
      //if obj is horiz OR very neighb in direction of slope: collide vs vert
      //if obj is horiz or vert neigh against direction of slope: collide vs. face
   
      int signx = t.signx;
      int signy = t.signy;
   
      if (oH == 0)
      {
         if (oV == 0)
         {
            //colliding with current tile
            
            double ox = (t.pos.x + (signx*t.xw)) - obj.pos.x;//(ox,oy) is the vector from the circle to 
            double oy = (t.pos.y + (signy*t.yw)) - obj.pos.y;//tile-circle's center
         
            double twid = t.xw*2;
            double trad = Math.sqrt(twid*twid + 0);//this gives us the radius of a circle centered on the tile's corner and extending to the opposite edge of the tile;
                                       //note that this should be precomputed at compile-time since it's constant
               
            double len = Math.sqrt(ox*ox + oy*oy);
            double pen = (len + obj.r) - trad;
            double lenP;
   
            if (0 < pen)
            {
               //find the smallest axial projection vector
               if (x < y)
               {              
                  //penetration in x is smaller
                  lenP = x;
                  y = 0;
                  
                  //get sign for projection along x-axis    
                  if ((obj.pos.x - t.pos.x) < 0)
                  {
                     x *= -1;
                  }
               }
               else
               {     
                  //penetration in y is smaller
                  lenP = y;
                  x = 0;
                  
                  //get sign for projection along y-axis    
                  if ((obj.pos.y - t.pos.y)< 0)
                  {
                     y *= -1;
                  }        
               }

               
               if (lenP < pen)
               {
                  obj.reportCollisionVsWorld(x,y,x/lenP, y/lenP, t);
                  
                  return COL_AXIS;
               }
               else
               {
                  //we can assume that len >0, because if we're here then
                  //(len + obj.r) > trad, and since obj.r <= trad
                  //len MUST be > 0
                  ox /= len;
                  oy /= len;

                  obj.reportCollisionVsWorld(ox*pen, oy*pen, ox, oy, t);
                  
                  return COL_OTHER;
               }
            }
            else
            {
               return COL_NONE;
            }
         }
         else
         {
            //colliding vertically
            if ((signy*oV) < 0)
            {        
               //colliding with face/edge
               obj.reportCollisionVsWorld(0,y*oV, 0, oV, t);
               
               return COL_AXIS;
            }
            else
            {
               //we could only be colliding vs the vertical tip
   
               //get diag vertex position
               double vx = t.pos.x - (signx*t.xw);
               double vy = t.pos.y + (oV*t.yw);
               
               double dx = obj.pos.x - vx;//calc vert->circle vector    
               double dy = obj.pos.y - vy;
               
               double len = Math.sqrt(dx*dx + dy*dy);
               double pen = obj.r - len;
               
               if (0 < pen)
               {
                  //vertex is in the circle; project outward
                  if (len == 0)
                  {
                     //project out vertically
                     dx = 0;
                     dy = oV;
                  }
                  else
                  {
                     dx /= len;
                     dy /= len;
                  }
   
                  obj.reportCollisionVsWorld(dx*pen, dy*pen, dx, dy, t);
                  
                  return COL_OTHER;
               }
            }
         }     
      }
      else if (oV == 0)
      {
         //colliding horizontally
         if ((signx*oH) < 0)
         {
            //colliding with face/edge
            obj.reportCollisionVsWorld(x*oH, 0, oH, 0, t);
            
            return COL_AXIS;
         }
         else
         {
            //we could only be colliding vs the horizontal tip

            //get diag vertex position
            double vx = t.pos.x + (oH*t.xw);
            double vy = t.pos.y - (signy*t.yw);
               
            double dx = obj.pos.x - vx;//calc vert->circle vector    
            double dy = obj.pos.y - vy;
               
            double len = Math.sqrt(dx*dx + dy*dy);
            double pen = obj.r - len;
            
            if (0 < pen)
            {
               //vertex is in the circle; project outward
               if (len == 0)
               {
                  //project out horizontally
                  dx = oH;
                  dy = 0;
               }
               else
               {
                  dx /= len;
                  dy /= len;
               }

               obj.reportCollisionVsWorld(dx*pen, dy*pen, dx, dy, t);
               
               return COL_OTHER;
            }  
         }
      }
      else
      {
         //colliding diagonally
         if ( 0 < ((signx*oH) + (signy*oV)) ) 
         {
            //the dotprod of slope normal and cell offset is strictly positive,
            //therefore obj is in the diagonal neighb pointed at by the normal, and
            //it cannot possibly reach/touch/penetrate the slope
            return COL_NONE;
         }
         else
         {
            //collide vs. vertex
            //get diag vertex position
            double vx = t.pos.x + (oH*t.xw);
            double vy = t.pos.y + (oV*t.yw);
            
            double dx = obj.pos.x - vx;//calc vert->circle vector    
            double dy = obj.pos.y - vy;
            
            double len = Math.sqrt(dx*dx + dy*dy);
            double pen = obj.r - len;
            
            if (0 < pen)
            {
               //vertex is in the circle; project outward
               if (len == 0)
               {
                  //project out by 45deg
                  dx = oH / SQRT2;
                  dy = oV / SQRT2;
               }
               else
               {
                  dx /= len;
                  dy /= len;
               }
   
               obj.reportCollisionVsWorld(dx*pen, dy*pen, dx, dy, t);
               
               return COL_OTHER;
            }
      
         }
      
      }
   
      return COL_NONE;
      
   }

   public int projCircle_45Deg(double x, double y, double oH, double oV, Circle obj, TileMapCell t)
   {
   
      //if we're colliding diagonally:
      // -if obj is in the diagonal pointed to by the slope normal: we can't collide, do nothing
      //  -else, collide vs. the appropriate vertex
      //if obj is in this tile: perform collision as for aabb-ve-45deg
      //if obj is horiz OR very neighb in direction of slope: collide only vs. slope
      //if obj is horiz or vert neigh against direction of slope: collide vs. face
      
      int signx = t.signx;
      int signy = t.signy; 
      
      if (oH == 0)
      {
         if (oV == 0)
         {
            //colliding with current tile
   
            double sx = t.sx;
            double sy = t.sy;
   
            double ox = (obj.pos.x - (sx*obj.r)) - t.pos.x;//this gives is the coordinates of the innermost
            double oy = (obj.pos.y - (sy*obj.r)) - t.pos.y;//point on the circle, relative to the tile center  
   
            //if the dotprod of (ox,oy) and (sx,sy) is negative, the innermost point is in the slope
            //and we need toproject it out by the magnitude of the projection of (ox,oy) onto (sx,sy)
            double dp = (ox*sx) + (oy*sy);
            
            if (dp < 0)
            {
               //collision; project delta onto slope and use this as the slope penetration vector
               sx *= -dp;//(sx,sy) is now the penetration vector
               sy *= -dp;
               
               double lenN = Math.sqrt(sx*sx + sy*sy);
               double lenP;
               
               //find the smallest axial projection vector
               if (x < y)
               {              
                  //penetration in x is smaller
                  lenP = x;
                  y = 0;
                  
                  //get sign for projection along x-axis    
                  if ((obj.pos.x - t.pos.x) < 0)
                  {
                     x *= -1;
                  }
               }
               else
               {     
                  //penetration in y is smaller
                  lenP = y;
                  x = 0;
                  
                  //get sign for projection along y-axis    
                  if ((obj.pos.y - t.pos.y)< 0)
                  {
                     y *= -1;
                  }        
               }
   
               if (lenP < lenN)
               {
                  obj.reportCollisionVsWorld(x,y,x/lenP, y/lenP, t);
                  
                  return COL_AXIS;
               }
               else
               {
                  obj.reportCollisionVsWorld(sx,sy,t.sx,t.sy,t);
                  
                  return COL_OTHER;
               }
            }        
   
         }
         else
         {
            //colliding vertically
            if ((signy*oV) < 0)
            {
               //colliding with face/edge
               obj.reportCollisionVsWorld(0,y*oV,0,oV,t);
               
               return COL_AXIS;
            }
            else
            {
               //we could only be colliding vs the slope OR a vertex
               //look at the vector form the closest vert to the circle to decide
   
               double sx = t.sx;
               double sy = t.sy;
   
               double ox = obj.pos.x - (t.pos.x - (signx*t.xw));//this gives is the coordinates of the innermost
               double oy = obj.pos.y - (t.pos.y + (oV*t.yw));//point on the circle, relative to the closest tile vert   
   
               //if the component of (ox,oy) parallel to the normal's righthand normal
               //has the same sign as the slope of the slope (the sign of the slope's slope is signx*signy)
               //then we project by the vertex, otherwise by the normal.
               //note that this is simply a VERY tricky/weird method of determining 
               //if the circle is in side the slope/face's voronoi region, or that of the vertex.                                  
               double perp = (ox*-sy) + (oy*sx);
               
               if (0 < (perp*signx*signy))
               {
                  //collide vs. vertex
                  double len = Math.sqrt(ox*ox + oy*oy);
                  double pen = obj.r - len;
                  
                  if (0 < pen)
                  {
                     //note: if len=0, then perp=0 and we'll never reach here, so don't worry about div-by-0
                     ox /= len;
                     oy /= len;
   
                     obj.reportCollisionVsWorld(ox*pen, oy*pen, ox, oy, t);
                     
                     return COL_OTHER;
                  }              
               }
               else
               {
                  //collide vs. slope
                  
                  //if the component of (ox,oy) parallel to the normal is less than the circle radius, we're
                  //penetrating the slope. note that this method of penetration calculation doesn't hold
                  //in general (i.e it won't work if the circle is in the slope), but works in this case
                  //because we know the circle is in a neighboring cell
                  double dp = (ox*sx) + (oy*sy);
                  double pen = obj.r - Math.abs(dp);//note: we don't need the abs because we know the dp will be positive, but just in case..
                  
                  if (0 < pen)
                  {
                     //collision; circle out along normal by penetration amount
                     obj.reportCollisionVsWorld(sx*pen, sy*pen, sx, sy, t);
                     
                     return COL_OTHER;
                  }
               }
            }
         }     
      }
      else if (oV == 0)
      {
         //colliding horizontally
         if ((signx*oH) < 0)
         {
            //colliding with face/edge
            obj.reportCollisionVsWorld(x*oH, 0, oH, 0, t);
            
            return COL_AXIS;
         }
         else
         {
            //we could only be colliding vs the slope OR a vertex
            //look at the vector form the closest vert to the circle to decide
   
            double sx = t.sx;
            double sy = t.sy;
   
            double ox = obj.pos.x - (t.pos.x + (oH*t.xw));//this gives is the coordinates of the innermost
            double oy = obj.pos.y - (t.pos.y - (signy*t.yw));//point on the circle, relative to the closest tile vert   
   
            //if the component of (ox,oy) parallel to the normal's righthand normal
            //has the same sign as the slope of the slope (the sign of the slope's slope is signx*signy)
            //then we project by the normal, otherwise by the vertex.
            //(NOTE: this is the opposite logic of the vertical case;
            // for vertical, if the perp prod and the slope's slope agree, it's outside.
            // for horizontal, if the perp prod and the slope's slope agree, circle is inside.
            //  ..but this is only a property of flahs' coord system (i.e the rules might swap
            // in righthanded systems))
            //note that this is simply a VERY tricky/weird method of determining 
            //if the circle is in side the slope/face's voronio region, or that of the vertex.                                  
            double perp = (ox*-sy) + (oy*sx);
            
            if ((perp*signx*signy) < 0)
            {
               //collide vs. vertex
               double len = Math.sqrt(ox*ox + oy*oy);
               double pen = obj.r - len;
               
               if (0 < pen)
               {
                  //note: if len=0, then perp=0 and we'll never reach here, so don't worry about div-by-0
                  ox /= len;
                  oy /= len;

                  obj.reportCollisionVsWorld(ox*pen, oy*pen, ox, oy, t);
                  
                  return COL_OTHER;
               }              
            }
            else
            {
               //collide vs. slope
               
               //if the component of (ox,oy) parallel to the normal is less than the circle radius, we're
               //penetrating the slope. note that this method of penetration calculation doesn't hold
               //in general (i.e it won't work if the circle is in the slope), but works in this case
               //because we know the circle is in a neighboring cell
               double dp = (ox*sx) + (oy*sy);
               double pen = obj.r - Math.abs(dp);//note: we don't need the abs because we know the dp will be positive, but just in case..
               
               if (0 < pen)
               {
                  //collision; circle out along normal by penetration amount
                  obj.reportCollisionVsWorld(sx*pen, sy*pen, sx, sy, t);
                  
                  return COL_OTHER;
               }
            }        
         }
      }
      else
      {
         //colliding diagonally
         if ( 0 < ((signx*oH) + (signy*oV)) ) 
         {
            //the dotprod of slope normal and cell offset is strictly positive,
            //therefore obj is in the diagonal neighb pointed at by the normal, and
            //it cannot possibly reach/touch/penetrate the slope
            return COL_NONE;
         }
         else
         {
            //collide vs. vertex
            //get diag vertex position
            double vx = t.pos.x + (oH*t.xw);
            double vy = t.pos.y + (oV*t.yw);
            
            double dx = obj.pos.x - vx;//calc vert->circle vector    
            double dy = obj.pos.y - vy;
            
            double len = Math.sqrt(dx*dx + dy*dy);
            double pen = obj.r - len;
            
            if (0 < pen)
            {
               //vertex is in the circle; project outward
               if (len == 0)
               {
                  //project out by 45deg
                  dx = oH / SQRT2;
                  dy = oV / SQRT2;
               }
               else
               {
                  dx /= len;
                  dy /= len;
               }
   
               obj.reportCollisionVsWorld(dx*pen, dy*pen, dx, dy, t);
               return COL_OTHER;
            }
      
         }
      
      }
   
      return COL_NONE;
   }

   public int projCircle_Half(double x, double y, double oH, double oV, Circle obj, TileMapCell t)
   {
      //if obj is in a neighbor pointed at by the halfedge normal,
      //we'll never collide (i.e if the normal is (0,1) and the obj is in the DL.D, or R neighbors)
      //
      //if obj is in a neigbor perpendicular to the halfedge normal, it might
      //collide with the halfedge-vertex, or with the halfedge side.
      //
      //if obj is in a neigb pointing opposite the halfedge normal, obj collides with edge
      //
      //if obj is in a diagonal (pointing away from the normal), obj collides vs vertex
      //
      //if obj is in the halfedge cell, it collides as with aabb
   
      int signx = t.signx;
      int signy = t.signy;
   
      double celldp = (oH*signx + oV*signy);//this tells us about the configuration of cell-offset relative to tile normal
      
      if (0 < celldp)
      {
         //obj is in "far" (pointed-at-by-normal) neighbor of halffull tile, and will never hit
         return COL_NONE;
      }
      else if (oH == 0)
      {
         if (oV == 0)
         {
            //colliding with current tile
            double r = obj.r;
            double ox = (obj.pos.x - (signx*r)) - t.pos.x;//this gives is the coordinates of the innermost
            double oy = (obj.pos.y - (signy*r)) - t.pos.y;//point on the circle, relative to the tile center
            
            //we perform operations analogous to the 45deg tile, except we're using 
            //an axis-aligned slope instead of an angled one..
            int sx = signx;
            int sy = signy;
            
            //if the dotprod of (ox,oy) and (sx,sy) is negative, the corner is in the slope
            //and we need toproject it out by the magnitude of the projection of (ox,oy) onto (sx,sy)
            double dp = (ox*sx) + (oy*sy);
            
            if (dp < 0)
            {
               //collision; project delta onto slope and use this to displace the object
               sx *= -dp;//(sx,sy) is now the projection vector
               sy *= -dp;     
                              
               double lenN = Math.sqrt(sx*sx + sy*sy);
               double lenP = Math.sqrt(x*x + y*y);
      
               if (lenP < lenN)
               {
                  obj.reportCollisionVsWorld(x,y,x/lenP, y/lenP,t);
   
                  return COL_AXIS;
               }
               else
               {     
                  obj.reportCollisionVsWorld(sx,sy,t.signx,t.signy, null);
   
                  return COL_OTHER;
               }
            }                    
         }
         else
         {
            //colliding vertically
   
            if (celldp == 0)
            {
               //double r = obj.r;
               double dx = obj.pos.x - t.pos.x;
                     
               //we're in a cell perpendicular to the normal, and can collide vs. halfedge vertex
               //or halfedge side
               if ((dx*signx) < 0)
               {
                  //collision with halfedge side
                  obj.reportCollisionVsWorld(0,y*oV,0,oV,t);
                  
                  return COL_AXIS;                 
               }
               else
               {
                  //collision with halfedge vertex
                  double dy = obj.pos.y - (t.pos.y + oV*t.yw);//(dx,dy) is now the vector from the appropriate halfedge vertex to the circle
                  
                  double len = Math.sqrt(dx*dx + dy*dy);
                  double pen = obj.r - len;
                  
                  if (0 < pen)
                  {
                     //vertex is in the circle; project outward
                     if (len == 0)
                     {
                        //project out by 45deg
                        dx = signx / SQRT2;
                        dy = oV / SQRT2;
                     }
                     else
                     {
                        dx /= len;
                        dy /= len;
                     }
                        
                     obj.reportCollisionVsWorld(dx*pen, dy*pen, dx, dy, t);
                     
                     return COL_OTHER;
                  }              
                  
               }
            }
            else
            {
               //due to the first conditional (celldp >0), we know we're in the cell "opposite" the normal, and so
               //we can only collide with the cell edge
               //collision with vertical neighbor
               obj.reportCollisionVsWorld(0,y*oV,0,oV,t);
               
               return COL_AXIS;
            }
            
         }
      }
      else if (oV == 0)
      {
         //colliding horizontally
         if (celldp == 0)
         {      
            //double r = obj.r;
            double dy = obj.pos.y - t.pos.y;
                     
            //we're in a cell perpendicular to the normal, and can collide vs. halfedge vertex
            //or halfedge side
            if ((dy*signy) < 0)
            {
               //collision with halfedge side
               obj.reportCollisionVsWorld(x*oH,0,oH,0,t);
               
               return COL_AXIS;                 
            }
            else
            {
               //collision with halfedge vertex
               double dx = obj.pos.x - (t.pos.x + oH*t.xw);//(dx,dy) is now the vector from the appropriate halfedge vertex to the circle
                  
               double len = Math.sqrt(dx*dx + dy*dy);
               double pen = obj.r - len;
               
               if (0 < pen)
               {
                  //vertex is in the circle; project outward
                  if (len == 0)
                  {
                     //project out by 45deg
                     dx = signx / SQRT2;
                     dy = oV / SQRT2;
                  }
                  else
                  {
                     dx /= len;
                     dy /= len;
                  }
                        
                  obj.reportCollisionVsWorld(dx*pen, dy*pen, dx, dy, t);
                  
                  return COL_OTHER;
               }              
                  
            }
         }
         else
         {        
            //due to the first conditional (celldp >0), we know w're in the cell "opposite" the normal, and so
            //we can only collide with the cell edge
            obj.reportCollisionVsWorld(x*oH, 0, oH, 0, t);
            
            return COL_AXIS;
         }     
      }
      else
      {     
         //colliding diagonally; we know, due to the initial (celldp >0) test which has failed
         //if we've reached this point, that we're in a diagonal neighbor on the non-normal side, so
         //we could only be colliding with the cell vertex, if at all.
   
         //get diag vertex position
         double vx = t.pos.x + (oH*t.xw);
         double vy = t.pos.y + (oV*t.yw);
            
         double dx = obj.pos.x - vx;//calc vert->circle vector    
         double dy = obj.pos.y - vy;
            
         double len = Math.sqrt(dx*dx + dy*dy);
         double pen = obj.r - len;
         
         if (0 < pen)
         {
            //vertex is in the circle; project outward
            if (len == 0)
            {
               //project out by 45deg
               dx = oH / SQRT2;
               dy = oV / SQRT2;
            }
            else
            {
               dx /= len;
               dy /= len;
            }
   
            obj.reportCollisionVsWorld(dx*pen, dy*pen, dx, dy, t);
            
            return COL_OTHER;
         }     
         
      }
      
      return COL_NONE;
      
   }
   
   public int projCircle_Full(double x, double y, double oH, double oV, Circle obj, TileMapCell t)
   {
      //if we're colliding vs. the current cell, we need to project along the
      //smallest penetration vector.
      //if we're colliding vs. horiz. or vert. neighb, we simply project horiz/vert
      //if we're colliding diagonally, we need to collide vs. tile corner
         
         if (oH == 0)
         {
            if (oV == 0)
            {
   
               //collision with current cell
               if (x < y)
               {              
                  //penetration in x is smaller; project in x
                  double dx = obj.pos.x - t.pos.x;//get sign for projection along x-axis
                  
                  //NOTE: should we handle the delta == 0 case?! and how? (project towards oldpos?)
                  if (dx < 0)
                  {
                     obj.reportCollisionVsWorld(-x,0,-1,0,t);
                     return COL_AXIS;
                  }
                  else
                  {
                     obj.reportCollisionVsWorld(x,0,1,0,t);
                     return COL_AXIS;
                  }
               }
               else
               {     
                  //penetration in y is smaller; project in y     
                  double dy = obj.pos.y - t.pos.y;//get sign for projection along y-axis
   
                  //NOTE: should we handle the delta == 0 case?! and how? (project towards oldpos?)               
                  if (dy < 0)
                  {
                     obj.reportCollisionVsWorld(0,-y,0,-1,t);
                     return COL_AXIS;
                  }
                  else
                  {
                     obj.reportCollisionVsWorld(0,y,0,1,t);
                     return COL_AXIS;
                  }           
               }              
            }
            else
            {
               //collision with vertical neighbor
               obj.reportCollisionVsWorld(0,y*oV,0,oV,t);
   
               return COL_AXIS;
            }
         }
         else if (oV == 0)
         {
            //collision with horizontal neighbor
            obj.reportCollisionVsWorld(x*oH,0,oH,0,t);
            return COL_AXIS;
         }
         else
         {        
            //diagonal collision
            
            //get diag vertex position
            double vx = t.pos.x + (oH*t.xw);
            double vy = t.pos.y + (oV*t.yw);
            
            double dx = obj.pos.x - vx;//calc vert->circle vector    
            double dy = obj.pos.y - vy;
            
            double len = Math.sqrt(dx*dx + dy*dy);
            double pen = obj.r - len;
            
            if (0 < pen)
            {
               //vertex is in the circle; project outward
               if (len == 0)
               {
                  //project out by 45deg
                  dx = oH / SQRT2;
                  dy = oV / SQRT2;
               }
               else
               {
                  dx /= len;
                  dy /= len;
               }
               
               obj.reportCollisionVsWorld(dx*pen, dy*pen, dx, dy, t);
               
               return COL_OTHER;
            }
         }
         
         return COL_NONE;
   }
   
   //this method detects a collision between a circle and an edge-based tilemap,
   //and (if collision is found) calls ResolveCircleTile() to resolve the collision
   //
   //the main difference bewteen circle and aabb collision is that circles need to 
   //collide against vertices as well as edges.
   //
   //since vertices in the tilemap are implicit, we need to do a bit more work than with AABBs.
   //the logic is based upon the cell-index offset form the circle to the tile.
   //
   //if the circle is colliding with the current tile, we use the same logic as for the AABB.
   //otherwise, we have to consider extra cases..
   
   public void collideVsTileMap(TileMap tiles)
   {
      double rad = this.r;
      TileMapCell c = tiles.getTile_V(pos);
      double tx = c.pos.x;
      double ty = c.pos.y;
      double txw = c.xw;
      double tyw = c.yw;
         
      double dx = (pos.x - tx);//tile->obj delta
      double dy = (pos.y - ty);
      
      int hitH = COL_NONE;
      int hitV = COL_NONE;
      int oH=0, oV=0;
      
      if (0 < c.ID)
      {
         //current tile is full!! 
         //for now, move object to oldpos; later, we'll need to determine projection direction and resolve
         //the collision
      
         double px = (txw + rad) - Math.abs(dx);//penetration depth in x   
         double py = (tyw + rad) - Math.abs(dy);//pen depth in y
   
         resolveCircleTile(px,py,0,0,this,c);
      }
   
      //test in y    
      boolean crossV = false;//these flags will indicate whether or not we should collide vs. the diagonal tile
         
      dy = (pos.y - ty);//tile->obj delta
      double py = (Math.abs(dy) + rad) - tyw;//pen depth in y
            
      if (0 < py)
      {
         crossV = true;
         
         // store edge, neighbor, and cell offset
         int eV;
         TileMapCell nV;
         
         if (dy < 0)
         {
            eV = c.eU;
            nV = c.nU;
            oV = 1;
         }
         else
         {
            eV = c.eD;
            nV = c.nD;
            oV = -1;
         }
            
         if (0 < eV)
         {
            //edge is solid or interesting
            if (eV == TileMapCell.EID_SOLID)
            {
               //we're colliding with a solid edge; resolve right away
               hitV = COL_AXIS;
               reportCollisionVsWorld(0,py*oV, 0, oV, nV);
            }
            else
            {
               //edge is interesting; resolve using tile-specific function
               hitV = resolveCircleTile(0,py,0,oV,this,nV);
            }
         }
      }
   
      // test in x
      boolean crossH = false;
      
      dx = (pos.x - tx);//tile->obj delta   
      double px = (Math.abs(dx) + rad) - txw;//penetration depth in x   

      if (0 < px)
      {
         crossH = true;//aabb crosses horizontal edge
         
         int eH;
         TileMapCell nH;
         
         if (dx < 0)
         {
            eH = c.eL;
            nH = c.nL;
            oH = 1;
         }
         else
         {
            eH = c.eR;
            nH = c.nR;
            oH = -1;       
         }
      
         if (0 < eH)
         {
            //edge is solid or interesting
            if (eH == TileMapCell.EID_SOLID)
            {
               //we're colliding with a solid edge; resolve right away
               hitH = COL_AXIS;
               reportCollisionVsWorld(px*oH, 0, oH, 0, nH);
            }
            else
            {
               //edge is interesting; resolve using tile-specific function
               hitH = resolveCircleTile(px,0,oH,0,this,nH);
            }
         }
      }     
      
      //we should collide vs. the diagonal ONLY if the following are true:
      //-object overlaps cell edge vertically but wasn't moved vertically to resolve this
      //-object overlaps cell edge horizontally but wasn't moved horizontally to resolve this
      //-adx and ady should both NOT be 0; if they ARE 0, the object is in the center of the tile
      //and thus (since objects must be smaller than tiles) should never be able to collide
      //with the diagonal

      //NOTE: the diagonal testing is VERY inefficient!
      
      if ((crossH)&&(hitH != COL_AXIS)&&(crossV)&&(hitV != COL_AXIS))
      {
         //we didn't collide with horiz or vert neighbors; test the diagonal neighbor
         //NOTE: just as we used the current cell's edges and the h/v neighbor's cell states
         //to test horizontally/vertically, we use the h/v neighbor's edges and the diagonal
         //neighb's cell state to test diagonally
         
         //note that we can assume that the object is hitting the diagonal edges. we KNOW this
         //due to the testD flag.

         TileMapCell dTile = null;//this should hold a pointer to the diagonal neighbot, IF we're colliding
         //boolean hit = false;//flag to indocate if we should resolve collision or not
         
         //the basic logic for a circle is: if EITHER edge is solid, we're colliding with the 
         //corner vertex and we need to project out. otherwise, if at least one edge is
         //non-off, we need to investigate.
         
         int eH=0, eV=0;
         
         if ((dx < 0) && (dy < 0))
         {
            //test top-left neighbor

            eH = c.nU.eL;
            eV = c.nL.eU;
            dTile = c.nU.nL;
         }
         else if ((dx < 0) && (0 < dy))
         {
            //test bottom-left neighbor

            eH = c.nD.eL;
            eV = c.nL.eD;
            dTile = c.nD.nL;
         }        
         else if ((0 < dx) && (0 < dy))
         {
            //test bottom-right neighbor
            
            eH = c.nD.eR;
            eV = c.nR.eD;
            dTile = c.nD.nR;

         }        
         else if ((0 < dx) && (dy < 0))
         {
            //test top-right neighbor
            
            eH = c.nU.eR;
            eV = c.nR.eU;
            dTile = c.nU.nR;
            
         }
         else
         {
            //object is in exact center of current cell.. we should never have reached here because
            //they could never have touched any cell edges
            System.out.println("WARNING!! we're colliding vs. diagonal but the object is in the dead center of it's cell!!");
         }

         
         if (0 < (eH + eV))
         {
            //at least one edge is non-off
            if ((eH == TileMapCell.EID_SOLID) || (eV == TileMapCell.EID_SOLID))
            {
               //at least one of the edges is solid; project out of the corresponding corner vertex
               double vx = dTile.pos.x + (oH*dTile.xw);
               double vy = dTile.pos.y + (oV*dTile.yw);
               
               dx = pos.x - vx;//calc vert->circle vector      
               dy = pos.y - vy;
               
               double len = Math.sqrt(dx*dx + dy*dy);
               double pen = r - len;
               
               if (0 < pen)
               {
                  //vertex is in the circle; project outward
                  if (len == 0)
                  {
                     //project out by 45deg
                     dx = oH / SQRT2;
                     dy = oV / SQRT2;
                  }
                  else
                  {
                     dx /= len;
                     dy /= len;
                  }

                  reportCollisionVsWorld(dx*pen, dy*pen, dx, dy, dTile);
               }
            }
            else
            {
               //at least one of the edges is interesting; investigate
               //note that we need to update the penetration info since 
               //we may have projected the object horiz/vert
               
               dx = (pos.x - dTile.pos.x);//tile->obj delta
               dy = (pos.y - dTile.pos.y);               
               px = (Math.abs(dx) + rad) - dTile.xw;//penetration depth in x  
               py = (Math.abs(dy) + rad) - dTile.yw;//penetration depth in y
               
               resolveCircleTile(px,py,oH,oV,this,dTile);
            }
         }
         else
         {
            //both edges are off; do nothing
         }
      }
   }   
}
