//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

package com.aleax.mufu.world2;

abstract class DynamicObject
{
   //these are used to report which type of collision was resolved   
   /** no collision was found/resolved */
   public static final int COL_NONE = 0;   
   /** collision was resolved along the x or y axis.. */
   public static final int COL_AXIS = 1;   
   /** tile-specific axis was used to repolve collision (i.e slope normal, etc.) */
   public static final int COL_OTHER = 2;
   
   //min/max values for various environemnt params
   public static final double MIN_FRICTION = 0;
   public static final double MAX_FRICTION = 1;   
   public static final double MIN_BOUNCE = 0;
   public static final double MAX_BOUNCE = 0.99;   
   public static final double MIN_GRAVITY = 0;
   public static final double MAX_GRAVITY = 1;
   
   /** Current position */
   final Vector2 pos;
   
   /** Last position (instead of storing velocity) */
   final Vector2 oldPos;
   
   /** Gravity - .3 is a bit much, .1 is a bit "on the moon". */
   public double gravity = 0.2;
   
   /** Drag - 0 means full drag, 1 is no drag */
   public double drag = 0.999999;
   
   /** Bounciness - must be in [0,1], where 1 means full bounce. but 1 seems to incite "the flubber effect" so use 0.9 as a practical upper bound */
   public double bounce = 0.3;
   
   /** Friction - must be in [0,1] */
   public double friction = 0.05;   

   //basically, these flags are uysed to indicate if an object has been moved.
   //COL_NONE means that it hasn't been moved.
   //COL_AXIS means that it has been moved so that it is no longer colliding with
   //one of the cell edges it was previously colliding with
   //COL_OTHER means it has been moved, but we don't know how (i.e it might still
   //be colliding with cell edges)                                         
   //.collision flags   
   
   /** Creates a DynamicObject with the given initial location */
   public DynamicObject(Vector2 pos)
   {
      this.pos = (Vector2)pos.clone();
      this.oldPos = (Vector2)pos.clone();
   }   
   
   public abstract void verlet();
   public abstract void collideVsWorldBounds();
   public abstract void collideVsTile(TileMapCell tile);
   public abstract void collideVsTileMap(TileMap tiles);
   public abstract void draw(VectorRenderer rend);
   
   /**
    *  Tests for a ray intersection with this DynamicObject
    *  
    *  @param out - where to store the intersection with the ray (if there is one)
    *  @param px - x coordinate of the ray origin
    *  @param py - x coordinate of the ray origin
    *  @param dx - x coordinate of point on ray (for direction)
    *  @param dy - x coordinate of point on ray (for direction)
    *  
    *  @return true if the ray intersects this DynamicObject, false otherwise
    */      
   public abstract boolean testRay(Vector2 out, double px, double py, double dx, double dy);
      
   protected final void integrateVerlet()
   {
      Vector2 p = this.pos;
      Vector2 o = this.oldPos;
         
      // we can't swap buffers since mcs/sticks point directly to vector2s..
      double ox = o.x; 
      double oy = o.y;
      double px = p.x;
      double py = p.y;
      
      o.x = p.x;      //get vector values
      o.y = p.y;      //p = position  
                      //o = oldposition
            
      //integrate 
      p.x += (drag*px) - (drag*ox);
      p.y += (drag*py) - (drag*oy) + gravity;   
   }
   
   /**
    * Call-back for when this object collides with something else. Subclasses overriding this method should call it.
    * 
    * @param px - projection vector x component
    * @param py - projection vector y component
    * @param dx - surface normal x component
    * @param dy - surface normal y component
    * @param obj - object hit
    */

   protected void reportCollisionVsWorld(double px, double py, double dx, double dy, Object obj)
   {
      System.out.println("collision reported to obj!");
      
      Vector2 p = this.pos;
      Vector2 o = this.oldPos;
      
      //calc velocity
      double vx = p.x - o.x;
      double vy = p.y - o.y;
      
      //find component of velocity parallel to collision normal
      double dp = (vx*dx + vy*dy);
      double nx = dp*dx;//project velocity onto collision normal
      
      double ny = dp*dy;//nx,ny is normal velocity
      
      double tx = vx-nx;//px,py is tangent velocity
      double ty = vy-ny;
      
      //we only want to apply collision response forces if the object is travelling into, and not out of, the collision
      double b,bx,by,fx,fy;
      
      if (dp < 0)
      {
         fx = tx*friction;
         fy = ty*friction;     
         
         b = 1+bounce;//this bounce constant should be elsewhere, i.e inside the object/tile/etc..
         
         bx = (nx*b);
         by = (ny*b);      
      }
      else
      {
         //moving out of collision, do not apply forces
         bx = by = fx = fy = 0;   
      }
      
      p.x += px;//project object out of collision
      p.y += py;
      
      o.x += px + bx + fx;//apply bounce+friction impulses which alter velocity
      o.y += py + by + fy;   
   }   
}
