//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

package com.aleax.mufu.world2;

/**
 * this object manages a grid of static AABB tiles.
 * 
 * NOTE: the tilemap is now transparently padded; it is 2 rows and columns larger than
 * any other module knows; the extra rows/cols are a solid border.
 * 
 * however, all client calls can remain the same since the tilemap handles the changes..
 */   

public final class TileMap
{
   private final double xw, yw;
   private final double tw, th;
   final int rows, cols;
   private final int fullRows, fullCols;
   private double minX, maxX, minY, maxY;
   private final TileMapCell[][] grid;
   private int buildStepsRemaining;
   final VectorRenderer rend;
   
   //rows/cols are the integer # of cells in each dimentsion; xw, yw are the halfwidths of each cell
   
   TileMap(int rows, int cols, double xw, double yw)
   {  
      this.xw = xw; //store tile halfwidths
      this.yw = yw;
      
      this.tw = 2*this.xw; //store tile dimensions
      this.th = 2*this.yw;
      this.rows = rows;
      this.cols = cols;
      this.fullRows = this.rows+2;
      this.fullCols = this.cols+2;
      
      this.minX = this.tw;//world bounds only include the inner part of the grid;
      this.minY = this.th;//not the outside edges.
      this.maxX = this.tw + (rows*this.tw);
      this.maxY = this.th + (cols*this.th);
               
      this.grid = new TileMapCell[fullCols][fullRows];
      this.buildStepsRemaining = 9;//state variable used to implement incremental build
      
      //debug renderer
      this.rend = new VectorRenderer();
      draw();
   }
   
   public void setMap(String map)
   {
      for (int row=0 ; row<rows ; row++)
      {
         for (int col=0 ; col<cols ; col++)
            setTileState(row, col, map.charAt(row*cols + col));
      }
   }
   
   //--- incremental construction
   
   //this returns true when the tilemap is fully built
   boolean building()
   {
      //DEBUG
      System.out.println("building tilemap.. steps remaining: " + this.buildStepsRemaining);
      
      //these are here just to make it easier to move the code from where it was..
      double x = this.xw;
      double y = this.yw;
   
      if (this.buildStepsRemaining == 9)
      {
         //build raw tiles    
         for (int i=0 ; i < fullRows; i++)
         {
            for (int j=0; j < fullCols; j++)
            {
               this.grid[i][j] = new TileMapCell(i,j,x,y,this.xw,this.yw);
               y += this.th;        
            }
            
            x += this.tw;
            y = this.yw;
         }  
         
         this.buildStepsRemaining--;
         return true;
      }
      else if (this.buildStepsRemaining == 8)
      {
         //link right
         for (int i=0 ; i < (fullRows-1); i++)
         {
            for (int j=0; j < fullCols; j++)
               this.grid[i][j].linkR( this.grid[i+1][j] );
         }     
         this.buildStepsRemaining--;
         return true;
      }
      else if (this.buildStepsRemaining == 7)
      {
         //link left
         for (int i= 1; i < fullRows; i++)
         {
            for (int j=0; j < fullCols; j++)
               this.grid[i][j].linkL( this.grid[i-1][j] );
         }     
         this.buildStepsRemaining--;
         return true;
      }  
      else if (this.buildStepsRemaining == 6)
      {
         //link down
         for (int i=0 ; i < fullRows; i++)
         {
            for (int j=0; j < (fullCols-1); j++)
               this.grid[i][j].linkD( this.grid[i][j+1] );
         }        
         this.buildStepsRemaining--;
         return true;
      }
      else if (this.buildStepsRemaining == 5)
      {
         //link up
         for (int i=0 ; i < fullRows; i++)
         {
            for (int j=1; j < fullCols; j++)
               this.grid[i][j].linkU( this.grid[i][j-1] );
         }        
         this.buildStepsRemaining--;
         return true;
      }
      else if (this.buildStepsRemaining == 4)
      {
         //fill top border tiles 
         for(int i = 0; i < fullRows; i++)
            this.grid[i][0].setState(TileMapCell.TID_FULL);       
         
         this.buildStepsRemaining--;
         return true;
      }
      else if (this.buildStepsRemaining == 3)
      {     
         //fill bottom border tiles
         for(int i = 0; i < fullRows; i++)
            this.grid[i][fullCols-1].setState(TileMapCell.TID_FULL);
         
         this.buildStepsRemaining--;
         return true;
      }
      else if (this.buildStepsRemaining == 2)
      {     
         //fill left border tiles
         for(int i = 0; i < fullCols; i++)
            this.grid[0][i].setState(TileMapCell.TID_FULL);
         
         this.buildStepsRemaining--;
         return true;
      }
      else if (this.buildStepsRemaining == 1)
      {     
         //fill right border tiles     
         for(int i = 0; i < fullCols; i++)
            this.grid[fullRows-1][i].setState(TileMapCell.TID_FULL);//right row
         
         this.buildStepsRemaining--;
         return true;
      }  
      else 
      {
         //we're done building the tilemap
         return false;
      }
   }
   
   //empties the grid
   //NOTE: only the cell's next/prev will be reset, each object's next/prev must also be reset
   void clearGrid()
   {
      for (int i=0 ; i <grid.length ; i++)
      {
         for (int j=0; j < grid[i].length ; j++)
         {
            grid[i+1][j+1].prev = null;
            grid[i+1][j+1].next = null;
         }
      }  
   }
        
   //-------------------------------- tile access operators -----------------------
   
   //returns a referance to the tile touching point x,y; scalar version
   TileMapCell getTile_S(double x, double y)
   {
      return this.grid[(int)Math.floor(x/this.tw)][(int)Math.floor(y/this.th)];
   }
   
   //vector version
   TileMapCell getTile_V(Vector2 p)
   {
      return this.grid[(int)Math.floor(p.x/this.tw)][(int)Math.floor(p.y/this.th)];
   }
   
   //index-based version
   TileMapCell getTile_I(int i,int j)
   {
      return this.grid[i][j]; //note!! this will break if i or j is out of bounds!!!
   }
   
   //fills vector v with grid coordinates (i.e the cell index) of the tile at point x,y (scalar version)
   void getIndex_S(Vector2 v, double x, double y)
   {
      v.x = Math.floor(x/this.tw);
      v.y = Math.floor(y/this.th);
   }
   
   void getIndex_V(Vector2 v, Vector2 p)
   {
      v.x = Math.floor(p.x/this.tw);
      v.y = Math.floor(p.y/this.th);
   }
   
   //these functions tokenize the tile states; currently they're assuming that
   //the current state of the tilemap (i.e dimensions) are constant, i.e
   //theyr'e not saved/loaded with the tile states.
   //
   //later we should change this.
   //
   //NOTE: the char with code 0 is apparently backspace or something.. so we pad the charcodes to get them in
   //    a "normal" range.. (alphanumerics start at 48, upper case at 65)
   
   // NOTE: the "\" works with input text, but acts like an escape character in code!! so, avoid it (ascii#92)
   private static final int CHAR_PAD = 48;
      
   /** @return a string of ascii characters, where each char is the tokenized description of a tile in the tilemap */
   
   String getTileStates()
   {
      //DEBUg
      System.out.println("getting tile states..");
      
      int n = this.rows;
      int m = this.cols;
      
      String output = "";
      
      for (int i=0 ; i < n; i++)
      {
         for (int j=0; j < m; j++)
            output += (char)(grid[i+1][j+1].ID + CHAR_PAD);         
      }  
      
      return output; 
   }
   
   //sets a single tile state
   void setTileState(int i, int j, int c)
   {
      //DEBUG
      System.out.println("-- tile: " + i + "," + j + " is being set to: " + c);
      
      this.grid[i+1][j+1].setState(c - CHAR_PAD);
      draw();
   }
   
   //each char in the string is assumed to be a tokenized tile-type ID
   void setTileStates(String instr)
   {
      //DEBUG
      System.out.println("setting tile states!!!: " + instr);
      
      for (int i=0 ; i < rows; i++)
      {
         for (int j=0; j < cols; j++)
            grid[i+1][j+1].setState(instr.charAt(i*cols + j) - CHAR_PAD);
      }  
      
      draw();
   }
   
   private void draw()
   {
      rend.clear();

      //draw grid bounds
      this.rend.setStyle(0,0x000000,100);
      this.rend.drawAABB_S(this.minX, this.maxX, this.minY, this.maxY);        
      
      rend.setStyle(0,0x797988,100);
      
      for (int i=0 ; i < fullRows; i++)
      {
         for (int j=0; j < fullCols; j++)
         {
            if (grid[i][j]!=null)
            {
               rend.buffer.beginFill(0x797988,100);
               grid[i][j].draw(rend);
               rend.buffer.endFill();
            }
         }
      }  
   }
   
   public boolean collideRay(Vector2 out, Vector2 p0, Vector2 p1)
   {
      //get starting cell and index
      TileMapCell c = getTile_V(p0);   
      double i0 = c.i;
      double j0 = c.j;
   
      double dx = p1.x - p0.x;//dx,dy is the ray direction vector
      double dy = p1.y - p0.y;
      double len = Math.sqrt(dx*dx + dy*dy);
      
      if (len != 0)
      {
         dx /= len;
         dy /= len;
      }
      else
      {
         System.out.println("Warning!! CollideRayvsTiles was passed a null vector instead of a ray to trace!");
         return false;
      }
      
      //if the length is greater than some (hacky) threshold, use voxel-walking;
      //otherwise, use AABB-based cell traversal
      //NOTE: for now, fuck AABB-based methods; alwasy use voxel-walking!
      
      //NOTE: we should probably collide vs. c here!!!
      
      // X,Y = starting cell coordinates
      double X = i0;
      double Y = j0;
         
      // stepX, stepY = sign of ray direction
      // tMaxX = value of t (ray parameter) at which ray crosses first x boundary
      // tMaxY = value of t (ray parameter) at which ray crosses first y boundary
      //tDeltaX = width of a cell in units of 't'
      // tDeltaY = height of a cell in units of 't'
      
      double stepX, stepY, tMaxX, tMaxY, tDeltaX, tDeltaY;

      //NOTE: we should test for the dx=0 and dy=0 cases, and handle them seperately

      if (dx < 0)
      {
         stepX = -1;
         tMaxX = ((c.pos.x - c.xw) - p0.x) / dx;
         tDeltaX = (2*c.xw) / -dx;
      }
      else if (0 < dx)
      {
         stepX = 1;
         tMaxX = ((c.pos.x + c.xw) - p0.x) / dx;
         tDeltaX = (2*c.xw) / dx;
      }
      else
      {
         //dx is 0, we should only walk in y
         stepX = 0;
         tMaxX = 100000000;//this is hacky, we SHOULD just handle this case seperately, but for now..
         tDeltaX = 0;      //we just want to make sure we never step in x
      }

      if (dy < 0)
      {
         stepY = -1;
         tMaxY = ((c.pos.y - c.yw) - p0.y) / dy;
         tDeltaY = (2*c.yw) / -dy;
      }
      else if (0 < dy)
      {
         stepY = 1;
         tMaxY = ((c.pos.y + c.yw) - p0.y) / dy;
         tDeltaY = (2*c.yw) / dy;               
      }
      else
      {
         //dy is 0, we should only walk in x
         stepY = 0;
         tMaxY = 100000000;//this is hacky, we SHOULD just handle this case seperately, but for now..
         tDeltaY = 0;      //we just want to make sure we never step in y
      }     
      
      double px = p0.x;//px,py is the current position of "us" as we walk along the ray
      double py = p0.y;

      //--perform voxel-traversal
      //DEBUG
      //ray_rend.setStyle(0,0x228888,100);
      
      //test the current cell
      if (c.testRay(out,px,py,dx,dy))
      {
         //we found a collision with the current tile..
         return true;   
      }     
      
      //NOTE: currently we calcuylate the ray-enters-cell position (px,py)
      //from scrach on each iteration; there SHOULD be an incremental way to 
      //do this, but i can't think of it right now (note that you can't just
      //set px,py to the first cell crossing and then increment by tdelta*(dx,dy)..
      //because we DON'T always move in t by the same amount..)
      
      double e;
      TileMapCell n;
      
      while( c != null )
      {  
         if (tMaxX < tMaxY)
         {
            //DEBUG
            //ray_rend.DrawCross_S(p0.x + tMaxX*dx, p0.y + tMaxX*dy);
            
            
            //we just stepped over an edge; test it
            //NOTE: when optimising, we should move the conditionals to outside the loop
            //(i.e have multiple versions of this loop, depending on the quadrant containing ray direction..
            if (stepX < 0)
            {
               //we crossed the left edge; get new edge and neighbor 
               e = c.eL;
               n = c.nL;               
            }
            else
            {
               //we crossed the right edge; get new edge and neighbor 
               e = c.eR;
               n = c.nR;
            }

            if (0 < e)
            {
               //edge is solid or interesting..
               px = p0.x + tMaxX*dx;
               py = p0.y + tMaxX*dy;
               
               if (e == TileMapCell.EID_SOLID)
               {
                  //collision!
                  out.x = px; 
                  out.y = py;
                  
                  return true;
               }
               else
               {                 
                  //the edge is "interesting"..investigate
                  if (n.testRay(out,px,py,dx,dy))
                     return true;
                  else
                  {
                     //no collision; keep walking
                  }
               }
            }
            else
            {
               //edge was empty; do nothing and proceed to the next cell
            }        
              

            //traverse the grid
            tMaxX += tDeltaX;
            X += stepX;
         }
         else
         {
            //DEBUG
            //ray_rend.DrawCross_S(p0.x + tMaxY*dx, p0.y + tMaxY*dy);
         
            if (stepY < 0)
            {
               //we crossed the up edge; get new edge and neighbor 
               e = c.eU;
               n = c.nU;               
            }
            else
            {
               //we crossed the down edge; get new edge and neighbor 
               e = c.eD;
               n = c.nD;
            }

            if (0 < e)
            {
               //edge is solid or interesting..
               px = p0.x + tMaxY*dx;
               py = p0.y + tMaxY*dy;
               
               if (e == TileMapCell.EID_SOLID)
               {
                  //collision!
                  out.x = px; 
                  out.y = py;
                  
                  return true;
               }
               else
               {                 
                  //the edge is "interesting"..investigate
                  if (n.testRay(out,px,py,dx,dy))
                     return true;
                  else
                  {
                     //no collision; keep walking
                  }
               }
            }
            else
            {
               //edge was empty; do nothing and proceed to the next cell
            }        
              
            //traverse the grid
            tMaxY += tDeltaY;
            Y += stepY;
         }     
         
         c = n;//get next tile
      }
   
      return false;           
   }
   
   //this method returns true if the ray from p0 through p1 hits the object obj BEFORE hitting the tiles, false
   //otherwise; it also fills out with the position of the intersection (a point on either a tile or the obj)
   public boolean queryRayObj(Vector2 out, Vector2 p0, Vector2 p1, DynamicObject obj)
   {
      //get starting cell and index
      TileMapCell c = getTile_V(p0);   
      double i0 = c.i;
      double j0 = c.j;
   
      double dx = p1.x - p0.x;//dx,dy is the ray direction vector
      double dy = p1.y - p0.y;
      double len = Math.sqrt(dx*dx + dy*dy);
      
      if (len != 0)
      {
         dx /= len;
         dy /= len;
      }
      else
      {
         System.out.println("Warning!! QueryRayObj was passed a null vector instead of a ray to trace!");
         return false;
      }
      
      // X,Y = starting cell coordinates
      double X = i0;
      double Y = j0;
         
      // stepX, stepY = sign of ray direction
      // tMaxX = value of t (ray parameter) at which ray crosses first x boundary
      // tMaxY = value of t (ray parameter) at which ray crosses first y boundary
      //tDeltaX = width of a cell in units of 't'
      // tDeltaY = height of a cell in units of 't'
      
      double stepX, stepY, tMaxX, tMaxY, tDeltaX, tDeltaY;

      //NOTE: we should test for the dx=0 and dy=0 cases, and handle them seperately

      if (dx < 0)
      {
         stepX = -1;
         tMaxX = ((c.pos.x - c.xw) - p0.x) / dx;
         tDeltaX = (2*c.xw) / -dx;
      }
      else if (0 < dx)
      {
         stepX = 1;
         tMaxX = ((c.pos.x + c.xw) - p0.x) / dx;
         tDeltaX = (2*c.xw) / dx;
      }
      else
      {
         //dx is 0, we should only walk in y
         stepX = 0;
         tMaxX = 100000000;//this is hacky, we SHOULD just handle this case seperately, but for now..
         tDeltaX = 0;      //we just want to make sure we never step in x
      }

      if (dy < 0)
      {
         stepY = -1;
         tMaxY = ((c.pos.y - c.yw) - p0.y) / dy;
         tDeltaY = (2*c.yw) / -dy;
      }
      else if (0 < dy)
      {
         stepY = 1;
         tMaxY = ((c.pos.y + c.yw) - p0.y) / dy;
         tDeltaY = (2*c.yw) / dy;               
      }
      else
      {
         //dy is 0, we should only walk in x
         stepY = 0;
         tMaxY = 100000000;//this is hacky, we SHOULD just handle this case seperately, but for now..
         tDeltaY = 0;      //we just want to make sure we never step in y
      }     
      
      double px = p0.x;//px,py is the current position of "us" as we walk along the ray
      double py = p0.y;

      //--perform voxel-traversal
      //DEBUG
      //ray_rend.SetStyle(0,0x228888,100);      
      
      //test the current cell
      if (c.testRay(out,px,py,dx,dy))
      {
         //we found a collision with the current tile..
         
         //test vs. object
         double tx = out.x;//cache ray-tile intersection point
         double ty = out.y;
         
         if (obj.testRay(out,p0.x,p0.y,dx,dy))
         {
            //ray also collides with object; check which is closer to ray origin
            //find the dotprod of obj-intersection->origin vs. raydirection and 
            //tile-intersection->origin vs. raydirection; the largest value (least negative) is the closest
            double dpO = (p0.x - out.x)*dx + (p0.y - out.y)*dy;
            double dpT = (p0.x - tx)*dx + (p0.y - ty)*dy;
            
            //DEBUG
            System.out.println("dpO: " + dpO + "  dpT: " + dpT);
            
            if (dpO < dpT)
            {
               //ray hits tile first
               out.x = tx;
               out.y = ty;
               return false;
            }
            else
            {
               //ray hits object first
               return true;
            }
         }
         else
         {
            //ray missses object; return ray-tile intersection
            out.x = tx;
            out.y = ty;
            return false;
         }
      }     
      
      //NOTE: currently we calcuylate the ray-enters-cell position (px,py)
      //from scrach on each iteration; there SHOULD be an incremental way to 
      //do this, but i can't think of it right now (note that you can't just
      //set px,py to the first cell crossing and then increment by tdelta*(dx,dy)..
      //because we DON'T always move in t by the same amount..)
      
      double e;
      TileMapCell n;
      
      while( c != null )
      {  
         if (tMaxX < tMaxY)
         {
            //DEBUG
            //ray_rend.DrawCross_S(p0.x + tMaxX*dx, p0.y + tMaxX*dy);
            
            //we just stepped over an edge; test it
            //NOTE: when optimising, we should move the conditionals to outside the loop
            //(i.e have multiple versions of this loop, depending on the quadrant containing ray direction..
            if (stepX < 0)
            {
               //we crossed the left edge; get new edge and neighbor 
               e = c.eL;
               n = c.nL;               
            }
            else
            {
               //we crossed the right edge; get new edge and neighbor 
               e = c.eR;
               n = c.nR;
            }

            if (0 < e)
            {
               //edge is solid or interesting..
               px = p0.x + tMaxX*dx;
               py = p0.y + tMaxX*dy;
               
               if (e == TileMapCell.EID_SOLID)
               {                 
                  //test vs. object
                  double tx = px;//cache ray-tile intersection point
                  double ty = py;
                  
                  if (obj.testRay(out,p0.x,p0.y,dx,dy))
                  {
                     //ray also collides with object; check which is closer to ray origin
                     //find the dotprod of obj-intersection->origin vs. raydirection and 
                     //tile-intersection->origin vs. raydirect????.ion; the largest value (least negative) is the closest
                     double dpO = (p0.x - out.x)*dx + (p0.y - out.y)*dy;
                     double dpT = (p0.x - tx)*dx + (p0.y - ty)*dy;

                     //DEBUG
                     System.out.println("dpO: " + dpO + "  dpT: " + dpT);

                     if (dpO < dpT)
                     {
                        //ray hits tile first
                        out.x = tx;
                        out.y = ty;
                        return false;
                     }
                     else
                     {
                        //ray hits object first
                        return true;
                     }
                  }
                  else
                  {
                     //ray missses object; return ray-tile intersection
                     out.x = tx;
                     out.y = ty;
                     return false;
                  }
               }
               else
               {                 
                  //the edge is "interesting"..investigate
                  if (n.testRay(out,px,py,dx,dy))
                  {
                     //test vs. object
                     double tx = out.x;//cache ray-tile intersection point
                     double ty = out.y;
                     
                     if (obj.testRay(out,p0.x,p0.y,dx,dy))
                     {
                        //ray also collides with object; check which is closer to ray origin
                        //find the dotprod of obj-intersection->origin vs. raydirection and 
                        //tile-intersection->origin vs. raydirection; the largest value (least negative) is the closest
                        double dpO = (p0.x - out.x)*dx + (p0.y - out.y)*dy;
                        double dpT = (p0.x - tx)*dx + (p0.y - ty)*dy;

                        //DEBUG
                        System.out.println("dpO: " + dpO + "  dpT: " + dpT);

                        if (dpO < dpT)
                        {
                           //ray hits tile first
                           out.x = tx;
                           out.y = ty;
                           return false;
                        }
                        else
                        {
                           //ray hits object first
                           return true;
                        }
                     }
                     else
                     {
                        //ray missses object; return ray-tile intersection
                        out.x = tx;
                        out.y = ty;
                        return false;
                     }
                  }
                  else
                  {
                     //no collision; keep walking
                  }
               }
            }
            else
            {
               //edge was empty; do nothing and proceed to the next cell
            }        
              
            //traverse the grid
            tMaxX += tDeltaX;
            X += stepX;
         }
         else
         {
            //DEBUG
            //ray_rend.DrawCross_S(p0.x + tMaxY*dx, p0.y + tMaxY*dy);
         
            if (stepY < 0)
            {
               //we crossed the up edge; get new edge and neighbor 
               e = c.eU;
               n = c.nU;               
            }
            else
            {
               //we crossed the down edge; get new edge and neighbor 
               e = c.eD;
               n = c.nD;
            }

            if (0 < e)
            {
               //edge is solid or interesting..
               px = p0.x + tMaxY*dx;
               py = p0.y + tMaxY*dy;
               
               if (e == TileMapCell.EID_SOLID)
               {
                  //test vs. object
                  double tx = px;//cache ray-tile intersection point
                  double ty = py;
                  
                  if (obj.testRay(out,p0.x,p0.y,dx,dy))
                  {
                     //ray also collides with object; check which is closer to ray origin
                     //find the dotprod of obj-intersection->origin vs. raydirection and 
                     //tile-intersection->origin vs. raydirection; the largest value (least negative) is the closest
                     double dpO = (p0.x - out.x)*dx + (p0.y - out.y)*dy;
                     double dpT = (p0.x - tx)*dx + (p0.y - ty)*dy;

                     //DEBUG
                     System.out.println("dpO: " + dpO + "  dpT: " + dpT);
   
                     if (dpO < dpT)
                     {
                        //ray hits tile first
                        out.x = tx;
                        out.y = ty;
                        return false;
                     }
                     else
                     {
                        //ray hits object first
                        return true;
                     }
                  }
                  else
                  {
                     //ray misses object; return ray-tile intersection
                     out.x = tx;
                     out.y = ty;
                     return false;
                  }
               }
               else
               {                 
                  //the edge is "interesting"..investigate
                  if (n.testRay(out,px,py,dx,dy))
                  {
                     //test vs. object
                     double tx = out.x;//cache ray-tile intersection point
                     double ty = out.y;
                     
                     if (obj.testRay(out,p0.x,p0.y,dx,dy))
                     {
                        //ray also collides with object; check which is closer to ray origin
                        //find the dotprod of obj-intersection->origin vs. raydirection and 
                        //tile-intersection->origin vs. raydirection; the largest value (least negative) is the closest
                        double dpO = (p0.x - out.x)*dx + (p0.y - out.y)*dy;
                        double dpT = (p0.x - tx)*dx + (p0.y - ty)*dy;
   
                        //DEBUG
                        System.out.println("dpO: " + dpO + "  dpT: " + dpT);
   
                        if (dpO < dpT)
                        {
                           //ray hits tile first
                           out.x = tx;
                           out.y = ty;
                           return false;
                        }
                        else
                        {
                           //ray hits object first
                           return true;
                        }
                     }
                     else
                     {
                        //ray missses object; return ray-tile intersection
                        out.x = tx;
                        out.y = ty;
                        return false;
                     }
                  }
                  else
                  {
                     //no collision; keep walking
                  }
               }
            }
            else
            {
               //edge was empty; do nothing and proceed to the next cell
            }        
              
            //traverse the grid
            tMaxY += tDeltaY;
            Y += stepY;
         }     
         
         c = n;//get next tile
      }
   
      //if we reach here, we need to test vs obj.. since we didn't hit any tiles...
      if (obj.testRay(out,p0.x,p0.y,dx,dy))
         return true;      
         
      // note: we should really return a flag, since there are THREE possible results: tile, obj, and NOTHING intersection..
      System.out.println("WARNING!! QueryRayObj() didn't find an intersection between the ray and ANYTHING!!");
      return false;  
   }   
}
