//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

package com.aleax.mufu.world2;

import java.util.*;

/** 
 * this object stores all the info for a tile; note that a lot of this is superfluous
 * (i.e any non-empty cell (i.e ID > 0) doesn't need drag/grav, and empty cells don't
 * really need position/xw/yw)
 */                   

public final class TileMapCell
{
   private static final double SQRT2 = Math.sqrt(2);
   
   //Tile types
   public static final int TID_EMPTY = 0;
   public static final int TID_FULL = 1;//fullAABB tile
   public static final int TID_45DEGpn = 2;//45-degree triangle, whose normal is (+ve,-ve)
   public static final int TID_45DEGnn = 3;//(+ve,+ve)
   public static final int TID_45DEGnp = 4;//(-ve,+ve)
   public static final int TID_45DEGpp = 5;//(-ve,-ve)
   public static final int TID_CONCAVEpn = 6;//1/4-circle cutout
   public static final int TID_CONCAVEnn = 7;
   public static final int TID_CONCAVEnp = 8;
   public static final int TID_CONCAVEpp = 9;
   public static final int TID_CONVEXpn = 10;//1/4/circle
   public static final int TID_CONVEXnn = 11;
   public static final int TID_CONVEXnp = 12;
   public static final int TID_CONVEXpp = 13;
   public static final int TID_22DEGpnS = 14;//22.5 degree slope
   public static final int TID_22DEGnnS = 15;
   public static final int TID_22DEGnpS = 16;
   public static final int TID_22DEGppS = 17;
   public static final int TID_22DEGpnB = 18;
   public static final int TID_22DEGnnB = 19;
   public static final int TID_22DEGnpB = 20;
   public static final int TID_22DEGppB = 21;
   public static final int TID_67DEGpnS = 22;//67.5 degree slope
   public static final int TID_67DEGnnS = 23;
   public static final int TID_67DEGnpS = 24;
   public static final int TID_67DEGppS = 25;
   public static final int TID_67DEGpnB = 26;
   public static final int TID_67DEGnnB = 27;
   public static final int TID_67DEGnpB = 28;
   public static final int TID_67DEGppB = 29;
   public static final int TID_HALFd = 30;//half-full tiles
   public static final int TID_HALFr = 31;
   public static final int TID_HALFu = 32;
   public static final int TID_HALFl = 33;

   //collision shape  "types"
   public static final int CTYPE_EMPTY = 0;
   public static final int CTYPE_FULL = 1;
   public static final int CTYPE_45DEG = 2;
   public static final int CTYPE_CONCAVE = 6;
   public static final int CTYPE_CONVEX = 10;
   public static final int CTYPE_22DEGs = 14;
   public static final int CTYPE_22DEGb = 18;
   public static final int CTYPE_67DEGs = 22;
   public static final int CTYPE_67DEGb = 26;
   public static final int CTYPE_HALF = 30;
   
   //---- edge state enumeration
   public static final int EID_OFF = 0;
   public static final int EID_INTERESTING = 1;
   public static final int EID_SOLID = 2;
   
   int ID;
   int CTYPE;
   final Vector2 pos;
   final double xw; //half-width
   final double yw; //half-height
   final double minx, maxx, miny, maxy;
   final int i,j; //tile coords in map
   
   int signx, signy;
   double sx, sy;
   
   //Neighbours up, right, down, left
   TileMapCell nU, nR, nD, nL;
   
   //edge states
   int eU, eR, eD, eL;

   // this is probably uselesss but should help while debugging..
   int objcounter = 0;
   
   //Linked list of tiles - not sure what this is for
   TileMapCell prev;
   TileMapCell next;
   
   final List objects = new LinkedList();

   public TileMapCell(double x, double y, double xw, double yw)
   {
      this(0,0, x, y, xw, yw);
   }
 
   public TileMapCell(int i, int j, double x, double y, double xw, double yw)
   {
      this.ID = TID_EMPTY; //all tiles start empty
      this.CTYPE = CTYPE_EMPTY;
      this.i = i;//store the index fo this tile in the grid
      this.j = j;
      this.nU = null;//init neighbor info
      this.nD = null;
      this.nL = null;
      this.nR = null;
      
      //edge info; all edges start off
      //NOTE: the format of edges will change as we try different collision detection methods..
      this.eU = EID_OFF;
      this.eD = EID_OFF;
      this.eL = EID_OFF;
      this.eR = EID_OFF;
      
      this.pos = new Vector2(x,y);  //setup collision properties
      this.xw = xw;
      this.yw = yw;
      this.minx = this.pos.x - this.xw;
      this.maxx = this.pos.x + this.xw;
      this.miny = this.pos.y - this.yw;
      this.maxy = this.pos.y + this.yw;
      
      //this stores tile-specific collision information
      this.signx = 0;
      this.signy = 0;
      this.sx = 0;
      this.sy = 0;
   }
   
   //these functions inits a tile by linking it to it's neighbors
   //note: border tiles have null neighbors (by default/as part of the tile-construction)
   //so we should simplyt NOT link them..
   public void linkU(TileMapCell t)
   {
      nU = t;
      
      //DEBUG
      //System.out.println("linked cell (" + this.i + "," + this.j + ") to cell (" + t.i + "," + t.j + ")");
      //System.out.println("this.nU: " + this.nU);
      //System.out.println("this.nU != null? " + (this.nU != null) );
   }
   public void linkD(TileMapCell t)
   {
      nD = t;
   }
   public void linkL(TileMapCell t)
   {
      nL = t;
   }
   public void linkR(TileMapCell t)
   {
      nR = t;
   }
      
   public String toString()
   {
      return "(" + i + "," + j + ")";
   }
   
   public void draw(VectorRenderer rend)
   {
      //TODO: drawing
      switch (ID)
      {
         case TID_EMPTY:
            break;
         case TID_FULL:
            rend.drawAABB(pos, xw, yw);
            break;
         case TID_45DEGpn:
            rend.drawTri_S(minx, miny, minx, maxy, maxx, maxy);
            break;
         case TID_45DEGnn:
            rend.drawTri_S(minx, maxy, maxx, maxy, maxx, miny);
            break;
         case TID_45DEGnp:
            rend.drawTri_S(minx, miny, maxx, miny, maxx, maxy);
            break;
         case TID_45DEGpp:
            rend.drawTri_S(minx, miny, minx, maxy, maxx, miny);
            break;
         case TID_CONCAVEpn:
            rend.buffer.moveTo(minx, maxy);
            rend.drawConcaveCCWArc_S(maxx, miny, minx, miny);
            rend.buffer.lineTo(minx, maxy);
            rend.buffer.lineTo(minx, miny);
            break;
         case TID_CONCAVEnn:
            rend.buffer.moveTo(minx, maxy);
            rend.drawConcaveCCWArc_S(minx, miny, minx, maxy);
            rend.buffer.lineTo(maxx, maxy);
            rend.buffer.lineTo(minx, maxy);
            break;
         case TID_CONCAVEnp:
            rend.buffer.moveTo(minx, miny);
            rend.buffer.lineTo(maxx, miny);
            rend.buffer.lineTo(maxx, maxy);
            rend.drawConcaveCCWArc_S(minx, maxy, maxx, maxy);
            break;
         case TID_CONCAVEpp:
            rend.buffer.moveTo(minx, maxy);
            rend.buffer.lineTo(minx, miny);
            rend.buffer.lineTo(maxx, miny);
            rend.drawConcaveCCWArc_S(maxx, maxy, maxx, miny);
            break;
         case TID_CONVEXpn:
            rend.buffer.moveTo(minx, miny);
            rend.buffer.lineTo(minx, maxy);
            rend.buffer.lineTo(maxx, maxy);
            rend.drawConcaveCCWArc_S(minx, maxy, maxx, maxy);
            break;
         case TID_CONVEXnn:
            rend.buffer.moveTo(maxx, miny);
            rend.drawConcaveCCWArc_S(maxx, maxy, maxx, miny);
            rend.buffer.lineTo(maxx, maxy);
            rend.buffer.lineTo(maxx, miny);
            break;
         case TID_CONVEXnp:
            rend.buffer.moveTo(minx, miny);
            rend.drawConcaveCCWArc_S(maxx, miny, minx, miny);
            rend.buffer.lineTo(maxx, miny);
            rend.buffer.lineTo(minx, miny);
            break;
         case TID_CONVEXpp:
            rend.buffer.moveTo(maxx, miny);
            rend.buffer.lineTo(minx, miny);
            rend.buffer.lineTo(minx, maxy);
            rend.drawConcaveCCWArc_S(minx, miny, minx, maxy);
            break;
         case TID_22DEGpnS:
            rend.drawTri_S(maxx, maxy, minx, maxy-yw, minx, maxy);
            break;
         case TID_22DEGnnS:
            rend.drawTri_S(minx, maxy, maxx, maxy-yw, maxx, maxy);
            break;
         case TID_22DEGnpS:
            rend.drawTri_S(minx, miny, maxx, miny+yw, maxx, miny);
            break;
         case TID_22DEGppS:
            rend.drawTri_S(maxx, miny, minx, miny+yw, minx, miny);
            break;
         case TID_22DEGpnB:
            rend.drawQuad_S(minx, miny, minx, maxy, maxx, maxy, maxx, miny+yw);
            break;
         case TID_22DEGnnB:
            rend.drawQuad_S(maxx, miny, maxx, maxy, minx, maxy, minx, miny+yw);
            break;
         case TID_22DEGnpB:
            rend.drawQuad_S(maxx, maxy, maxx, miny, minx, miny, minx, maxy-yw);
            break;
         case TID_22DEGppB:
            rend.drawQuad_S(minx, maxy, minx, miny, maxx, miny, maxx, maxy-yw);
            break;
         case TID_67DEGpnS:
            rend.drawTri_S(maxx-xw, maxy, minx, maxy, minx, miny);
            break;
         case TID_67DEGnnS:
            rend.drawTri_S(minx+xw, maxy, maxx, maxy, maxx, miny);
            break;
         case TID_67DEGnpS:
            rend.drawTri_S(minx+xw, miny, maxx, miny, maxx, maxy);
            break;
         case TID_67DEGppS:
            rend.drawTri_S(maxx-xw, miny, minx, miny, minx, maxy);
            break;
         case TID_67DEGpnB:
            rend.drawQuad_S(minx, miny, minx, maxy, maxx, maxy, maxx-xw, miny);
            break;
         case TID_67DEGnnB:
            rend.drawQuad_S(maxx, miny, maxx, maxy, minx, maxy, minx+xw, miny);
            break;
         case TID_67DEGnpB:
            rend.drawQuad_S(maxx, maxy, maxx, miny, minx, miny, minx+xw, maxy);
            break;
         case TID_67DEGppB:
            rend.drawQuad_S(minx, maxy, minx, miny, maxx, miny, maxx-xw, maxy);
            break;
         case TID_HALFd:
            rend.drawQuad_S(minx, miny+yw, minx, maxy, maxx, maxy, maxx, miny+yw);
            break;
         case TID_HALFr:
            rend.drawQuad_S(maxx, miny, maxx, maxy, maxx-xw, maxy, maxx-xw, miny);
            break;
         case TID_HALFu:
            rend.drawQuad_S(minx, miny, minx, maxy-yw, maxx, maxy-yw, maxx, miny);
            break;
         case TID_HALFl:
            rend.drawQuad_S(minx, miny, minx, maxy, minx+xw, maxy, minx+xw, miny);
            break;
      }
   }
   
   public void drawOld(VectorRenderer rend)
   {
      //0,0x882222,100 debug drawing funciton..
      int thick = 1;
      
      rend.clear();
      rend.setStyle(2*thick,0x882222,100);
      
      if (0 < ID)
         rend.drawPlus(pos);
      
      if (1 < ID)
      {
         if (ID < TID_CONCAVEpn)
         {
            //45deg slope; draw it
            if ((ID%2)==0)
            {
               //slope x=y
               rend.drawLine_S(minx, miny, maxx, maxy);    
            }
            else
            {
               //slope x = -y
               rend.drawLine_S(minx, maxy, maxx, miny);
            }
         }
         else if (ID < TID_22DEGpnS)
         {
            //circular tile
            
            if (ID == TID_CONCAVEpn || (ID == TID_CONVEXnp))
            {
               //concave 1/4-circle 
               rend.drawConcaveCCWArc_S(maxx, miny, minx, miny);
            }
            else if (ID == TID_CONCAVEnn || (ID == TID_CONVEXpp))
            {
               //concave 1/4-circle
               rend.drawConcaveCCWArc_S(minx, miny, minx, maxy); 
            }
            else if (ID == TID_CONCAVEnp || (ID == TID_CONVEXpn))
            {
               //concave 1/4-circle
               rend.drawConcaveCCWArc_S(minx, maxy, maxx, maxy); 
            }
            else if (ID == TID_CONCAVEpp || (ID == TID_CONVEXnn))
            {
               //concave 1/4-circle
               rend.drawConcaveCCWArc_S(maxx, maxy, maxx, miny); 
            }  
            
         }
         else if (ID < TID_67DEGpnS)
         {
            //22.5deg tile
            if ((ID == TID_22DEGpnS) || (ID == TID_22DEGnpB))
            {
               rend.drawLine_S(minx, pos.y, maxx, maxy);
            }
            else if (ID == TID_22DEGpnB || (ID == TID_22DEGnpS))
            {
               rend.drawLine_S(minx, miny, maxx, pos.y);
            }
            else if (ID == TID_22DEGnnS || (ID == TID_22DEGppB))
            {
               rend.drawLine_S(minx, maxy, maxx, pos.y);
            }
            else if (ID == TID_22DEGnnB || (ID == TID_22DEGppS))
            {
               rend.drawLine_S(minx, pos.y, maxx, miny);
            }        
         }
         else if (ID < TID_HALFd)
         {
            //67deg slope  
            if (ID == TID_67DEGpnS || (ID == TID_67DEGnpB))
            {
               rend.drawLine_S(minx, miny, pos.x, maxy);   
            }
            else if (ID == TID_67DEGpnB || (ID == TID_67DEGnpS))
            {
               rend.drawLine_S(pos.x, miny, maxx, maxy);   
            }
            else if (ID == TID_67DEGnnS || (ID == TID_67DEGppB))
            {
               rend.drawLine_S(maxx, miny, pos.x, maxy);
            }
            else if (ID == TID_67DEGnnB || (ID == TID_67DEGppS))
            {
               rend.drawLine_S(pos.x, miny, minx, maxy);
            }        
         }
         else
         {
            //half-full
            if ((ID == TID_HALFu) || (ID == TID_HALFd))
            {
               rend.drawLine_S(minx, pos.y, maxx, pos.y);
            }
            else if ((ID == TID_HALFl) || (ID == TID_HALFr))
            {
               rend.drawLine_S(pos.x, miny, pos.x, maxy);
            }
         }
      }
      
      //draw edges slightly offset 
      thick = 0;
      
      if (eU == EID_OFF)
      {
         //rend.setStyle(0,0x228822,100);
         //rend.drawLine_S(minx, miny + thick, maxx, miny + thick);
      }
      else if (eU == EID_INTERESTING)
      {
         rend.setStyle(0,0xAAAAAA,100);
         rend.drawLine_S(minx, miny + thick, maxx, miny + thick);
      }
      else if (eU == EID_SOLID)
      {
         rend.setStyle(0,0x882222,100);
         rend.drawLine_S(minx, miny + thick, maxx, miny + thick);
      }
      else
      {
         //error!!
         rend.setStyle(4,0x000000,100);
         rend.drawLine_S(minx, miny + thick, maxx, miny + thick);
      }
      
      if (eD == EID_OFF)
      {
         //rend.setStyle(0,0x228822,100);
         //rend.drawLine_S(minx, maxy - thick, maxx, maxy - thick);
      }
      else if (eD == EID_INTERESTING)
      {
         rend.setStyle(0,0xAAAAAA,100);
         rend.drawLine_S(minx, maxy - thick, maxx, maxy - thick);
      }
      else if (eD == EID_SOLID)
      {
         rend.setStyle(0,0x882222,100);
         rend.drawLine_S(minx, maxy - thick, maxx, maxy - thick);
      }
      else
      {
         //error!!
         rend.setStyle(4,0x000000,100);
         rend.drawLine_S(minx, maxy - thick, maxx, maxy - thick);
      }
   
   
      if (eL == EID_OFF)
      {
         //rend.setStyle(0,0x228822,100);
         //rend.drawLine_S(minx + thick, miny, minx + thick, maxy);
      }
      else if (eL == EID_INTERESTING)
      {
         rend.setStyle(0,0xAAAAAA,100);
         rend.drawLine_S(minx + thick, miny, minx + thick, maxy);
      }
      else if (eL == EID_SOLID)
      {
         rend.setStyle(0,0x882222,100);
         rend.drawLine_S(minx + thick, miny, minx + thick, maxy);
      }
      else
      {
         //error!!
         rend.setStyle(4,0x000000,100);
         rend.drawLine_S(minx + thick, miny, minx + thick, maxy);
      }
   
      if (eR == EID_OFF)
      {
         //rend.setStyle(0,0x228822,100);
         //rend.drawLine_S(maxx - thick, miny, maxx - thick, maxy);
      }
      else if (eR == EID_INTERESTING)
      {
         rend.setStyle(0,0xAAAAAA,100);
         rend.drawLine_S(maxx - thick, miny, maxx - thick, maxy);
      }
      else if (eR == EID_SOLID)
      {
         rend.setStyle(0,0x882222,100);
         rend.drawLine_S(maxx - thick, miny, maxx - thick, maxy);
      }
      else
      {
         //error!!
         rend.setStyle(4,0x000000,100);
         rend.drawLine_S(maxx - thick, miny, maxx - thick, maxy);
      }
   }
   
   //these functions are used to update the cell
   //note: ID is assumed to NOT be "empty" state..
   //if it IS the empty state, the tile clears itself
   public void setState(int ID)
   {
      if (ID == TID_EMPTY)
         clear();
      else
      {
         //set tile state to a non-emtpy value, and update it's edges and those of the neighbors
         this.ID = ID;
         updateType();
         updateEdges();//this also draws...
         updateNeighbors();//broadcasts changes to neighboring cells
      }        
   }
   
   void clear()
   {
      //tile was on, turn it off
      ID = TID_EMPTY;
      updateType();
      updateEdges();//we don't really need to do this, as this tile's edge states are based only on it's neighbors' states, not on iself
      updateNeighbors();
   }
   
   //this converts a tile from implicitly-defined (via ID), to explicit (via properties)
   public boolean updateType()
   {
      if (0 < ID)
      {
         //tile is non-empty; collide
         if (ID < CTYPE_45DEG)
         {
            //TID_FULL
            CTYPE = CTYPE_FULL;
            signx = 0;
            signy = 0;
            sx = 0;
            sy = 0;
         }
         else if (ID < CTYPE_CONCAVE)
         {
            //45deg
            CTYPE = CTYPE_45DEG;
            
            if (ID == TID_45DEGpn)
            {
               signx = 1;
               signy = -1;
               sx = signx / SQRT2;//get slope _unit_ normal
               sy = signy / SQRT2;//since normal is (1,-1), length is sqrt(1*1 + -1*-1) = sqrt(2)                       
            }
            else if (ID == TID_45DEGnn)
            {
               signx = -1;
               signy = -1;
               sx = signx / SQRT2;//get slope _unit_ normal
               sy = signy / SQRT2;//since normal is (1,-1), length is sqrt(1*1 + -1*-1) = sqrt(2)           
               
            }
            else if (ID == TID_45DEGnp)
            {
               signx = -1;
               signy = 1;
               sx = signx / SQRT2;//get slope _unit_ normal
               sy = signy / SQRT2;//since normal is (1,-1), length is sqrt(1*1 + -1*-1) = sqrt(2)           
            }
            else if (ID == TID_45DEGpp)
            {
               signx = 1;
               signy = 1;
               sx = signx / SQRT2;//get slope _unit_ normal
               sy = signy / SQRT2;//since normal is (1,-1), length is sqrt(1*1 + -1*-1) = sqrt(2)           
            }           
            else
            {
               System.out.println("BAAAD TILE!!!!!: ID=" + ID + " ("+ i + "," + j + ")");
               return false;
            }           
         }
         else if (ID < CTYPE_CONVEX)
         {
            //concave
            CTYPE = CTYPE_CONCAVE;
            if (ID == TID_CONCAVEpn)
            {
               signx = 1;
               signy = -1;
               sx = 0;
               sy = 0;
            }
            else if (ID == TID_CONCAVEnn)
            {
               signx = -1;
               signy = -1;
               sx = 0;
               sy = 0;
            }
            else if (ID == TID_CONCAVEnp)
            {
               signx = -1;
               signy = 1;
               sx = 0;
               sy = 0;
            }  
            else if (ID == TID_CONCAVEpp)
            {
               signx = 1;
               signy = 1;
               sx = 0;
               sy = 0;
            }
            else
            {
               System.out.println("BAAAD TILE!!!!!: ID=" + ID + " ("+ i + "," + j + ")");
               return false;
            }
         }
         else if (ID < CTYPE_22DEGs)
         {
            //convex
            CTYPE = CTYPE_CONVEX;
            if (ID == TID_CONVEXpn)
            {
               signx = 1;
               signy = -1;
               sx = 0;
               sy = 0;
            }
            else if (ID == TID_CONVEXnn)
            {
               signx = -1;
               signy = -1;
               sx = 0;
               sy = 0;
            }
            else if (ID == TID_CONVEXnp)
            {
               signx = -1;
               signy = 1;
               sx = 0;
               sy = 0;
            }
            else if (ID == TID_CONVEXpp)
            {
               signx = 1;
               signy = 1;
               sx = 0;
               sy = 0;
            }
            else
            {
               System.out.println("BAAAD TILE!!!!!: ID=" + ID + " ("+ i + "," + j + ")");
               return false;
            }
         }
         else if (ID < CTYPE_22DEGb)
         {
            //22deg small
            CTYPE = CTYPE_22DEGs;
            if (ID == TID_22DEGpnS)
            {
               signx = 1;
               signy = -1;
               double slen = Math.sqrt(2*2 + 1*1);
               sx = (signx*1) / slen;
               sy = (signy*2) / slen;          
            }
            else if (ID == TID_22DEGnnS)
            {
               signx = -1;
               signy = -1;
               double slen = Math.sqrt(2*2 + 1*1);
               sx = (signx*1) / slen;
               sy = (signy*2) / slen;
            }  
            else if (ID == TID_22DEGnpS)
            {
               signx = -1;
               signy = 1;
               double slen = Math.sqrt(2*2 + 1*1);
               sx = (signx*1) / slen;
               sy = (signy*2) / slen;
            }
            else if (ID == TID_22DEGppS)
            {
               signx = 1;
               signy = 1;
               double slen = Math.sqrt(2*2 + 1*1);
               sx = (signx*1) / slen;
               sy = (signy*2) / slen;
            }
            else
            {
               System.out.println("BAAAD TILE!!!!!: ID=" + ID + " ("+ i + "," + j + ")");
               return false;
            }                    
         }
         else if (ID < CTYPE_67DEGs)
         {                                             
            //22deg big
            CTYPE = CTYPE_22DEGb;
            if (ID == TID_22DEGpnB)
            {
               signx = 1;
               signy = -1;
               double slen = Math.sqrt(2*2 + 1*1);
               sx = (signx*1) / slen;
               sy = (signy*2) / slen; 
            }
            else if (ID == TID_22DEGnnB)
            {
               signx = -1;
               signy = -1;
               double slen = Math.sqrt(2*2 + 1*1);
               sx = (signx*1) / slen;
               sy = (signy*2) / slen; 
            }
            else if (ID == TID_22DEGnpB)
            {
               signx = -1;
               signy = 1;
               double slen = Math.sqrt(2*2 + 1*1);
               sx = (signx*1) / slen;
               sy = (signy*2) / slen; 
            }
            else if (ID == TID_22DEGppB)
            {
               signx = 1;
               signy = 1;
               double slen = Math.sqrt(2*2 + 1*1);
               sx = (signx*1) / slen;
               sy = (signy*2) / slen; 
            }
            else
            {
               System.out.println("BAAAD TILE!!!!!: ID=" + ID + " ("+ i + "," + j + ")");
               return false;
            }                       
         }
         else if (ID < CTYPE_67DEGb)
         {
                                                
            //67deg small
            CTYPE = CTYPE_67DEGs;
            if (ID == TID_67DEGpnS)
            {
               signx = 1;
               signy = -1;
               double slen = Math.sqrt(2*2 + 1*1);
               sx = (signx*2) / slen;
               sy = (signy*1) / slen; 
            }
            else if (ID == TID_67DEGnnS)
            {
               signx = -1;
               signy = -1;
               double slen = Math.sqrt(2*2 + 1*1);
               sx = (signx*2) / slen;
               sy = (signy*1) / slen;
            }
            else if (ID == TID_67DEGnpS)
            {
               signx = -1;
               signy = 1;
               double slen = Math.sqrt(2*2 + 1*1);
               sx = (signx*2) / slen;
               sy = (signy*1) / slen;
            }  
            else if (ID == TID_67DEGppS)
            {
               signx = 1;
               signy = 1;
               double slen = Math.sqrt(2*2 + 1*1);
               sx = (signx*2) / slen;
               sy = (signy*1) / slen;
            }
            else
            {
               System.out.println("BAAAD TILE!!!!!: ID=" + ID + " ("+ i + "," + j + ")");
               return false;
            }                          
         }
         else if (ID < CTYPE_HALF)
         {
                           
            //67deg big
            CTYPE = CTYPE_67DEGb;
            if (ID == TID_67DEGpnB)
            {
               signx = 1;
               signy = -1;
               double slen = Math.sqrt(2*2 + 1*1);
               sx = (signx*2) / slen;
               sy = (signy*1) / slen;
            }
            else if (ID == TID_67DEGnnB)
            {
               signx = -1;
               signy = -1;
               double slen = Math.sqrt(2*2 + 1*1);
               sx = (signx*2) / slen;
               sy = (signy*1) / slen;
            }
            else if (ID == TID_67DEGnpB)
            {
               signx = -1;
               signy = 1;
               double slen = Math.sqrt(2*2 + 1*1);
               sx = (signx*2) / slen;
               sy = (signy*1) / slen;
            }  
            else if (ID == TID_67DEGppB)
            {
               signx = 1;
               signy = 1;
               double slen = Math.sqrt(2*2 + 1*1);
               sx = (signx*2) / slen;
               sy = (signy*1) / slen;
            }  
            else
            {
               System.out.println("BAAAD TILE!!!!!: ID=" + ID + " ("+ i + "," + j + ")");
               return false;
            }
         }
         else
         {
            //half-full tile
            CTYPE = CTYPE_HALF;
            if (ID == TID_HALFd)
            {
               signx = 0;
               signy = -1;
               sx = signx;
               sy = signy;
            }
            else if (ID == TID_HALFu)
            {
               signx = 0;
               signy = 1;
               sx = signx;
               sy = signy;
            }
            else if (ID == TID_HALFl)
            {
               signx = 1;
               signy = 0;
               sx = signx;
               sy = signy;
            }
            else if (ID == TID_HALFr)
            {
               signx = -1;
               signy = 0;
               sx = signx;
               sy = signy;
            }
            else
            {
               System.out.println("BAAD TILE!!!: ID=" + ID + " ("+ i + "," + j + ")");
               return false;
            }                             
         }
      }
      else
      {
         //TID_EMPTY
         CTYPE = CTYPE_EMPTY;
         signx = 0;
         signy = 0;
         sx = 0;
         sy = 0;
      }
      
      return true;
   }

   //this method updates neighbor's edge states
   //(i.e if this tile is activated, it's neighbor's edges must be updated to reflect the change..)
   //note that we could simply call UpdateEdges() on all neighbors (so we don't duplicate code/etc..)
   //this is very inefficient, but for now we care more about ease of implementation
   void updateNeighbors()
   {
      if (nU != null)
         nU.updateEdges();

      if (nD != null)
         nD.updateEdges();

      if (nL != null)
         nL.updateEdges();

      if (nR != null)
         nR.updateEdges();  
   }
   
   //updates the tiles edges; if the current tile is empty, and it's neighbor is full,
   //then it's edges should be turned on
   //NOTE: most of this logic belongs in an editor.. i.e the relationship between
   //tile states and edge states is arbitrary at run-time, all this logic is doing is
   //enforcing certain behaviours we want when editing (i.e toggleable solid tiles)
   private void updateEdges()
   {
      //the new rules for determining edge state are quite complicated.   
      TileMapCell n = nU;
      
      if (ID == TID_EMPTY)
      {
         if (n==null || n.ID == TID_EMPTY)
         {  
            //edge is off
            eU = EID_OFF;
         }
         else if (n.ID == TID_FULL)
         {
            eU = EID_SOLID;
         }
         else if (((n.signy*-1) <= 0) || n.ID == TID_67DEGpnS || n.ID == TID_67DEGnnS)
         {
            //nieghbor's surface points towards us; edge is interesting
            eU = EID_INTERESTING;
            //note that the <= is supposed to flag edges sharesd with half-fulls as interesting, but it 
            //might also have unwanted negative sideeffects
         }
         else
         {
            //neighbors surface points away; edge is solid
            eU = EID_SOLID;
         }        
      }
      else if (ID == TID_FULL)
      {
         //edge will either be off or interesting
         if (n==null || n.ID == TID_EMPTY)
         {
            eU = EID_OFF;
         }
         else if (n.ID == TID_FULL)
         {
            eU = EID_OFF;
         }
         else if ((n.signy*-1) <= 0 || n.ID == TID_67DEGpnS || n.ID == TID_67DEGnnS)
         {
            eU = EID_INTERESTING;
         }
         else
         {
            eU = EID_OFF;
         }
      }
      else
      {
         //edges pointed at by this cell's normal can be off, interesting, or solid.   
         //edges opposite this cell's normal are off or interesting
         if (0 <= (signy*-1))
         {
            if (n==null || n.ID == TID_EMPTY)
            {
               eU = EID_OFF;
            }
            else if (n.ID == TID_FULL)
            {
               eU = EID_SOLID;
            }
            else if ((n.signy*-1) <= 0 || n.ID == TID_67DEGpnS || n.ID == TID_67DEGnnS)
            {
               //nieghbor's surface points towards us; edge is interesting
               eU = EID_INTERESTING;
               //note that the <= is supposed to flag edges sharesd with half-fulls as interesting, but it 
               //might also have unwanted negative sideeffects
            }
            else
            {
               //neighbors surface points away; edge is solid
               eU = EID_SOLID;
            }
         }
         else
         {   
            //edges pointing away from the cell normal can be off or interesting,
            //OR solid if the cell is 22/67 small
            
            if (ID == TID_67DEGppS || ID == TID_67DEGnpS)
            {
               if (n==null || n.ID == TID_EMPTY)
               {
                  eU = EID_OFF;
               }
               else if (n.ID == TID_FULL)
               {
                  eU = EID_SOLID;
               }           
               else if ((n.signy*-1) <= 0 || n.ID == TID_67DEGpnS || n.ID == TID_67DEGnnS)
               {
                  eU = EID_INTERESTING;
               }
               else if (0 < (n.signy*-1) || n.ID == TID_FULL)
               {
                  eU = EID_SOLID;
               }
               else
               {
                  eU = EID_OFF;
               }           
            }
            else 
            {
                  
               if (n.ID == TID_FULL)
               {
                  eU = EID_OFF;
               }
               else if (n.ID == TID_EMPTY)
               {
                  eU = EID_OFF;
               }
               else if ((n.signy*-1) <= 0 || n.ID == TID_67DEGpnS || n.ID == TID_67DEGnnS)
               {
                  eU = EID_INTERESTING;
               }
               else
               {
                  eU = EID_OFF;
               }
            }
         }  
      }
   
      n = nD;
      
      if (ID == TID_EMPTY)
      {
         if (n==null || n.ID == TID_EMPTY)
         {
            //edge is off
            eD = EID_OFF;
         }     
         else if (n.ID == TID_FULL)
         {
            eD = EID_SOLID;
         }
         else if ((n.signy*1) <= 0 || n.ID == TID_67DEGppS || n.ID == TID_67DEGnpS)
         {
            //nieghbor's surface points towards us; edge is interesting
            eD = EID_INTERESTING;
            //note that the <= is supposed to flag edges sharesd with half-fulls as interesting, but it 
            //might also have unwanted negative sideeffects
         }
         else
         {
            //neighbors surface points away; edge is solid
            eD = EID_SOLID;
         }        
      }
      else if (ID == TID_FULL)
      {
         //edge will either be off or interesting
         if (n==null || n.ID == TID_EMPTY)
         {
            eD = EID_OFF;
         }     
         else if (n.ID == TID_FULL)
         {
            eD = EID_OFF;
         }
         else if ((n.signy*1) <= 0 || n.ID == TID_67DEGppS || n.ID == TID_67DEGnpS)
         {
            eD = EID_INTERESTING;
         }
         else
         {
            eD = EID_OFF;
         }
      }
      else
      {
         //edges pointed at by this cell's normal can be off, interesting, or solid.   
         //edges opposite this cell's normal are off or interesting
         if (0 <= (signy*1))
         {
            if (n==null || n.ID == TID_EMPTY)
            {
               eD = EID_OFF;
            }
            else if (n.ID == TID_FULL)
            {
               eD = EID_SOLID;
            }
            else if ((n.signy*1) <= 0 || n.ID == TID_67DEGppS || n.ID == TID_67DEGnpS)
            {
               //nieghbor's surface points towards us; edge is interesting
               eD = EID_INTERESTING;
               //note that the <= is supposed to flag edges sharesd with half-fulls as interesting, but it 
               //might also have unwanted negative sideeffects
            }
            else
            {
               //neighbors surface points away; edge is solid
               eD = EID_SOLID;
            }
         }
         else
         {   
            if (ID == TID_67DEGpnS || ID == TID_67DEGnnS)
            {
               if (n==null || n.ID == TID_EMPTY)
               {
                  eD = EID_OFF;
               }
               else if (n.ID == TID_FULL)
               {
                  eD = EID_SOLID;
               }           
               else if ((n.signy*1) <= 0 || n.ID == TID_67DEGppS || n.ID == TID_67DEGnpS)
               {
                  eD = EID_INTERESTING;
               }
               else if (0 < (n.signy*1) || n.ID == TID_FULL)
               {
                  eD = EID_SOLID;
               }
               else
               {
                  eD = EID_OFF;
               }           
            }
            else 
            {
               if (n==null || n.ID == TID_EMPTY)
               {
                  eD = EID_OFF;
               }        
               else if (n.ID == TID_FULL)
               {
                  eD = EID_OFF;
               }
               else if ((n.signy*1) <= 0 || n.ID == TID_67DEGppS || n.ID == TID_67DEGnpS)
               {
                  eD = EID_INTERESTING;
               }
               else
               {
                  eD = EID_OFF;
               }
            }
         }  
      }
   
      n = nR;
      
      if (ID == TID_EMPTY)
      {
         if (n==null || n.ID == TID_EMPTY)
         {
            //edge is off
            eR = EID_OFF;
         }     
         else if (n.ID == TID_FULL)
         {
            eR = EID_SOLID;
         }
         else if ((n.signx*1) <= 0 || n.ID == TID_22DEGpnS || n.ID == TID_22DEGppS)
         {
            //nieghbor's surface points towards us; edge is interesting
            eR = EID_INTERESTING;
            //note that the <= is supposed to flag edges sharesd with half-fulls as interesting, but it 
            //might also have unwanted negative sideeffects
         }
         else
         {
            //neighbors surface points away; edge is solid
            eR = EID_SOLID;
         }        
      }
      else if (ID == TID_FULL)
      {
         //edge will either be off or interesting
         if (n==null || n.ID == TID_EMPTY)
         {
            eR = EID_OFF;
         }     
         else if (n.ID == TID_FULL)
         {
            eR = EID_OFF;
         }
         else if ((n.signx*1) <= 0 || n.ID == TID_22DEGpnS || n.ID == TID_22DEGppS)
         {
            eR = EID_INTERESTING;
         }
         else
         {
            eR = EID_OFF;
         }
      }
      else
      {
         //edges pointed at by this cell's normal can be off, interesting, or solid.   
         //edges opposite this cell's normal are off or interesting
         if (0 <= (signx*1))
         {
            //DEBUG
            //var mc = CreateMC("EMPTY_MC","j");
            //mc.lineStyle(2,0x228822,100);
            ///mc.moveTo(pos.x, pos.y);
            //mc.lineTo(pos.x, pos.y - yw);        
            
            if (n==null || n.ID == TID_EMPTY)
            {
               eR = EID_OFF;
            }
            else if (n.ID == TID_FULL)
            {
               eR = EID_SOLID;
            }
            else if ((n.signx*1) <= 0 || n.ID == TID_22DEGpnS || n.ID == TID_22DEGppS)
            {
               //nieghbor's surface points towards us; edge is interesting
               eR = EID_INTERESTING;
               //note that the <= is supposed to flag edges sharesd with half-fulls as interesting, but it 
               //might also have unwanted negative sideeffects
            }
            else
            {
               //neighbors surface points away; edge is solid
               eR = EID_SOLID;
            }
         }
         else
         {
            if (ID == TID_22DEGnnS || ID == TID_22DEGnpS)
            {
               if (n==null || n.ID == TID_EMPTY)
               {
                  eR = EID_OFF;
                  
                  //DEBUG
                  //var mc = CreateMC("EMPTY_MC","j");
                  //mc.lineStyle(2,0x882222,100);
                  //mc.moveTo(pos.x, pos.y);
                  //mc.lineTo(pos.x - xw, pos.y - yw);             
                  
               }
               else if (n.ID == TID_FULL)
               {
                  eR = EID_SOLID; 
               }           
               else if ((n.signx*1) <= 0 || n.ID == TID_22DEGpnS || n.ID == TID_22DEGppS)
               {
                  eR = EID_INTERESTING;             
               }
               else if (n.ID == TID_FULL || (0 < (n.signx*1)) )
               {
                  eR = EID_SOLID;             
               }
               else
               {
                  eR = EID_OFF;
               }           
            }
            else 
            {
               if (n==null || n.ID == TID_EMPTY)
               {
                  eR = EID_OFF;
               }        
               else if (n.ID == TID_FULL)
               {
                  eR = EID_OFF;
               }
               else if ((n.signx*1) <= 0 || n.ID == TID_22DEGpnS || n.ID == TID_22DEGppS)
               {
                  eR = EID_INTERESTING;
               }
               else
               {
                  eR = EID_OFF;
               }
            }
         }  
      }
   
      n = nL;
      
      if (ID == TID_EMPTY)
      {
         if (n==null || n.ID == TID_EMPTY)
         {
            //edge is off
            eL = EID_OFF;
         }     
         else if (n.ID == TID_FULL)
         {
            eL = EID_SOLID;
         }
         else if ((n.signx*-1) <= 0 || n.ID == TID_22DEGnnS || n.ID == TID_22DEGnpS)
         {
            //nieghbor's surface points towards us; edge is interesting
            eL = EID_INTERESTING;
            //note that the <= is supposed to flag edges sharesd with half-fulls as interesting, but it 
            //might also have unwanted negative sideeffects
         }
         else
         {
            //neighbors surface points away; edge is solid
            eL = EID_SOLID;
         }        
      }
      else if (ID == TID_FULL)
      {
         //edge will either be off or interesting
         if (n==null || n.ID == TID_EMPTY)
         {
            eL = EID_OFF;
         }     
         else if (n.ID == TID_FULL)
         {
            eL = EID_OFF;
         }
         else if ((n.signx*-1) <= 0 || n.ID == TID_22DEGnnS || n.ID == TID_22DEGnpS)
         {
            eL = EID_INTERESTING;
         }
         else
         {
            eL = EID_OFF;
         }
      }
      else
      {
         //edges pointed at by this cell's normal can be off, interesting, or solid.   
         //edges opposite this cell's normal are off or interesting
         if (0 <= (signx*-1))
         {
            if (n==null || n.ID == TID_EMPTY)
            {
               eL = EID_OFF;
            }
            else if (n.ID == TID_FULL)
            {
               eL = EID_SOLID;
            }
            else if ((n.signx*-1) <= 0 || n.ID == TID_22DEGnnS || n.ID == TID_22DEGnpS)
            {
               //nieghbor's surface points towards us; edge is interesting
               eL = EID_INTERESTING;
               //note that the <= is supposed to flag edges sharesd with half-fulls as interesting, but it 
               //might also have unwanted negative sideeffects
            }
            else
            {
               //neighbors surface points away; edge is solid
               eL = EID_SOLID;
            }
         }
         else
         {
            if (ID == TID_22DEGpnS || ID == TID_22DEGppS)
            {
               if (n==null || n.ID == TID_EMPTY)
               {
                  eL = EID_OFF;
               }
               else if (n.ID == TID_FULL)
               {
                  eL = EID_SOLID;
               }
               else if ((n.signx*-1) <= 0 || n.ID == TID_22DEGnnS || n.ID == TID_22DEGnpS)
               {
                  eL = EID_INTERESTING;
               }
               else if (0 < (n.signx*-1) || n.ID == TID_FULL)
               {
                  eL = EID_SOLID;
               }
               else
               {
                  eL = EID_OFF;
               }           
               
            }
            else 
            {                    
               if (n==null || n.ID == TID_EMPTY)
               {
                  eL = EID_OFF;
               }        
               else if (n.ID == TID_FULL)
               {
                  eL = EID_OFF;
               }
               else if ((n.signx*-1) <= 0 || n.ID == TID_22DEGnnS || n.ID == TID_22DEGnpS)
               {
                  eL = EID_INTERESTING;
               }
               else
               {
                  eL = EID_OFF;
               }
            }
         }  
      }
   }
   
   //this method inserts a tile into the cell's linkedlist
   void insertObj(TileMapCell obj)
   {
      //insert the object at the head of the list
      
      obj.next = next;//hook object into head of list
      obj.prev = this;
      
      prev = null;//just in case..
      
      if (next != null)
      {
         //not sure if the conditional is necessary, but just in case..
         next.prev = obj;//set previous list head to point to object
      }
      next = obj;//set obj as list head
      
      objcounter++;//update the numbre of objects in this cell
   }
   
   //note: this illustrates that these methods should NOT belogn to the cell, since
   //this one has NOTHING to do with the cell!!
   void removeObj(TileMapCell obj)
   {
      //remove the object from the list   
      obj.prev.next = obj.next;
      if (obj.next != null)
      {
         //not sure if the conditional is necessary, but just in case..
         obj.next.prev = obj.prev;//the next object in the lsit is the new head
      }  
      
      obj.next = null;
      obj.prev = null;  
      
      objcounter--;//update the number of objects in this cell
      
   }   
   
   
   //this method detects intersection between a ray and a tile;
   //note that we can assume the ray didn't enter along a "full" tile edge
   //if intersection is found, the point of intersection is stored in out
   //px,py is the point where the ray enters the tile; dx,dy is the ray's (unit)
   //direction vector, and t is the tile
   
   public boolean testRay(Vector2 out, double px, double py, double dx, double dy)
   {
      switch (CTYPE)
      {
         case TileMapCell.CTYPE_FULL:
            return testRay_Full(out,px,py,dx,dy);
         case TileMapCell.CTYPE_45DEG:
            return testRay_45Deg(out,px,py,dx,dy);
         case TileMapCell.CTYPE_CONCAVE:
            return testRay_Concave(out,px,py,dx,dy);
         case TileMapCell.CTYPE_CONVEX:
            return testRay_Convex(out,px,py,dx,dy);
         case TileMapCell.CTYPE_22DEGs:
            return testRay_22DegS(out,px,py,dx,dy);
         case TileMapCell.CTYPE_22DEGb:
            return testRay_22DegB(out,px,py,dx,dy);
         case TileMapCell.CTYPE_67DEGs:
            return testRay_67DegS(out,px,py,dx,dy);
         case TileMapCell.CTYPE_67DEGb:
            return testRay_67DegB(out,px,py,dx,dy);
         case TileMapCell.CTYPE_HALF:
            return testRay_Half(out,px,py,dx,dy);
         default:
            System.out.println("TestRayTile() was called with an empty (or unknown) tile!: ID=" + ID + " ("+ i + "," + j + ")");
            return false;
      }
   }

   private boolean testRay_67DegB(Vector2 out, double px, double py, double dx, double dy)
   {
      //DEBUG
      System.out.println("--- intersecting ray vs. 67DegB tile!");      
      
      double sx = this.sx;
      double sy = this.sy;
      int signx = this.signx;
      int signy = this.signy;
      
      //we need to first test if we started on the tile's half-edge
      double wx = pos.x - px;//the test point is the corner joining the slpoe to the halfedge
      double wy = pos.y - py;
      
      if ((wy*signy <= 0) && (0 <= wx*signx))
      {
         //DEBUG
         System.out.println("collision with half-edge!");
         
         //the point at which we entered the cell is in a half-edge, so we have an instant collision
         out.x = px;
         out.y = py;
         return true;
      }
      else
      {
         
         if (0 <= (sx*dx + sy*dy))
         {
            //ray cannot hit tile as it is moving "away" from (OR PARALLEL TO!!) the slope
            System.out.println("...ray and slope are parallel!");
            return false;
         }
         else
         {
            //test for intersection between ray and slope; use ray-vs-lineseg intersection
            //
            // from geometry algorithms, the parameter describing the intersection point on the
            // tile's slope-lineseg is:
            //
            // s = (vy*wx - vx*wy) / (vx*uy - vy*ux)
            //
            // where v is the ray's direction, u is the lineseg's UNNORMALIZED! direction, and w is a vector from 
            // a point on the ray to a point on the lineseg.
            
            //we're going to use a weird version: calculate the parameter for the lineseg from the 
            //tile's center to one of it's corners; if the abs value of the parameter is <= 1, then
            //the ray intersects the slope lineseg
   
            //we now need to shift the test point to the midpoint of the slope lineseg
            double vx = signx*0.5*xw;
            wx += vx;
   
            double ux = -0.5*signy*xw;//the tile's normal's RHnormal gives us the tile's slope vector
            double uy = signx*yw;//(note that we could precalc this!)
            
            double s = (dy*wx - dx*wy) / (dx*uy - dy*ux);//note that the denominator will only be 0 if the ray and lineseg
                                             //are parallel, which we already checked for above..
            
            //DEBUG
            System.out.println("s: " + s);
            
            
            if (Math.abs(s) <= 1)
            {
               //collision
               out.x = (pos.x + vx) + s*ux;
               out.y = pos.y + s*uy;
               return true;
            }
            else
               return false;
         }
      }
   }

   private boolean testRay_67DegS(Vector2 out, double px, double py, double dx, double dy)
   {
      //DEBUG
      System.out.println("--- intersecting ray vs. 67DegS tile!");
      
      double sx = this.sx;
      double sy = this.sy;
      int signx = this.signx;
      int signy = this.signy;
      
      //we need to first test if we started on the tile's half-edge
      double wx = pos.x - px;//the test point is the corner joining the slpoe to the halfedge
      double wy = (pos.y-signy*yw) - py;
      
      if ((0 <= wx*signx) && (0 <= wy*signy))
      {
         //DEBUG
         System.out.println("collision with half-edge!");
         
         //the point at which we entered the cell is in a half-edge, so we have an instant collision
         out.x = px;
         out.y = py;
         return true;
      }
      else
      {
         if (0 <= (sx*dx + sy*dy))
         {
            //ray cannot hit tile as it is moving "away" from (OR PARALLEL TO!!) the slope
            System.out.println("...ray and slope are parallel!");
            return false;
         }
         else
         {
            //test for intersection between ray and slope; use ray-vs-lineseg intersection
            //
            // from geometry algorithms, the parameter describing the intersection point on the
            // tile's slope-lineseg is:
            //
            // s = (vy*wx - vx*wy) / (vx*uy - vy*ux)
            //
            // where v is the ray's direction, u is the lineseg's UNNORMALIZED! direction, and w is a vector from 
            // a point on the ray to a point on the lineseg.
            
            //we're going to use a weird version: calculate the parameter for the lineseg from the 
            //tile's center to one of it's corners; if the abs value of the parameter is <= 1, then
            //the ray intersects the slope lineseg
   
            //we now need to shift the test point to the midpoint of the slope lineseg
            wy += signy*yw;//the test point is the corner joining the slpoe to the halfedge
            double vx = signx*0.5*xw;
            wx -= vx;
   
            double ux = -0.5*signy*xw;//the tile's normal's RHnormal gives us the tile's slope vector
            double uy = signx*yw;//(note that we could precalc this!)
            
            double s = (dy*wx - dx*wy) / (dx*uy - dy*ux);//note that the denominator will only be 0 if the ray and lineseg
                                             //are parallel, which we already checked for above..
            
            //DEBUG
            System.out.println("s: " + s);
            
            
            if (Math.abs(s) <= 1)
            {
               //collision
               out.x = (pos.x  - vx)+ s*ux;
               out.y = pos.y + s*uy;
               return true;
            }
            else
               return false;
         }
      }
   }

   private boolean testRay_22DegB(Vector2 out, double px, double py, double dx, double dy)
   {
      //DEBUG
      System.out.println("--- intersecting ray vs. 22DegB tile!");
      
      double sx = this.sx;
      double sy = this.sy;
      int signx = this.signx;
      int signy = this.signy;
      
      //we need to first test if we started on the tile's half-edge
      double wx = pos.x - px;//the test point is the corner joining the slpoe to the halfedge
      double wy = pos.y - py;
      
      if ((wx*signx <= 0) && (0 <= wy*signy))
      {
         //DEBUG
         System.out.println("collision with half-edge!");
         
         //the point at which we entered the cell is in a half-edge, so we have an instant collision
         out.x = px;
         out.y = py;
         return true;
      }
      else
      {
         
         if (0 <= (sx*dx + sy*dy))
         {
            //ray cannot hit tile as it is moving "away" from (OR PARALLEL TO!!) the slope
            System.out.println("...ray and slope are parallel!");
            return false;
         }
         else
         {
            //test for intersection between ray and slope; use ray-vs-lineseg intersection
            //
            // from geometry algorithms, the parameter describing the intersection point on the
            // tile's slope-lineseg is:
            //
            // s = (vy*wx - vx*wy) / (vx*uy - vy*ux)
            //
            // where v is the ray's direction, u is the lineseg's UNNORMALIZED! direction, and w is a vector from 
            // a point on the ray to a point on the lineseg.
            
            //we're going to use a weird version: calculate the parameter for the lineseg from the 
            //tile's center to one of it's corners; if the abs value of the parameter is <= 1, then
            //the ray intersects the slope lineseg
   
            //we now need to shift the test point to the midpoint of the slope lineseg
            double vy = signy*0.5*yw;
            wy += vy;
   
            double ux = -signy*xw;//the tile's normal's RHnormal gives us the tile's slope vector
            double uy = 0.5*signx*yw;//(note that we could precalc this!)
            
            double s = (dy*wx - dx*wy) / (dx*uy - dy*ux);//note that the denominator will only be 0 if the ray and lineseg
                                             //are parallel, which we already checked for above..
            
            //DEBUG
            System.out.println("s: " + s);
            
            
            if (Math.abs(s) <= 1)
            {
               //collision
               out.x = pos.x + s*ux;
               out.y = (pos.y + vy) + s*uy;
               return true;
            }
            else
               return false;
         }
      }
   }

   private boolean testRay_22DegS(Vector2 out, double px, double py, double dx, double dy)
   {
      //DEBUG
      System.out.println("--- intersecting ray vs. 22DegS tile!");
      
      
      double sx = this.sx;
      double sy = this.sy;
      int signx = this.signx;
      int signy = this.signy;
      
      //we need to first test if we started on the tile's half-edge
      double wx = (pos.x-signx*xw) - px;//the test point is the corner joining the slpoe to the halfedge
      double wy = pos.y - py;
      
      if ((0 <= wx*signx) && (0 <= wy*signy))
      {
         //DEBUG
         System.out.println("collision with half-edge!");
         
         //the point at which we entered the cell is in a half-edge, so we have an instant collision
         out.x = px;
         out.y = py;
         return true;
      }
      else
      {
         
         if (0 <= (sx*dx + sy*dy))
         {
            //ray cannot hit tile as it is moving "away" from (OR PARALLEL TO!!) the slope
            System.out.println("...ray and slope are parallel!");
            return false;
         }
         else
         {
            //test for intersection between ray and slope; use ray-vs-lineseg intersection
            //
            // from geometry algorithms, the parameter describing the intersection point on the
            // tile's slope-lineseg is:
            //
            // s = (vy*wx - vx*wy) / (vx*uy - vy*ux)
            //
            // where v is the ray's direction, u is the lineseg's UNNORMALIZED! direction, and w is a vector from 
            // a point on the ray to a point on the lineseg.
            
            //we're going to use a weird version: calculate the parameter for the lineseg from the 
            //tile's center to one of it's corners; if the abs value of the parameter is <= 1, then
            //the ray intersects the slope lineseg
   
            //we now need to shift the test point to the midpoint of the slope lineseg
            wx += signx*xw;//the test point is the corner joining the slpoe to the halfedge
            double vy = signy*0.5*yw;
            wy -= vy;
   
            double ux = -signy*xw;//the tile's normal's RHnormal gives us the tile's slope vector
            double uy = 0.5*signx*yw;//(note that we could precalc this!)
            
            double s = (dy*wx - dx*wy) / (dx*uy - dy*ux);//note that the denominator will only be 0 if the ray and lineseg
                                             //are parallel, which we already checked for above..
            
            //DEBUG
            System.out.println("s: " + s);
            
            
            if (Math.abs(s) <= 1)
            {
               //collision
               out.x = pos.x + s*ux;
               out.y = (pos.y - vy) + s*uy;
               return true;
            }
            else
               return false;
         }
      }
   }

   private boolean testRay_Half(Vector2 out, double px, double py, double dx, double dy)
   {
      //DEBUG
      System.out.println("--- intersecting ray vs. half tile!");
      
      double sx = this.signx;
      double sy = this.signy;
      
      //we need to first test if we started on one of the tile's half-edges
      double wx = pos.x - px;
      double wy = pos.y - py;
      
      if (0 <= (wx*sx + wy*sy))
      {
         //DEBUG
         System.out.println("collision with half-edge!");
         
         //the point at which we entered the cell is in a half-edge, so we have an instant collision
         out.x = px;
         out.y = py;
         return true;
      }
      else
      {
         
         if (0 <= (sx*dx + sy*dy))
         {
            //ray cannot hit tile as it is moving "away" from (OR PARALLEL TO!!) the slope
            System.out.println("...ray and slope are parallel!");
            return false;
         }
         else
         {
            //test for intersection between ray and slope; use ray-vs-lineseg intersection
            //
            // from geometry algorithms, the parameter describing the intersection point on the
            // tile's slope-lineseg is:
            //
            // s = (vy*wx - vx*wy) / (vx*uy - vy*ux)
            //
            // where v is the ray's direction, u is the lineseg's UNNORMALIZED! direction, and w is a vector from 
            // a point on the ray to a point on the lineseg.
            
            //we're going to use a weird version: calculate the parameter for the lineseg from the 
            //tile's center to one of it's corners; if the abs value of the parameter is <= 1, then
            //the ray intersects the slope lineseg
            
            double ux = sy*xw;//this gives is the vector from the tile center to a halfedge corner; note
            double uy = sx*yw;//that we could precalc this!
            
   
            double s = (dy*wx - dx*wy) / (dx*uy - dy*ux);//note that the denominator will only be 0 if the ray and lineseg
                                             //are parallel, which we already checked for above..
            
            //DEBUG
            System.out.println("s: " + s);
            
            
            if (Math.abs(s) <= 1)
            {
               //collision
               out.x = pos.x + s*ux;
               out.y = pos.y + s*uy;
               return true;
            }
            else
               return false;
         }
      }
   }

   private boolean testRay_Convex(Vector2 out, double px, double py, double dx, double dy)
   {
      //DEBUG
      System.out.println("--- intersecting ray vs. convex tile!");
      
      
      int sx = signx;
      int sy = signy;
      
      //we use gamasutra swept-circle code to test for collision;
      //we know that the ray doens't collide initially (since it starts on a cell edge)
      //
            
      double abx = px - (pos.x - (sx*xw));//(abx,aby) should now be the vector from the circle's center to the
      double aby = py - (pos.y - (sy*yw)) ;//ray position px,py
      
      double a = (dx*dx)+(dy*dy);// const SCALAR a = vab.dot(vab); u*u coefficient
      double b = 2*( (dx*abx) + (dy*aby) );// const SCALAR b = 2*vab.dot(AB); u coefficient
      double rad = xw*2;
      double c = ((abx*abx)+(aby*aby)) - (rad*rad);// const SCALAR c = AB.dot(AB) - rab*rab; constant term
            
      double q = b*b - (4*a*c);// const SCALAR q = b*b - 4*a*c; 
      
      if (0 <= q)
      {
         //if ( q >= 0 )
      
         double sq = Math.sqrt(q);// const SCALAR sq = sqrt(q);
         double d = 1/(2*a);// const SCALAR d = 1 / (2*a);
         double u0 = (-b + sq) * d;
         double u1 = (-b - sq) * d;

         //DEBUG
         System.out.println("roots: " + u0 + "," + u1);

         if ( u1 < u0 )
         {
            //u1 is time of first collision
            out.x = px + u1*dx;
            out.y = py + u1*dy;
         }
         else
         {
            //u0 is time of first collision; u1 is our solution
            out.x = px + u0*dx;
            out.y = py + u0*dy;              
         }
      
         return true;//real roots
      }
      else
      {
         //DEBUG
         System.out.println("complex: " + q);
         
         //complex roots
         return false;
      }     
   }

   //NOTE: this is currently buggy!! specifically, we reject any rays which
   //have slopes pointing in the same direction as the tile normal; this is tru for 45deg tiles,
   //but for concave tiles, if the ray STARTS in the tile between the diagonal and the circle surface,
   //this breaks.. however, we need this kind of condition or else we'll find
   //intersection points outside of the tile!!
   
   
   private boolean testRay_Concave(Vector2 out, double px, double py, double dx, double dy)
   {
      //DEBUG
      System.out.println("--- intersecting ray vs. concave tile!");
      
      
      int sx = signx;
      int sy = signy;
      
      if (0 <= (sx*dx + sy*dy))
      {
         //ray cannot hit tile as it is moving "away" from (OR PARALLEL TO!!) the slope
         System.out.println("...ray and slope are parallel!");
         return false;
      }
      else
      {
         //test for intersection between ray and slope; use ray-vs-lineseg intersection
         //
         // from geometry algorithms, the parameter describing the intersection point on the
         // tile's slope-lineseg is:
         //
         // s = (vy*wx - vx*wy) / (vx*uy - vy*ux)
         //
         // where v is the ray's direction, u is the lineseg's UNNORMALIZED! direction, and w is a vector from 
         // a point on the ray to a point on the lineseg.
         
         //we're going to use a weird version: calculate the parameter for the lineseg from the 
         //tile's center to one of it's corners; if the abs value of the parameter is <= 1, then
         //the ray intersects the slope lineseg
         
         double ux = sx*xw;//this gives is the vector from the tile center to a corner; note
         double uy = -sy*yw;//that we could precalc this!
         
         double wx = pos.x - px;
         double wy = pos.y - py;
         
         double s = (dy*wx - dx*wy) / (dx*uy - dy*ux);//note that the denominator will only be 0 if the ray and lineseg
                                          //are parallel, which we already checked for above..
         
         //DEBUG
         System.out.println("s: " + s);
            
         if (Math.abs(s) <= 1)
         {
            //the ray collides with the diagonal so it MUST collide with the circle surface; find the
            //last time of intersection between the ray and the circle
            //(use swept-circle code from gamasutra)
            
            double abx = -ux - wx;//(abx,aby) should now be the vector from the circle's center to the
            double aby = uy - wy ;//ray position px,py
            
            double a = (dx*dx)+(dy*dy);// const SCALAR a = vab.dot(vab); u*u coefficient
            double b = 2*( (dx*abx) + (dy*aby) );// const SCALAR b = 2*vab.dot(AB); u coefficient
            double rad = xw*2;
            double c = ((abx*abx)+(aby*aby)) - (rad*rad);// const SCALAR c = AB.dot(AB) - rab*rab; constant term
                  
            double q = b*b - (4*a*c);// const SCALAR q = b*b - 4*a*c; 
            
            if (0 <= q)
            {
               //if ( q >= 0 )
            
               double sq = Math.sqrt(q);// const SCALAR sq = sqrt(q);
               double d = 1/(2*a);// const SCALAR d = 1 / (2*a);
               double u0 = (-b + sq) * d;
               double u1 = (-b - sq) * d;
            
               if ( u1 < u0 )
               {
                  //u1 is time of first collision; u0 is our solution
                  out.x = px + u0*dx;
                  out.y = py + u0*dy;
               }
               else
               {
                  //u0 is time of first collision; u1 is our solution
                  out.x = px + u1*dx;
                  out.y = py + u1*dy;              
               }
            
               return true;//real roots
            }
            else
            {
               //complex roots
               return false;
            }     
   
         }
         else
            return false;
      }
   }

   private boolean testRay_45Deg(Vector2 out, double px, double py, double dx, double dy)
   {
      //DEBUG
      System.out.println("--- intersecting ray vs. 45deg tile!");
      
      int sx = signx;
      int sy = signy;
      
      if (0 <= (sx*dx + sy*dy))
      {
         //ray cannot hit tile as it is moving "away" from (OR PARALLEL TO!!) the slope
         System.out.println("...ray and slope are parallel!");
         return false;
      }
      else
      {
         //test for intersection between ray and slope; use ray-vs-lineseg intersection
         //
         // from geometry algorithms, the parameter describing the intersection point on the
         // tile's slope-lineseg is:
         //
         // s = (vy*wx - vx*wy) / (vx*uy - vy*ux)
         //
         // where v is the ray's direction, u is the lineseg's UNNORMALIZED! direction, and w is a vector from 
         // a point on the ray to a point on the lineseg.
         
         //we're going to use a weird version: calculate the parameter for the lineseg from the 
         //tile's center to one of it's corners; if the abs value of the parameter is <= 1, then
         //the ray intersects the slope lineseg
         
         double ux = sx*xw;//this gives is the vector from the tile center to a corner; note
         double uy = -sy*yw;//that we could precalc this!
         
         double wx = pos.x - px;
         double wy = pos.y - py;
         
         double s = (dy*wx - dx*wy) / (dx*uy - dy*ux);//note that the denominator will only be 0 if the ray and lineseg
                                          //are parallel, which we already checked for above..
         
         //DEBUG
         System.out.println("s: " + s);
         
         
         if (Math.abs(s) <= 1)
         {
            //collision
            out.x = pos.x + s*ux;
            out.y = pos.y + s*uy;
            return true;
         }
         else
            return false;
      }
   }

   //this method should NEVER be called, because
   //rays are clipped to solid edges, and full tiles are ALL solid edges, so
   //if we call this, it means that either the ray has tunneled through
   //a solid edge or through a tile adjacent to this full tile..
   private boolean testRay_Full(Vector2 out, double px, double py, double dx, double dy)
   {
      System.out.println("WARNING!!!!! testRay_Full was called -- this should NEVER happen unless ray has tunneled!!");
      return false;
   }   
}
